#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "detri2.h"

using  namespace detri2;

//==============================================================================

int Triangulation::get_vertex_metric(Vertex *v)
{
  if (OMT_domain) {
    // The metric is defined by a background mesh.
    // Locate the vertex in this background mesh.
    // assert(v->on_dm.tri != NULL);
    //int loc = OMT_domain->locate_point(v, v->on_dm, 1, 0);
    int loc = OMT_domain->locate_point(v, v->on_dm, 0);

    if (loc == LOC_ON_VERT) {
      // Direct copy the data.
      Vertex *vdm = v->on_dm.org();
      for (int i = 0; i < 3; i++) {
        v->mtr[i] = vdm->mtr[i];
      }
      v->val = vdm->val; // Will not be used in the future.
      return 1;
    }

    if ((loc == LOC_IN_OUTSIDE) || (loc == LOC_ENC_SEG)) {
      assert(0); // The background mesh does not cover the mesh domain.
    }

    if (v->on_dm.tri->is_hulltri()) {
      if (loc == LOC_ON_EDGE) {
        v->on_dm = v->on_dm.esym();
      } else {
        assert(0); // Unknown case.
      }
    }
    assert(!v->on_dm.tri->is_hulltri());

    Vertex *pa = v->on_dm.org();
    Vertex *pb = v->on_dm.dest();
    Vertex *pc = v->on_dm.apex();
    double A  = get_tri_area(pa, pb, pc);
    double Wa = get_tri_area( v, pb, pc);
    double Wb = get_tri_area(pa,  v, pc);
    double Wc = get_tri_area(pa, pb,  v);

    //*val = (pa->val * Wa + pb->val * Wb + pc->val * Wc) / A;
    if (op_metric == METRIC_EUCLIDEAN) { // default
      // [2018-11-3] We will use mtr in the future.
      if ((pa->val > 0.) && (pb->val > 0.) && (pc->val > 0.)) {
        v->val = (pa->val * Wa + pb->val * Wb + pc->val * Wc) / A;
      } else {
        v->val = 0.0; // not defined.
      }
    } else if (op_metric == METRIC_ISO) {
      v->mtr[0] = (pa->mtr[0] * Wa + pb->mtr[0] * Wb + pc->mtr[0] * Wc) / A;
      v->mtr[1] = 0.0;
      v->mtr[2] = v->mtr[0];
    } else if (op_metric == METRIC_ANISO) {
      // ...
    } else if (op_metric == METRIC_HDE) {
      for (int i = 0; i < 3; i++) {
        v->mtr[i] = (pa->mtr[i] * Wa + pb->mtr[i] * Wb + pc->mtr[i] * Wc) / A;
      }
    } else {
      // assert(0); // Unknown metric
    }
  } else { // No background mesh.
    // The metric is given by implicit functions.
    // to do ...
    assert(0);
  }

  return 1;
}

//==============================================================================

REAL Triangulation::get_norm(Vertex *v1, Vertex *v2)
{
  REAL norm = 0.;

  if (op_metric == METRIC_EUCLIDEAN) {
    REAL vx = v2->crd[0] - v1->crd[0];
    REAL vy = v2->crd[1] - v1->crd[1];
    norm = (vx * vx + vy * vy);
  } else if (op_metric == METRIC_ISO) {
    REAL vx = v2->crd[0] - v1->crd[0];
    REAL vy = v2->crd[1] - v1->crd[1];
    norm = (vx * vx + vy * vy);
  } else if (op_metric == METRIC_ANISO) {
    // ...
  } else if (op_metric == METRIC_HDE) {
    REAL vec[5];
    get_vertex_metric(v1);
    get_vertex_metric(v2);
    // We assume the values of v1 and v2 are not normalised
    REAL nrm1 = sqrt(v1->mtr[1]*v1->mtr[1] + v1->mtr[2]*v1->mtr[2] + 1.0);
    REAL nrm2 = sqrt(v2->mtr[1]*v2->mtr[1] + v2->mtr[2]*v2->mtr[2] + 1.0);
    for (int i = 1; i < 3; i++) v1->mtr[i] /= nrm1;
    for (int i = 1; i < 3; i++) v2->mtr[i] /= nrm2;
    vec[0] = (v2->crd[0] - v1->crd[0]) / io_diagonal;
    vec[1] = (v2->crd[1] - v1->crd[1]) / io_diagonal;
    vec[2] = (v2->mtr[0] - v1->mtr[0]) / io_max_metric * op_hde_s1; // f
    vec[3] = (v2->mtr[1] - v1->mtr[1]) * op_hde_s2; // fx
    vec[4] = (v2->mtr[2] - v1->mtr[2]) * op_hde_s2; // fy
    for (int i = 0; i < 5; i++) {
      norm += (vec[i] * vec[i]);
    }
  } else {
    //assert(0); // Unknown metric
  }

  return norm;
}

REAL Triangulation::get_distance(Vertex *v1, Vertex *v2)
{
  REAL distsum = 0.0;

  if (op_ada_use_intpoints > 0) {
    // Get interpolating points between v1 and v2.
    Vertex IntV[16], *mid;
    REAL s = 1.0 / ((REAL)(op_ada_use_intpoints + 1));
    for (int k = 0; k < op_ada_use_intpoints; k++) {
      mid = &(IntV[k]);
      mid->init();
      REAL t = ((REAL) (k+1) * s);
      mid->crd[0] = v1->crd[0] + t * (v2->crd[0] - v1->crd[0]);
      mid->crd[1] = v1->crd[1] + t * (v2->crd[1] - v1->crd[1]);
    } // k
    distsum = sqrt(get_norm(v1, &(IntV[0])));
    int i = 1;
    for (; i < op_ada_use_intpoints; i++) {
      distsum += sqrt(get_norm(&(IntV[i-1]), &(IntV[i])));
    }
    distsum += sqrt(get_norm(&(IntV[i-1]), v2));
  } else {
    distsum = sqrt(get_norm(v1, v2));
  }

  return distsum;
}

// angle at v0 in [0,pi].
REAL Triangulation::get_angle(Vertex* v0, Vertex* v1, Vertex* v2)
{
  // Use the law of cosines: cos(C) = (a^2 + b^2 - c^2) / (2ab);
  // see https://en.wikipedia.org/wiki/Law_of_cosines

  REAL a = get_distance(v0, v1);
  REAL b = get_distance(v0, v2);
  REAL c = get_distance(v1, v2);

  REAL cosC = (a*a + b*b - c*c) / (2.0*a*b);
  if (cosC > 1.0) cosC = 1.0;
  if (cosC < -1.0) cosC = -1.0;

  return acos(cosC);
}

//==============================================================================

REAL Triangulation::get_tri_area(Vertex* pa, Vertex* pb, Vertex* pc)
{
  REAL detleft = (pa->crd[0] - pc->crd[0]) * (pb->crd[1] - pc->crd[1]);
  REAL detright = (pa->crd[1] - pc->crd[1]) * (pb->crd[0] - pc->crd[0]);
  REAL det = detleft - detright;
  return 0.5 * fabs(det);
}

//==============================================================================
// Return the shortest edge length at this vertex.
REAL Triangulation::get_vertex_size(Vertex *pt)
{
  REAL lmin = 0., l;
  TriEdge E = pt->adj;
  do {
    if (E.dest() != tr_infvrt) { 
      l = get_distance(pt, E.dest());
      if (lmin == 0.) {
        lmin = l;
      } else {
        lmin = (l < lmin) ? l : lmin;
      }
    }
    E = E.eprev_esym();
  } while (E.tri != pt->adj.tri);
  assert(lmin > 0.);
  return lmin;
}

REAL Triangulation::get_vertex_size_average(Vertex *pt)
{
  REAL lmax, lmin, lave, l;

  int ecount = 0;
  lmax = lmin = lave = 0.0;
  TriEdge E = pt->adj;
  do {
    // Skip the edge if it is a hull edge.
    if (E.dest() != tr_infvrt) { 
      l = get_distance(pt, E.dest());
      if (ecount == 0) {
        lmax = lmin = l;
      } else {
        lmax = (l > lmax) ? l : lmax;
        lmin = (l < lmin) ? l : lmin;
      }
      lave += l;
      ecount++;
    }
    E = E.eprev_esym();
  } while (E.tri != pt->adj.tri);

  lave -= (lmax + lmin);
  lave /= (ecount - 2);
  return lave;
}

//==============================================================================
// Given two triangles [a,b,c] and [b,a,d], check if the edge [a,b]
//   is locally regular, i.e., locally Delaunay in Euclidean metric.
//   Return 1 (non-regular), 0 (either regular or degenerate, co-hyperplane)

REAL Triangulation::regular_test(Vertex* pa, Vertex* pb, Vertex* pc, Vertex* pd)
{
  if ((op_metric == METRIC_EUCLIDEAN) || (op_metric == METRIC_ISO)) {
    //printf("  O3d: (%d, %d, %d, %d)\n", tt[0].org()->idx, tt[0].dest()->idx, tt[0].apex()->idx, tt[1].apex()->idx);
    return Orient3d(pa, pb, pc, pd) * op_dt_nearest;
    //printf("  O3d = %g, op_dt_nearest=%d\n", ori, op_dt_nearest);
  }

  REAL l_ab = get_distance(pa, pb);
  REAL l_cd = get_distance(pc, pd);
  REAL l_bc = get_distance(pb, pc);
  REAL l_ca = get_distance(pc, pa);
  REAL l_ad = get_distance(pa, pd);
  REAL l_db = get_distance(pd, pb);
  
  REAL cosa = (l_ca*l_ca + l_ad*l_ad - l_cd*l_cd) / (2.0 * l_ca*l_ad);
  REAL cosb = (l_bc*l_bc + l_db*l_db - l_cd*l_cd) / (2.0 * l_bc*l_db);
  REAL cosc = (l_bc*l_bc + l_ca*l_ca - l_ab*l_ab) / (2.0 * l_bc*l_ca);
  REAL cosd = (l_ad*l_ad + l_db*l_db - l_ab*l_ab) / (2.0 * l_ad*l_db);
  
  // Rounding
  if (cosa >  1.0) cosa =  1.0;
  if (cosa < -1.0) cosa = -1.0;
  if (cosb >  1.0) cosb =  1.0;
  if (cosb < -1.0) cosb = -1.0;
  if (cosc >  1.0) cosc =  1.0;
  if (cosc < -1.0) cosc = -1.0;
  if (cosd >  1.0) cosd =  1.0;
  if (cosd < -1.0) cosd = -1.0;
    
  REAL angA = acos(cosa);
  REAL angB = acos(cosb);
  REAL angC = acos(cosc);
  REAL angD = acos(cosd);
  
  REAL sum_ab = angA + angB;
  REAL sum_cd = angC + angD;
  if ((fabs(sum_ab - PI) / PI) < 1.e-4) sum_ab = PI; // Roundoff
  if ((fabs(sum_cd - PI) / PI) < 1.e-4) sum_cd = PI;

  // Is regular (need flip)?
  if ((sum_ab < PI) && (sum_ab > sum_cd)) {
    return 1.0; //     ori = 1.0; // need to flip.
  } else {
    return 0.0; // is regular
  }
}

//==============================================================================

int Triangulation::lawson_flip(Vertex *pt, int hullflag, arraypool *fqueue)
{
  TriEdge E, N, tt[4];
  Vertex *delpt = NULL;
  REAL ori;
  int ishullflip;
  int fcount = 0;

  if (fqueue->objects == 0) {
    if (pt != NULL) {
      // Collect the link edges of pt.
      E = pt->adj;
      do {
        N = E.enext();
        N.set_edge_infect();
        * (TriEdge *) fqueue->alloc() = N;
        E = E.eprev_esym(); // ccw rotate
      } while (E.tri != pt->adj.tri);
    } else {
      // Collect all edges of the triangulation.
      for (int i = 0; i < tr_tris->used_items; i++) {
        E.tri = (Triang *) tr_tris->get(i);
        if (E.tri->is_deleted()) continue;
        if (!E.tri->is_hulltri()) {
          for (E.ver = 0; E.ver < 3; E.ver++) {
            if (!E.esym().tri->is_infected()) {
              E.set_edge_infect();
              * (TriEdge *) fqueue->alloc() = E;
            }
          }
          E.tri->set_infect();
        }
      }
      // Uninfect all triangles.
      for (int i = 0; i < tr_tris->used_items; i++) {
        E.tri = (Triang *) tr_tris->get(i);
        if (E.tri->is_deleted()) continue;
        if (!E.tri->is_hulltri()) {
          E.tri->clear_infect();
        }
      }
    }
  }

  if (op_db_verbose > 2) {
    printf("    Lawson flipping %d edges.\n", fqueue->objects);
  }

  for (int i = 0; i < fqueue->used_items; i++) {
    TriEdge *pte = (TriEdge *) fqueue->get(i);
    if (pte->tri->is_deleted()) continue;
    if (!pte->is_edge_infected()) continue; // is it still in queue?
    pte->clear_edge_infect();
    if (pt != NULL) {
      if (pte->apex() != pt) continue;
    }
    // Check if this edge is locally regular.
    ori = 0; ishullflip = 0;
    tt[0] = *pte;
    tt[1] = tt[0].esym();
    if (!tt[0].tri->is_hulltri()) {
      if (!tt[1].tri->is_hulltri()) { 
        // An interior edge.
        if (op_db_verbose > 3) {
          printf("      O3d: (%d, %d, %d, %d)\n", tt[0].org()->idx, tt[0].dest()->idx, tt[0].apex()->idx, tt[1].apex()->idx);
        }
        /*
        if ((op_metric == METRIC_EUCLIDEAN) || (op_metric == METRIC_ISO)) {
          //printf("  O3d: (%d, %d, %d, %d)\n", tt[0].org()->idx, tt[0].dest()->idx, tt[0].apex()->idx, tt[1].apex()->idx);
          ori = Orient3d(tt[0].org(), tt[0].dest(), tt[0].apex(), tt[1].apex())
              * op_dt_nearest;
          //printf("  O3d = %g, op_dt_nearest=%d\n", ori, op_dt_nearest);
        } else {
          ori = regular_test(tt[0].org(), tt[0].dest(), tt[0].apex(), tt[1].apex());
        }
        */
        ori = regular_test(tt[0].org(), tt[0].dest(), tt[0].apex(), tt[1].apex());
      }
    } else {
      if (tt[1].tri->is_hulltri()) {
        // Only do flip if it is the Delaunay case.
        if (hullflag) {
          // Check if an exterior flip (for convexity) is needed.
          if (tt[0].org() != tr_infvrt) tt[0] = tt[0].esym();
          assert(tt[0].org() == tr_infvrt);
          tt[1] = tt[0].esym();
          if (op_db_verbose > 3) {
            printf("      O2d: (%d, %d, %d)\n", tt[1].apex()->idx, tt[1].org()->idx, tt[0].apex()->idx);
          }
          ori = Orient2d(tt[1].apex(), tt[1].org(), tt[0].apex());
          ishullflip = 1;
        }
      }
    }
    if (ori > 0) {      
      // This edge is either locally non-regular or non-convex.
      if (op_no_incremental_flip && !ishullflip) {
        continue; // Ignore this flip.
      }
      int fflag = FLIP_UNKNOWN;
      if (flip(tt, &delpt, fflag, fqueue)) {
        fcount++;
      } 
    }
  } // i

  fqueue->clean(); // clean it.
  return fcount;
}

//==============================================================================

int Triangulation::get_edge_bisector(Vertex *pa, Vertex *pb, Vertex *mid,
                                     REAL *lo, REAL *hi, int iter)
{ // Using the simple bisector algorithm.
  REAL split = 0.5 * (*lo + *hi);
  if (op_db_verbose > 2) {
    printf("      lo = %g, hi = %g, split = %g, iter = %d\n", *lo, *hi, split, iter);
  }
  mid->crd[0] = (1.0 - split) * pa->crd[0] + split * pb->crd[0];
  mid->crd[1] = (1.0 - split) * pa->crd[1] + split * pb->crd[1];
  mid->crd[2] = (1.0 - split) * pa->crd[2] + split * pb->crd[2]; // No use
  
  if (!get_vertex_metric(mid)) {
    assert(0); // debug it.
    return 0;
  }

  REAL dist_to_a = get_distance(pa, mid);
  REAL dist_to_b = get_distance(pb, mid);
  if (op_db_verbose > 2) {
    printf("      dist_to_a=%g, dist_to_b=%g\n", dist_to_a, dist_to_b);
  }

  // Check if this is the midpoint (or different to it not much).
  REAL diff = dist_to_a - dist_to_b;
  if ((fabs(diff) / (dist_to_a + dist_to_a)) < 0.3) {
    return iter; // Found! Return the midpoint.
  }

  if (iter > 0) {
    int next_iter = iter - 1;
    if (dist_to_a > dist_to_b) {
      // move it towards to pa.
      *hi = split;
    } else {
      // move it towards to pb.
      *lo = split;
    }
    return get_edge_bisector(pa, pb, mid, lo, hi, next_iter);
  } else {
    return 0; // Not converged. Return the midpoint.
  }
}

//==============================================================================
int Triangulation::split_edge(TriEdge &E, double t, Vertex** ppt, arraypool* fqueue)
{
  Vertex *endpts[2];
  endpts[0] = E.org();
  endpts[1] = E.dest();

  if (op_db_verbose > 2) {
    printf("      Split edge [%d, %d]\n", endpts[0]->idx, endpts[1]->idx);
  }

  Vertex *newvrt = (Vertex *) tr_steiners->alloc();
  newvrt->init();
  newvrt->idx = io_firstindex + (ct_in_vrts + tr_steiners->objects - 1);
  newvrt->typ = STEINERVERTEX;

  if (t > 0. && t < 1.) {
    newvrt->crd[0] = (1.-t) * endpts[0]->crd[0] + t * endpts[1]->crd[0];
    newvrt->crd[1] = (1.-t) * endpts[0]->crd[1] + t * endpts[1]->crd[1];
    newvrt->crd[2] = (1.-t) * endpts[0]->crd[2] + t * endpts[1]->crd[2]; // no use.
  } else {
    // Get the bisector point on this edge.
    REAL lo = 0., hi = 1.0;
    int it = get_edge_bisector(endpts[0], endpts[1], newvrt, &lo, &hi, op_edge_bisect_iter);
    if (op_db_verbose > 3) {
      REAL split = 0.5 * (lo + hi);
      //printf("      lo = %g, hi = %g, split = %g, iterations = %d, .\n", lo, hi, split,
      //       op_edge_bisect_iter - it);
    }
  }

  // Validate if the new vertex does not result inverted triangles.
  TriEdge tt[4];
  tt[0] = E;
  tt[1] = tt[0].esym();
  Vertex *pa = tt[0].org();
  Vertex *pb = tt[0].dest();
  Vertex *pc = tt[0].apex();
  Vertex *pd = tt[1].apex();
  REAL o1, o2, o3, o4;
  if (pc != tr_infvrt) {
    o1 = Orient2d(pa, newvrt, pc);
    o2 = Orient2d(newvrt, pb, pc);
  } else {
    o1 = o2 = 1.0;
  }
  if (pd != tr_infvrt) {
    o3 = Orient2d(pb, newvrt, pd);
    o4 = Orient2d(newvrt, pa, pd);
  } else {
    o3 = o4 = 1.0;
  }
  if ((o1 > 0.) && (o2 > 0.) && (o3 > 0.) && (o4 > 0.)) {
    // Insert the point.
    int fflag = FLIP_24;
    if (flip(tt, &newvrt, fflag, fqueue)) {
      lawson_flip(newvrt, 0, fqueue);
    } else {
      assert(0);
    }
    if (ppt != NULL) {
      *ppt = newvrt;
    }
    return 1;
  } else {
    if (op_db_verbose > 2) {
      printf("      Reject a new point on edge [%d,%d].\n", endpts[0]->idx, endpts[1]->idx);
    }
    newvrt->set_deleted();
    tr_steiners->dealloc((void *) newvrt);
    return 0;
  }
}

//==============================================================================
// Split edges which have too long size (wrt. the metric field).
// It assumes that: op_target_length > 0.
// Return the number of split edges

int Triangulation::split_edges()
{
  if (op_target_length <= 0.0) return 0;

  if (op_db_verbose) {
    printf("Splitting edges longer than target length (%g)...\n", op_target_length);
  }
  arraypool *edge_list = new arraypool(sizeof(Triang), 10);
  double len, diff, ratio;
  int split_count = 0;
  TriEdge E;
  
  TriEdge Emax;
  double max_len = 0.0;

  // Collect all edges to be split of the triangulation.
  for (int i = 0; i < tr_tris->used_items; i++) {
    E.tri = (Triang *) tr_tris->get(i);
    if (E.tri->is_deleted()) continue;
    if (!E.tri->is_hulltri()) {
      for (E.ver = 0; E.ver < 3; E.ver++) {
        if (!E.esym().tri->is_infected()) {
          len = get_distance(E.org(), E.dest());
          diff = len - op_target_length;
          ratio = len / op_target_length; // Roundoff
          if ((ratio > op_edge_split_factor) && (diff > 0)) {
            if (len > max_len) { // Record the maximum edge length.
              Emax = E;
              max_len = len;
            }
            Triang *seg = (Triang *) edge_list->alloc();
            seg->init();
            seg->vrt[0] = E.org();
            seg->vrt[1] = E.dest();
            seg->val = len; // Remember the length (for debugging).
          }
        }
      }
      E.tri->set_infect();
    }
  }
  // Uninfect all triangles.
  for (int i = 0; i < tr_tris->used_items; i++) {
    E.tri = (Triang *) tr_tris->get(i);
    if (E.tri->is_deleted()) continue;
    if (!E.tri->is_hulltri()) {
      E.tri->clear_infect();
    }
  }

  if (edge_list->objects > 0) {
    if (op_db_verbose) {
      printf("  Found %d long edges, maxlen at (%d,%d) = %g.\n", 
            edge_list->objects, Emax.org()->idx, Emax.dest()->idx, max_len);
    }
    //if (op_db_verbose > 2) {
      //FILE *fout = fopen("split_edges.lua", "w");
      //HDE_dump_splitting_edges(edge_list, fout);
      //fclose(fout);
    //}
    //mesh_phase = 4;
    arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);

    for (int i = 0; i < edge_list->objects; i++) {
      Triang *seg = (Triang *) edge_list->get(i);
      if (get_edge(seg->vrt[0], seg->vrt[1], E)) {
        if (op_db_verbose > 1) {
          printf("    splitting edge [%d,%d] len=%g\n", seg->vrt[0]->idx,
                 seg->vrt[1]->idx, seg->val);
        }
        if (split_edge(E, 0.5, NULL, fqueue)) { // t = 0.
          split_count++;
        }
      }
    }

    if (op_db_verbose > 1) {
      printf("  Split %d edges\n", split_count);
    }

    delete fqueue;
    //mesh_phase = 0;
  }

  delete edge_list;
  return split_count;
}

//==============================================================================
// Mesh coarsening (for adaptive mesh refinement)
// Return the number of removed vertices.

int Triangulation::coarsen_mesh()
{
  if (op_db_verbose) {
    printf("Mesh coarsening...\n");
  }
  arraypool *delptlist = new arraypool(sizeof(Vertex*), 8);

  if (op_coarse_mesh) { // -R
    // Delete all vertices except fixed vertices.
    if (op_db_verbose > 1) {
      printf("  Collecting all free vertices...\n");
    }
    for (int i = 0; i < ct_in_vrts; i++) {
      Vertex *v = &(in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      if (!v->is_fixed()) { //if (v->typ != RIDGEVERTEX) {
        * (Vertex **) delptlist->alloc() = v;
      }
    }
  }
  else {
  //if (op_metric == METRIC_ISO) { // -m
    if (op_db_verbose > 1) {
      printf("  Collecting vertices based on metric %d\n", op_metric);
    }
    double lmin, diff, tol;
    for (int i = 0; i < ct_in_vrts; i++) {
      Vertex *v = &(in_vrts[i]);
      if (v->typ == UNUSEDVERTEX) continue;
      assert(v->adj.tri != NULL); // Not a dangling vertex.
      //if (v->adj.tri != NULL) { // Not a dangling vertex.
      //if (((v->typ == FREEVERTEX) ||
      //     (v->typ == SEGMENTVERTEX)) &&
      //     ((v->tag <= 2) || (v->tag > 100))) { // do not split tag in [2,100].
      if (!v->is_fixed()) {
        lmin = get_vertex_size(v);
        //lmin = get_vertex_size_average(v);
        if (op_metric == METRIC_EUCLIDEAN) {
          if (get_vertex_metric(v)) {
            REAL val = v->val;
            if (val > 0.0) {
              diff = lmin - val;
              tol = fabs(diff) / val;
              if ((tol > 0.2) && (diff < 0)) {
                // Delete this vertex.
                v->set_infect();
                * (Vertex **) delptlist->alloc() = v;
              }
            }
          }
        } else {
          // Compare lmin with the desired unit edge length.
          // to be done...
        }
        if ((op_minlen > 0) && !v->is_infected()) { // -L#
          // Is this edge too short w.r.t. the minimum edge length option.
          diff = lmin - op_minlen;
          tol = fabs(diff) / op_minlen;
          if ((tol > 0.2) && (diff < 0)) {
            // Delete this vertex.
            v->set_infect();
            * (Vertex **) delptlist->alloc() = v;
          }
        }
      }
      //}
    }
    if (op_db_verbose > 1) {
      printf("  Collected %d input vertices.\n", delptlist->objects);
    }
    // Do the same check for Steiner points.
    if (tr_steiners != NULL) {
      // There are Steiner points.
      int bak_count = delptlist->objects;
      for (int i = 0; i < tr_steiners->used_items; i++) {
        Vertex *v = (Vertex *) tr_steiners->get(i);
        if (v->is_deleted()) continue;
        assert(v->typ == STEINERVERTEX);
        if (!v->is_fixed()) {
          lmin = get_vertex_size(v);
          //lmin = get_vertex_size_average(v);
          if (op_metric == METRIC_EUCLIDEAN) {
            if (get_vertex_metric(v)) {
              REAL val = v->val;
              diff = lmin - val;
              tol = fabs(diff) / val;
              if ((tol > 0.2) && (diff < 0)) {
                // Delete this vertex.
                v->set_infect();
                * (Vertex **) delptlist->alloc() = v;
              }
            }
          } else {
            // Compare lmin with the desired unit edge length.
            // to be done...
          }
          if ((op_minlen > 0) && !v->is_infected()) { // -L#
            diff = lmin - op_minlen;
            tol = fabs(diff) / op_minlen;
            if ((tol > 0.2) && (diff < 0)) {
              // Delete this vertex.
              v->set_infect();
              * (Vertex **) delptlist->alloc() = v;
            }
          }
        } // if (!v->is_fixed())
      }
      if (op_db_verbose > 1) {
        printf("  Collected %d Steiner vertices.\n", delptlist->objects - bak_count);
      }
    }
    
    if (0) { // [2018-11-10] temporarily disable this option.
      // Remove vertices which are too close to boundary.
      TriEdge E;
      for (int i = 0; i < tr_tris->objects; i++) {
        E.tri = (Triang *) tr_tris->get(i);
        if (E.tri->is_deleted()) continue;
        if (E.tri->is_hulltri() || E.tri->is_exterior()) continue;
        for (E.ver = 0; E.ver < 3; E.ver++) {
          if (E.is_segment()) {
            if (!E.apex()->is_infected()) {
              //if (((E.apex()->typ == FACETVERTEX) ||
              //     (E.apex()->typ == SEGMENTVERTEX)) &&
              //    ((E.apex()->tag <= 0) || (E.apex()->tag > 100))) {  // do not split (0,100].
              if (!E.apex()->is_fixed()) {
                // Check if this triangle has a bad angle at boundary.
                Vertex *e1 = E.org();
                Vertex *e2 = E.dest();
                Vertex *e3 = E.apex();
                double V12[2], V23[2], V31[2];
                V12[0] = e2->crd[0] - e1->crd[0];
                V12[1] = e2->crd[1] - e1->crd[1];
                V23[0] = e3->crd[0] - e2->crd[0];
                V23[1] = e3->crd[1] - e2->crd[1];
                V31[0] = e1->crd[0] - e3->crd[0];
                V31[1] = e1->crd[1] - e3->crd[1];
                double V12_norm = sqrt(V12[0]*V12[0]+V12[1]*V12[1]);
                double V23_norm = sqrt(V23[0]*V23[0]+V23[1]*V23[1]);
                double V31_norm = sqrt(V31[0]*V31[0]+V31[1]*V31[1]);
                double cos_e1 = (V12[0]*V31[0]+V12[1]*V31[1]) / (V12_norm*V31_norm);
                double cos_e2 = (V12[0]*V23[0]+V12[1]*V23[1]) / (V12_norm*V23_norm);
                // rounding
                if (cos_e1 > 1) cos_e1 = 1.0;
                if (cos_e1 < -1) cos_e1 = -1.0;
                if (cos_e2 > 1) cos_e2 = 1.0;
                if (cos_e2 < -1) cos_e2 = -1.0;
                double ang_e1 = acos(cos_e1);
                double ang_e2 = acos(cos_e2);
                if (((ang_e1 * 6.0) < PI) || ((ang_e2 * 6.0) < PI)) {
                  // Delete this vertex.
                  e3->set_infect();
                  * (Vertex **) delptlist->alloc() = e3;
                }
              }
            }
          }
        }
      }
    } // if (0)
  } // if (op_metric)

  int rcount = 0;
  if (delptlist->objects > 0) {
    if (op_db_verbose > 1) {
      printf("  Removing %d vertices...\n", (int) delptlist->objects);
    }
    //mesh_phase = 4;
    arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
    for (int i = 0; i < delptlist->objects; i++) {
      Vertex *v = * (Vertex **) delptlist->get(i);
      if (!v->is_deleted()) {
        if (v->typ != UNUSEDVERTEX) {
          if (op_db_verbose > 1) {
            printf("  Removing vertex [%d] typ(%d)\n", v->idx, (int) v->typ);
          }
          if (remove_point(v, fqueue)) {
            rcount++;
          }
          lawson_flip(NULL, 0, fqueue);
        }
      }
    }
    delete fqueue;
    //mesh_phase = 0;
    if (op_db_verbose > 1) {
      printf("  Removed %d vertices\n", rcount);
    }
  }

  delete delptlist;
  return rcount;
}

//==============================================================================

int Triangulation::get_powercell_mass_centers(Vertex *massptlist)
{
  if (op_db_verbose > 1) {
    printf("  Calculating %d orthocenters\n", tr_tris->objects + ct_hullsize);
  }
  assert(OMT_domain != NULL);

  int i, idx;
  idx = io_firstindex;
  // Calculate circumcenters for hull triangles.
  for (i = 0; i < tr_tris->used_items; i++) {
    Triang* tri = (Triang *) tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted() || tri->is_hulltri()) continue;
    get_tri_orthocenter(tri);
    tri->idx = idx; // hulltri is also indexed.
    idx++;
  }
  // Calculate bisectors for hull triangles.
  for (i = 0; i < tr_tris->used_items; i++) {
    Triang* tri = (Triang *) tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted()) continue;
    if (tri->is_hulltri()) { // A hull triangle.
      get_hulltri_orthocenter(tri);
    }
  }

  if (op_db_verbose > 1) {
    printf("  Calculating %d power cells and mass centers\n",
           ct_in_vrts + ((tr_steiners != NULL) ? tr_steiners->objects : 0));
  }
  int mcount = 0;

  // Calculate power cells and mass centers.
  idx = 0;
  for (i = 0; i < ct_in_vrts; i++) {
    Vertex *masspt = &(massptlist[idx]);
    masspt->init();
    idx++;

    Vertex *mesh_vert = &(in_vrts[i]);
    if (mesh_vert->typ == UNUSEDVERTEX) continue;
    if (mesh_vert->is_fixed()) continue;
    //printf("  idx = %d\n", i);
    Vertex *ptlist = NULL;
    int ptnum = 0;

    if (get_powercell(mesh_vert, &ptlist, &ptnum)) {
      // Calculate its mass center
      get_mass_center(ptlist, ptnum, masspt->crd);
      masspt->Pair = mesh_vert; // Remember it.
      mcount++;
      delete [] ptlist;
    } // if (Tr->get_powercell
  }

  if (tr_steiners != NULL) {
    for (i = 0; i < tr_steiners->used_items; i++) {
      Vertex *masspt = &(massptlist[idx]);
      masspt->init();
      idx++;
    
      Vertex *mesh_vert = (Vertex *) tr_steiners->get(i);
      if (mesh_vert->is_deleted()) continue;
      if (mesh_vert->is_fixed()) continue;
      //printf("  idx = %d\n", i);
      Vertex *ptlist = NULL;
      int ptnum = 0;

      if (get_powercell(mesh_vert, &ptlist, &ptnum)) {
        get_mass_center(ptlist, ptnum, masspt->crd);
        masspt->Pair = mesh_vert; // Remember it.
        mcount++;
        delete [] ptlist;
      } // if (Tr->get_powercell
    }
  }

  return mcount;
}

int Triangulation::get_laplacian_centers(Vertex *massptlist)
{
  if (op_db_verbose > 1) {
    printf("  Calculating %d Laplacian centers\n",
           ct_in_vrts + ((tr_steiners != NULL) ? tr_steiners->objects : 0));
  }
  int mcount = 0;

  int idx = 0;
  for (int i = 0; i < ct_in_vrts; i++) {
    Vertex *masspt = &(massptlist[idx]);
    masspt->init();
    idx++;

    Vertex *mesh_vertex = &(in_vrts[i]);
    if (mesh_vertex->typ == UNUSEDVERTEX) continue;
    //printf("  idx = %d fixed(%d)\n", i, mesh_vertex->is_fixed());
    if (mesh_vertex->is_fixed()) continue;

    double mcx = 0., mcy = 0.;
    int deg = 0; // debgree
    bool is_seg_vertex = false;
    TriEdge E = mesh_vertex->adj;
    do {
      if (E.is_segment()) {
        is_seg_vertex = true;
        break;
      }
      if (!E.tri->is_hulltri()) {
        deg++;
        mcx += E.dest()->crd[0];
        mcy += E.dest()->crd[1];
      }
      E = E.eprev_esym(); // CCW
    } while (E.tri != mesh_vertex->adj.tri);

    if (!is_seg_vertex) {
      //if (deg <= 2) continue;
      mcx /= deg;
      mcy /= deg;
      // Save this point.
      masspt->crd[0] = mcx;
      masspt->crd[1] = mcy;
      masspt->Pair = mesh_vertex; // Remember it.
      mcount++;
    }
  }

  if (tr_steiners != NULL) {
    for (int i = 0; i < tr_steiners->used_items; i++) {
      Vertex *masspt = &(massptlist[idx]);
      masspt->init();
      idx++;
    
      Vertex *mesh_vertex = (Vertex *) tr_steiners->get(i);
      if (mesh_vertex->is_deleted()) continue;
      //printf("  idx = %d fixed(%d)\n", i, mesh_vertex->is_fixed());
      if (mesh_vertex->is_fixed()) continue;

      double mcx = 0., mcy = 0.;
      int deg = 0; // debgree
      bool is_seg_vertex = false;
      TriEdge E = mesh_vertex->adj;
      do {
        if (E.is_segment()) {
          is_seg_vertex = true;
          break;
        }
        if (!E.tri->is_hulltri()) {
          deg++;
          mcx += E.dest()->crd[0];
          mcy += E.dest()->crd[1];
        }
        E = E.eprev_esym(); // CCW
      } while (E.tri != mesh_vertex->adj.tri);

      if (!is_seg_vertex) {
        //if (deg <= 2) continue;
        mcx /= deg;
        mcy /= deg;
        // Save this point.
        masspt->crd[0] = mcx;
        masspt->crd[1] = mcy;
        masspt->Pair = mesh_vertex; // Remember it.
        mcount++;
      }
    }
  }

  return mcount;
}

//==============================================================================

int Triangulation::smooth_vertices(int criterion)
{
  /*
  if (op_db_verbose > 1) {
    printf("  Calculating %d orthocenters\n", tr_tris->objects + ct_hullsize);
  }
  assert(OMT_domain != NULL);
  
  int i, idx;
  idx = io_firstindex;
  // Calculate circumcenters for hull triangles.
  for (i = 0; i < tr_tris->used_items; i++) {
    Triang* tri = (Triang *) tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted() || tri->is_hulltri()) continue;
    get_tri_orthocenter(tri);
    tri->idx = idx; // hulltri is also indexed.
    idx++;
  }
  // Calculate bisectors for hull triangles.
  for (i = 0; i < tr_tris->used_items; i++) {
    Triang* tri = (Triang *) tr_tris->get(i);
    // Ignore exterior triangles.
    if (tri->is_deleted()) continue;
    if (tri->is_hulltri()) { // A hull triangle.
      get_hulltri_orthocenter(tri);
    }
  }

  int ptcount = ct_in_vrts;
  if (tr_steiners) {
    ptcount += tr_steiners->objects;
  }
  Vertex *massptlist = new Vertex[ptcount];
  int mcount = 0; // count the number of moved vertices.

  if (op_db_verbose > 1) {
    printf("  Calculating %d power cells and mass centers\n", ptcount);
  }

  // Calculate power cells and mass centers.
  idx = 0;
  for (i = 0; i < ct_in_vrts; i++) {
    Vertex *masspt = &(massptlist[idx]);
    masspt->init();
    idx++;

    Vertex *mesh_vert = &(in_vrts[i]);
    if (mesh_vert->typ == UNUSEDVERTEX) continue;
    if (mesh_vert->is_fixed()) continue;
    //printf("  idx = %d\n", i);
    Vertex *ptlist = NULL;
    int ptnum = 0;

    if (get_powercell(mesh_vert, &ptlist, &ptnum)) {
      // Calculate its mass center
      get_mass_center(ptlist, ptnum, masspt->crd);
      masspt->Pair = mesh_vert; // Remember it.
      mcount++;
      delete [] ptlist;
    } // if (Tr->get_powercell
  }

  if (tr_steiners != NULL) {
    for (i = 0; i < tr_steiners->used_items; i++) {
      Vertex *masspt = &(massptlist[idx]);
      masspt->init();
      idx++;
    
      Vertex *mesh_vert = (Vertex *) tr_steiners->get(i);
      if (mesh_vert->is_deleted()) continue;
      if (mesh_vert->is_fixed()) continue;
      //printf("  idx = %d\n", i);
      Vertex *ptlist = NULL;
      int ptnum = 0;

      if (get_powercell(mesh_vert, &ptlist, &ptnum)) {
        get_mass_center(ptlist, ptnum, masspt->crd);
        masspt->Pair = mesh_vert; // Remember it.
        mcount++;
        delete [] ptlist;
      } // if (Tr->get_powercell
    }
  } 
  */

  int ptcount = ct_in_vrts;
  if (tr_steiners) {
    ptcount += tr_steiners->objects;
  }
  Vertex *massptlist = new Vertex[ptcount];
  int mcount = 0; // count the number of moved vertices.

  if (criterion == SMOOTH_LAPLACIAN) {
    printf("\nUsing Lapacian smoother\n");
    mcount = get_laplacian_centers(massptlist);
  } else if (criterion == SMOOTH_CVT) {
    printf("\nUsing CVT smoother\n");
    mcount = get_powercell_mass_centers(massptlist);
  } else {
    printf("Smooth option not available yet.\n");
    delete [] massptlist;
    return 0;
  }

  if (op_db_verbose > 1) {
    printf("  Relocating %d vertices.\n", mcount);
  }
  arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);

  for (int i = 0; i < ptcount; i++) {
    Vertex *mcpt = &(massptlist[i]);
    if (mcpt->Pair == NULL) continue; 

    Vertex *mesh_vert = mcpt->Pair;
    if (op_db_verbose > 2) {
      printf("    Relocating vertex %d.\n", mesh_vert->idx);
    }

    // Check if we can directly move it.
    bool bflag = true;
    int hullflag = 0; // Check if it is a hull vertex.
    TriEdge E = mesh_vert->adj;
    do {
      if (!E.tri->is_hulltri()) {
        if (Orient2d(E.dest(), E.apex(), mcpt) <= 0.) {
          bflag = false; break;
        } else {
          hullflag = 1;
        }
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vert->adj.tri);   

    if (bflag) {
      REAL x = mcpt->crd[0];
      REAL y = mcpt->crd[1];
      if (op_db_verbose > 2) {
        printf("    Directly move to its new position - hull(%d).\n", hullflag);
      }
      mesh_vert->crd[0] = x;
      mesh_vert->crd[1] = y;
      mesh_vert->crd[2] = x*x + y*y - mesh_vert->wei;

      // Do flips to recover Delaunayness.
      TriEdge N;
      E = mesh_vert->adj;
      do {
        E.set_edge_infect(); // A star edge
        * (TriEdge *) fqueue->alloc() = E;
        N = E.enext();       // A link edge
        N.set_edge_infect();
        * (TriEdge *) fqueue->alloc() = N;
        E = E.eprev_esym(); // ccw
      } while (E.tri != mesh_vert->adj.tri);

      lawson_flip(NULL, hullflag, fqueue);
      continue;
    }

    if (op_db_verbose > 2) {
      printf("    Delete-and-Insert to its new position.\n");
    }
    // Bakup
    mcpt->idx = mesh_vert->idx;
    mcpt->tag = mesh_vert->tag;
    mcpt->wei = mesh_vert->wei;
    mcpt->val = mesh_vert->val;
    mcpt->typ = mesh_vert->typ;

    // Remove this vertex from the triangulation.
    if (!remove_point(mesh_vert, fqueue)) {
      //assert(0); // This should always succeed.
      if (op_db_verbose > 2) {
        printf("    Failed to remove vertex %d\n", mesh_vert->idx);
      }
      // Recover Delaunay at this vertex.
      // push the link edges of mesh_vert.
      TriEdge N;
      E = mesh_vert->adj;
      do {
        E.set_edge_infect(); // A star edge
        * (TriEdge *) fqueue->alloc() = E;
        N = E.enext();       // A link edge
        N.set_edge_infect();
        * (TriEdge *) fqueue->alloc() = N;
        E = E.eprev_esym(); // ccw
      } while (E.tri != mesh_vert->adj.tri);

      lawson_flip(NULL, hullflag, fqueue);
      continue;
    }

    lawson_flip(NULL, hullflag, fqueue); // hullflag = 0

    // Insert a new vertex at (x,y).
    Vertex *newpt;
    if (mcpt->typ == STEINERVERTEX) {
      newpt = (Vertex *) tr_steiners->alloc();
      newpt->init();
    } else {
      newpt = mesh_vert;
      assert(newpt->typ == UNUSEDVERTEX);
      ct_unused_vrts++;
    }
    REAL x = mcpt->crd[0];
    REAL y = mcpt->crd[1];
    newpt->crd[0] = x;
    newpt->crd[1] = y;
    newpt->crd[2] = x*x + y*y - mcpt->wei;
    newpt->idx = mcpt->idx;
    newpt->wei = mcpt->wei; // the weight
    newpt->val = mcpt->val; // mesh size
    newpt->typ = mcpt->typ;
    //newpt->Pair = mcpt->Pair;

    TriEdge tt[4];
    E.tri = NULL;
    int loc = locate_point(newpt, E, 1); // Stop at segment.
    tt[0] = E;
    if (loc == LOC_IN_OUTSIDE) {
      int fflag = FLIP_13;
      flip(tt, &newpt, fflag, fqueue);
    } else if (loc == LOC_IN_TRI) {
      int fflag = FLIP_13;
      flip(tt, &newpt, fflag, fqueue);
    } else if (loc == LOC_ON_EDGE) {
      int fflag = FLIP_24;
      flip(tt, &newpt, fflag, fqueue);
    } else {
      assert(0); // Not possible.
    }
    lawson_flip(newpt, hullflag, fqueue);
  } // i

  //if (check_mesh(0,0) > 0) {
  //  printf("!!! Mesh invalid after relocating vertices.\n");
  //  assert(0);
  //}

  delete fqueue;
  delete [] massptlist;
  return 1;
}
