#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "detri2.h"

using  namespace detri2;

//==============================================================================

// On return, "E" is the TriEdge containing the point "pt".
//
// The returned value is one of the following:
//  1, in triangle (may be outside, if this is a hull triangle).
//  2, on edge, return by (E.org(), E.dest()).
//  3, on vertex, returned by E.org().
//  4, try to cross a segment (maybe interior) (used by delaunay refinement).
//
// If "rndflag" is set, adjaust the exact location to approximate location,
//  i.e., on edge, and on vertex.
//
// See the function insert_point() for an example.
//
//int Triangulation::locate_point(Vertex *pt, TriEdge &E, int rndflag, int encflag)
int Triangulation::locate_point(Vertex *pt, TriEdge &E, int encflag)
{
  if (op_db_verbose > 1) {
    printf("  Locating point [%d]\n", pt->idx);
  }

  if ((E.tri == NULL) || E.tri->is_deleted()) {
    // Try to search from the recent visited triangle.
    if ((tr_recnttri != NULL) && !tr_recnttri->is_deleted()) {
      E.tri = tr_recnttri; E.ver = 0;
    } else {
      E = tr_infvrt->adj;
    }
  }

  if (E.tri->is_hulltri()) {
    // Get a non-hull triangle.
    for (E.ver = 0; E.ver < 3; E.ver++) {
      if (E.apex() == tr_infvrt) break; 
    }
    assert(E.ver < 3);
    E = E.esym();
  }

  // Select an edge such that pt lies to CCW of it.
  for (E.ver = 0; E.ver < 3; E.ver++) {
    if (Orient2d(E.org(), E.dest(), pt) > 0) break;
  }

  // Let E = [a,b,c] and p lies to the CCW of [a->b].
  do {

    REAL ori1 = Orient2d(E.dest(), E.apex(), pt);
    REAL ori2 = Orient2d(E.apex(),  E.org(), pt);

    if (ori1 > 0) {
      if (ori2 > 0) {
        break; // Found.
      } else if (ori2 < 0) {
        E.ver = _eprev_tbl[E.ver]; 
      } else { // ori2 == 0
        E.ver = _eprev_tbl[E.ver]; 
        return LOC_ON_EDGE; // ONEDGE  p lies on edge [c,a]
      }
    } else if (ori1 < 0) {
      if (ori2 > 0) {
        E.ver = _enext_tbl[E.ver]; 
      } else if (ori2 < 0) {
        // Randomly choose one.
        if (rand() % 2) { // flipping a coin.
          E.ver = _enext_tbl[E.ver];
        } else {
          E.ver = _eprev_tbl[E.ver];
        }
      } else { // ori2 == 0
        E.ver = _enext_tbl[E.ver]; 
      }
    } else { // ori1 == 0
      if (ori2 > 0) {
        E.ver = _enext_tbl[E.ver]; // p lies on edge [b,c].
        return LOC_ON_EDGE; // ONEDGE
      } else if (ori2 < 0) {
        E.ver = _eprev_tbl[E.ver]; 
      } else { // ori2 == 0
        E.ver = _eprev_tbl[E.ver]; // p is coincident with apex.
        return LOC_ON_VERT; // ONVERTEX Org(E)
      }
    }

    E = E.esym(); // Go to the adjacent triangle.

    if (E.is_segment()) {
      if (encflag) {
        return LOC_ENC_SEG; // Do not cross a boundary edge.
      }
    }
  } while (!E.tri->is_hulltri());

  tr_recnttri = E.tri; // Remember this triangle for next point location.

  if (E.tri->is_hulltri()) {
    // pt lies at the exiterior of this hull edge [v1, v2].
    if (tr_nonconvex) {
      // This triangulation is non-convex, the search might get stuck
      //   at an interior hole or a non-convex segment.
      printf("!! search non-convex triangulation (globally)\n");
      int j;
      for (j = 0; j < tr_tris->used_items; j++) {
        Triang *t = (Triang *) tr_tris->get(j);
        if (t->is_deleted()) continue;
        if (t->is_hulltri()) continue;
        Vertex *pa = t->vrt[0];
        Vertex *pb = t->vrt[1];
        Vertex *pc = t->vrt[2];
        double A  = get_tri_area(pa, pb, pc);
        double Wa = get_tri_area(pt, pb, pc) / A;
        double Wb = get_tri_area(pa, pt, pc) / A;
        double Wc = get_tri_area(pa, pb, pt) / A;
        if (fabs((Wa + Wb + Wc) - 1.0) < 1.e-3) {
          E.tri = t; //break;
          //return locate_point(pt, E, rndflag, encflag);
          return locate_point(pt, E, encflag);
        }
      }
    }
    return LOC_IN_OUTSIDE;
  } // if (E.tri->is_hulltri())

  // The point lies exactly inside the triangle (NOT in outside).
  //assert(!E.tri->is_hulltri());

  /*
  if (rndflag > 0) {
    // Approximating the location of the point w.r.t. this triangle.
    Vertex *v1 = E.org();
    Vertex *v2 = E.dest();
    Vertex *v3 = E.apex();
    double a1 = get_tri_area(pt, v1, v2);
    double a2 = get_tri_area(pt, v2, v3);
    double a3 = get_tri_area(pt, v3, v1);
    double triarea = a1 + a2 + a3;
    // Round the areas
    if (fabs(a1 / triarea - 1.0) < 1e-5) a1 = 0.0;
    if (fabs(a2 / triarea - 1.0) < 1e-5) a2 = 0.0;
    if (fabs(a3 / triarea - 1.0) < 1e-5) a3 = 0.0;
    if (a1 == 0) {
      if (a2 == 0) {
        //if (a3 > 0) {
          assert(a3 > 0);
          E = E.enext(); // E.enextself(); // On vertex E.dest().
          return LOC_ON_VERT; // return 3;
        //}
      } else if (a3 == 0) {
        return LOC_ON_VERT; // return 3; // On vertex E.org().
      } else {
        return LOC_ON_EDGE; // return 2; // On edge.
      }
    } else if (a2 == 0) {
      if (a3 == 0) {
        E = E.eprev(); // E.enext2self(); // On vertex E.apex().
        return LOC_ON_VERT; // return 3;
      } else {
        E = E.enext(); //E.enextself(); // On edge.
        return LOC_ON_EDGE; // return 2;
      }
    } else if (a3 == 0) {
      E = E.eprev(); // E.enext2self(); // On edge.
      return LOC_ON_EDGE; // return 2;
    }
  } // if (rndflag > 0)
  */

  return LOC_IN_TRI; // P lies inside E = [a,b,c]
}

//==============================================================================

int Triangulation::sort_vertices(Vertex* vrtarray, int arysize, Vertex**& permutarray)
{
  permutarray = new Vertex*[arysize];
  int randindex, i;

  if (op_db_verbose) {
    printf("Sorting vertices...\n");
  }

  if (so_nosort | so_norandom) { // -SN or -SR
    for (i = 0; i < arysize; i++) {
      permutarray[i] = &vrtarray[i];
    }
  } else {
    // Randomly permute the vertices.
    srand(arysize);
    for (i = 0; i < arysize; i++) {
      randindex = rand() % (i + 1); 
      permutarray[i] = permutarray[randindex];
      permutarray[randindex] = &vrtarray[i];
    }
  }

  if (!so_nosort && !so_nobrio) { // no -SN or -SB
    hilbert_init(2);
    brio_multiscale_sort2(permutarray, arysize, so_brio_threshold,
                          so_brio_ratio, so_hilbert_order, so_hilbert_limit,
                          io_xmin, io_xmax, io_ymin, io_ymax);
  }

  return 1;
}

//==============================================================================

int Triangulation::first_tri(Vertex **ptlist, int ptnum)
{
  int i, j, it = 0;
  REAL ori = 0.0;
  Vertex *swappt;

  if (so_nosort) {
    // Assume pt[0] and pt[1] are distinct.
    // Search the third non-collinear vertex.
    for (i = 2; i < ptnum; i++) {
      ori = Orient2d(ptlist[0], ptlist[1], ptlist[i]);
      if (ori != 0) break;
    }
    if (i == ptnum) return 0; // Failed
    // If i > 2, swap it to 2.
    if (i > 2) {
      swappt = ptlist[2];
      ptlist[2] = ptlist[i];
      ptlist[i] = swappt;
    }
  } else {
  // Randomly select 3 points.
    do {
      // Randomly select three vertices from the iuput list.
      for (i = 0; i < 3; i++) {
        // Swap ith and jth element.
        j = rand() % (ptnum - i);
        swappt = ptlist[i];
        ptlist[i] = ptlist[j];
        ptlist[j] = swappt;
      }
      ori = Orient2d(ptlist[0], ptlist[1], ptlist[2]);
      if (ori != 0) break;
      it++;
    } while (it < 10);

    if (it >= 10) return 0; // Failed.
  }

  if (ori < 0) {
    // Swap the first two vertices.
    swappt = ptlist[0];
    ptlist[0] = ptlist[1];
    ptlist[1] = swappt;
  }

  if (op_db_verbose > 1) {
    printf("  First triangle [%d,%d,%d]\n", ptlist[0]->idx, ptlist[1]->idx,
           ptlist[2]->idx);
  }

  if (tr_tris == NULL) {
    // Estimate the final mesh size.
    int est_size = 0;
    if (tr_segs == NULL) {
      est_size = ct_in_vrts * 4; // Delaunay triangulation case. 
    } else {
      est_size = 8192; // temp
    }
    int log2objperblk = 0;
    while (est_size >>= 1) log2objperblk++;
    tr_tris = new arraypool(sizeof(Triang), log2objperblk);
  }

  first_triangle(ptlist[0], ptlist[1], ptlist[2]);

  return 1;
}

//==============================================================================

int Triangulation::incremental_delaunay()
{
  if (op_db_verbose) {
    printf("Incremental Delaunay construction...\n");
  }

  Vertex** vrtarray = NULL;
  int arysize = ct_in_vrts;

  sort_vertices(in_vrts, ct_in_vrts, vrtarray);

  if (!first_tri(vrtarray, ct_in_vrts)) {
    return 0;
  }

  arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
  TriEdge E = tr_infvrt->adj, tt[4];
  int loc;

  for (int i = 3; i < arysize; i++) {
    if (op_db_verbose > 1) {
      printf("  Inserting vertex %d: %d\n", i+1, vrtarray[i]->idx);
    }
    //loc = locate_point(vrtarray[i], E, 0, 0); // encflag = 0
    loc = locate_point(vrtarray[i], E, 0); // encflag = 0
    if (loc != LOC_ON_VERT) { // ON_VERTEX
      REAL ori = 1.0;
      if (loc == LOC_IN_TRI) { //if ((loc == 1) && !E.tri->is_hulltri()) {
        // Make sure this vertex is regular w.r.t. E.
        // An interior edge.
        ori = Orient3d(E.org(), E.dest(), E.apex(), vrtarray[i])
            * op_dt_nearest;
      }
      if (ori > 0) {
        // Insert vertex
        tt[0] = E;
        if (loc == LOC_IN_OUTSIDE) { // IN outise
          int fflag = FLIP_13;
          flip(tt, &(vrtarray[i]), fflag, fqueue);
        } else if (loc == LOC_IN_TRI) { // IN_TRAING
          int fflag = FLIP_13;
          flip(tt, &(vrtarray[i]), fflag, fqueue);
        } else if (loc == LOC_ON_EDGE) { // ON_EDGE
          int fflag = FLIP_24;
          //tt[1] = tt[0].esym();
          flip(tt, &(vrtarray[i]), fflag, fqueue);
        } else {
          printf(" loc = %d\n", loc);
          assert(0); // not possible for encflag = 0
        }
        lawson_flip(vrtarray[i], 1, fqueue); // hullflag = 1
        E = vrtarray[i]->adj; // For next point location.
      } else {
        if (op_db_verbose > 1) {
          printf("  Skip a redundant vertex %d\n", vrtarray[i]->idx);
        }
      }
    } else {
      if (op_db_verbose) {
        printf("Warning:  Vertex %d is coincident with %d\n",
               vrtarray[i]->idx, E.org()->idx);
      }
    }
  }

  delete fqueue;
  delete [] vrtarray;
  return 1;
}
