#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "detri2.h"

using  namespace detri2;

//==============================================================================

int Triangulation::line_line_intersection(double X0, double Y0, double X1, double Y1,
                             double X2, double Y2, double X3, double Y3,
                             double *t1, double *t2)
{
  double Ux = X1 - X0;
  double Uy = Y1 - Y0;
  double Vx = X3 - X2;
  double Vy = Y3 - Y2;
  double Wx = X2 - X0;
  double Wy = Y2 - Y0;
  double det = Ux*Vy - Uy*Vx;

  double absolut = fabs(Ux*Vy) + fabs(Uy*Vx);
  if (fabs(det) / absolut < 1e-6) {
    printf("Warning: Two lines are nearly parallel.\n");
    *t1 = *t2 = 0.0;
    return 0;
  }

  *t1 = (Wx*Vy - Wy*Vx) / det;
  *t2 = (Ux*Wy - Uy*Wx) / det;
  return 1;
}

//==============================================================================

// Get the circumcenter (or orthocenter) of 3 (weighted) vertices, U, V, W.
int Triangulation::get_orthocenter(REAL Ux, REAL Uy, REAL U_weight,
                                   REAL Vx, REAL Vy, REAL V_weight,
                                   REAL Wx, REAL Wy, REAL W_weight,
                                   REAL* Cx, REAL* Cy, REAL* r2)
{
  /*
  if (op_dt_nearest < 0) { // -u option
    // This is the furtherest weighted Delaunay.
    // The weights must be transformed.
    REAL h_u = -(Ux * Ux + Uy * Uy - U_weight); // The transformed height
    REAL h_v = -(Vx * Vx + Vy * Vy - V_weight);
    REAL h_w = -(Wx * Wx + Wy * Wy - W_weight);
    U_weight = (Ux * Ux + Uy * Uy) - h_u; // The transfomred weight
    V_weight = (Vx * Vx + Vy * Vy) - h_v;
    W_weight = (Wx * Wx + Wy * Wy) - h_w;
  }
  */

  REAL Px = Vx - Ux;
  REAL Py = Vy - Uy;
  REAL Qx = Wx - Ux;
  REAL Qy = Wy - Uy;
  REAL Px2_plus_Py2 = Px * Px + Py * Py + U_weight - V_weight; // + w_0 - w_p
  REAL Qx2_plus_Qy2 = Qx * Qx + Qy * Qy + U_weight - W_weight; // + w_0 - w_q

  REAL det = Px * Qy - Qx * Py;
  REAL dx, dy;
  if ((fabs(det) / (fabs(Px2_plus_Py2) + fabs(Qx2_plus_Qy2))) < 1e-8) {
    // The triangle is nearly degenerated.
    det = 0;
    if (op_db_verbose) {
      printf("!! Warning: triangle is (nearly) degenerated.\n");
      printf("  The mass center is used.\n");
    }
    dx = (Ux + Vx + Wx) / 3.0;
    dy = (Uy + Vy + Wy) / 3.0;
    dx -= Ux;
    dy -= Uy;
  } else {
    REAL denominator = 0.5 / det; //  0.5 / (Px * Qy - Qx * Py);
    dx = (Qy * Px2_plus_Py2 - Py * Qx2_plus_Qy2) * denominator;
    dy = (Px * Qx2_plus_Qy2 - Qx * Px2_plus_Py2) * denominator;
  }

  // The power (of radius).
  REAL rr2 = dx * dx + dy * dy - U_weight;

  if (Cx != NULL) {
    *Cx = Ux + dx;
    *Cy = Uy + dy;
    *r2 = rr2;
    if (op_db_verbose > 2) {
      printf("  (Cx, Cy): %g %g r(%g) r2(%g)\n", *Cx, *Cy, fabs(rr2), rr2);
    }
  } else {
    // Only print the result (for debug).
    REAL cx, cy, r, h;
    cx = Ux + dx;
    cy = Uy + dy;
    r = sqrt(fabs(rr2));
    h = cx*cx + cy*cy - rr2;
    printf("  (Cx, Cy): %g %g r(%g) r2(%g) h(%g)\n", cx, cy, r, rr2, h);
  }

  return det != 0.0;
}

int Triangulation::get_bissector(REAL Ux, REAL Uy, REAL U_weight,
                                 REAL Vx, REAL Vy, REAL V_weight,
                                 REAL* Cx, REAL* Cy, REAL* r2)   // Cx, Cy, radius^2
{
  /*
  if (op_dt_nearest < 0) { // -u option
    // This is the furtherest weighted Delaunay.
    // The weights must be transformed.
    REAL h_u = -(Ux * Ux + Uy * Uy - U_weight); // The transformed height
    REAL h_v = -(Vx * Vx + Vy * Vy - V_weight);
    U_weight = (Ux * Ux + Uy * Uy) - h_u; // The transfomred weight
    V_weight = (Vx * Vx + Vy * Vy) - h_v;
  }
  */

  REAL Qx = Vx - Ux;
  REAL Qy = Vy - Uy;

  REAL Qx2_plus_Qy2 = Qx*Qx + Qy*Qy;
  REAL t = 0.5 * (1 + (U_weight - V_weight) / Qx2_plus_Qy2);

  // (cx, cy) is the bisect point (midpoint) on the line [e1->e2].
  REAL cx = Ux + t * Qx;
  REAL cy = Uy + t * Qy;
  REAL rr2 = t*t * (Qx*Qx + Qy*Qy) - U_weight;

  if (Cx != NULL) {
    *Cx = cx;
    *Cy = cy;
    *r2 = rr2;
    if (op_db_verbose > 2) {
      printf("  (Cx, Cy): %g %g r(%g) r2(%g)\n", *Cx, *Cy, fabs(rr2), rr2);
    }
  } else {
    // Only print the result.
    REAL r, h;
    r = sqrt(fabs(rr2));
    h = cx*cx + cy*cy - rr2;
    printf("  (Cx, Cy): %g %g r(%g) r2(%g) h(%g)\n", cx, cy, r, rr2, h);
  }

  return 1;
}

//==============================================================================
// see doc/voronoi-edge-boundary-cut-2018-11-1.pdf

int Triangulation::get_boundary_cut_dualedge(Vertex& In_pt, Vertex& Out_pt, TriEdge& S)
{
  if (op_db_verbose > 3) {
    printf("  S.org(%d) -> S.dest(%d)\n", S.org()->idx, S.dest()->idx);
    printf("  In (%g,%g) - Out (%g,%g)\n", In_pt.crd[0], In_pt.crd[1], Out_pt.crd[0], Out_pt.crd[1]);
  }

  REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
  assert(ss < 0); // CW orientation
  REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
  REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
  if (op_db_verbose > 3) {
    printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
    printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
  }

  if (s1 < 0) {
    do {
      // Must go left of S.org();
      TriEdge searchEdge = S.esym();
      //printf("DBG: adjust S to the left of S.org().\n");
      TriEdge workedge = searchEdge.enext();
      while (true) {
        if (op_db_verbose > 3) {
          printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                 workedge.dest()->idx, workedge.apex()->idx);
        }
        assert(workedge.org() == S.org());
        searchEdge = workedge.esym_enext(); // CW
        if (searchEdge.is_segment()) break;
        if ((searchEdge.esym()).tri->is_hulltri()) break;
        workedge = searchEdge;
      }
      S = searchEdge.esym(); // Update S.
      s1 = Orient2d(&In_pt, &Out_pt, S.org());
    } while (s1 < 0);
  } else if (s2 > 0) {
    // Muts go right of S.dest();
    //printf("DBG: adjust S to the right of S.dest().\n");
    do {
      TriEdge searchEdge = S.esym();
      TriEdge workedge = searchEdge.eprev();
      while (true) {
        if (op_db_verbose > 3) {
          printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                 workedge.dest()->idx, workedge.apex()->idx);
        }
        assert(workedge.dest() == S.dest());
        searchEdge = workedge;
        if (searchEdge.is_segment()) break;
        if ((searchEdge.esym()).tri->is_hulltri()) break;
        workedge = searchEdge.esym_eprev();
      }
      S = searchEdge.esym(); // Update S
      s2 = Orient2d(&In_pt, &Out_pt, S.dest());
    } while (s2 > 0);
  }

  if (op_db_verbose > 3) {
    printf("  Found background boundary S: [%d,%d,%d]\n\n",
           S.org()->idx, S.dest()->idx, S.apex()->idx);
  }

  return 1;
}

//==============================================================================

int Triangulation::get_hulltri_orthocenter(Triang* hulltri)
{
  TriEdge E; E.tri = hulltri;
  for (E.ver = 0; E.ver < 3; E.ver++) {
    if (E.apex() == tr_infvrt) break;
  }
  assert(E.ver < 3);

  REAL Ux, Uy, U_weight;
  REAL Vx, Vy, V_weight;

  Vertex *v1 = E.org();
  Vertex *v2 = E.dest();

  Ux = v1->crd[0];
  Uy = v1->crd[1];
  Vx = v2->crd[0];
  Vy = v2->crd[1];

  // Translate heights to weights.
  U_weight = Ux*Ux + Uy*Uy - v1->crd[2];
  V_weight = Vx*Vx + Vy*Vy - v2->crd[2];

  if (op_db_verbose > 2) {
    printf("A Hull edge:\n");
    printf(" [%d]: %g,%g,%g, r(%g)\n", v1->idx, Ux, Uy, U_weight, sqrt(fabs(U_weight)));
    printf(" [%d]: %g,%g,%g, r(%g)\n", v2->idx, Vx, Vy, V_weight, sqrt(fabs(V_weight)));
  }

  get_bissector(Ux, Uy, U_weight,
                Vx, Vy, V_weight,
                &E.tri->cct[0], &E.tri->cct[1], &E.tri->cct[2]);

  if (OMT_domain) {
    // The bisector lies exactly in the edge [v1, v2].
    // We move it along the edge normal towards exterior of the triangulaiton.
    TriEdge N = E.esym();
    // The dual vertex of N.tri was already calcualted.
    assert(N.tri->on.tri != NULL); 
    if (op_db_verbose > 2) {
      printf("  Bisector on the hull edge:  [%g,%g]\n", E.tri->cct[0], E.tri->cct[1]);
      printf("  Ccenter of hull triangle N: [%g,%g]\n", N.tri->cct[0], N.tri->cct[1]);
      //printf("p:show_vector(%g,%g,0,%g,%g,0)\n",
      //       N.tri->cct[0], N.tri->cct[1], E.tri->cct[0], E.tri->cct[1]);
    }

    // Get the edge normal towards exterior (rotated 90 deg).
    double outerN[2];
    outerN[0] =  (E.org()->crd[1] - E.dest()->crd[1]);
    outerN[1] = -(E.org()->crd[0] - E.dest()->crd[0]);
    double len = sqrt(outerN[0]*outerN[0] + outerN[1]*outerN[1]);
    outerN[0] /= len;
    outerN[1] /= len;

    if (op_dt_nearest < 0) { // The farthest-point Voronoi diagram
      // change direction of the outer normal.
      outerN[0] = -outerN[0];
      outerN[1] = -outerN[1];
    }

    len = get_distance(E.org(), E.dest());

    if (!N.tri->is_dual_in_exterior() && !N.tri->is_dual_on_bdry()) {
      // This dual vertex lies inside the domain.
      if (!N.is_segment()) {
        // Move the dual vertex to the boundary of the OMT_domain.
        // For this we need a far point outside the OMT_domain.
        double scale = len;
        /*
        Vertex Mass_pt; // pt;
        REAL Wx = N.apex()->crd[0];
        REAL Wy = N.apex()->crd[1];
        Mass_pt.crd[0] = (Ux + Vx + Wx) / 3.;
        Mass_pt.crd[1] = (Uy + Vy + Wy) / 3.;
        TriEdge S = N.tri->on; // Must be an interior triangle of OMT_domain.
        int loc = OMT_domain->locate_point(&Mass_pt, S, 0, 0);
        */
        Vertex In_pt; // The bisector of this hull edge.
        In_pt.init();
        In_pt.crd[0] = E.tri->cct[0];
        In_pt.crd[1] = E.tri->cct[1];
        TriEdge S = N.tri->on; // Must be an interior triangle of OMT_domain.
        if (op_db_verbose > 2) {
          printf("  N.tri->on (cct) background tri: [%d,%d,%d]\n", S.org()->idx, S.dest()->idx, S.apex()->idx);
        }
        //int loc = OMT_domain->locate_point(&In_pt, S, 0, 0);
        int loc = OMT_domain->locate_point(&In_pt, S, 0);
        if (op_db_verbose > 2) {
          printf("  In_pt on background tri: [%d,%d,%d]\n", S.org()->idx, S.dest()->idx, S.apex()->idx);
        }
        assert(loc != LOC_IN_OUTSIDE);
        Vertex Out_pt;
        Out_pt.init();
        while (1) { // we must find a point.
          scale = scale * 2.0;
          Out_pt.crd[0] = E.tri->cct[0] + scale * outerN[0];
          Out_pt.crd[1] = E.tri->cct[1] + scale * outerN[1];
          // Search the Out_pt, stop at the first segment.
          //loc = OMT_domain->locate_point(&Out_pt, S, 0, 1);
          loc = OMT_domain->locate_point(&Out_pt, S, 1);
          if (loc == LOC_IN_OUTSIDE) {
            assert(S.apex() == OMT_domain->tr_infvrt);
            break;
          } else if (loc == LOC_IN_TRI) {
            if (S.tri->is_exterior()) break;
          } else if (loc == LOC_ENC_SEG) {
            assert(S.is_segment());
            break; // hit a segment.
          }
        } // while (1)
        if (op_db_verbose > 3) {
          printf("  Background boundary: [%d,%d,%d]\n", S.org()->idx, S.dest()->idx, S.apex()->idx);
        }
        get_boundary_cut_dualedge(In_pt, Out_pt, S);
        /*
        // Check if S [org, dest] cuts the line [In_pt, Out_pt].
        //===================== Subroutine start =================================
        // see doc/voronoi-edge-boundary-cut-2018-11-1.pdf
        REAL ss = Orient2d(S.org(), S.dest(), &In_pt);
        assert(ss < 0); // Bisect_pt must be CW oriented (inside of OMT_domain).
        REAL s1 = Orient2d(&In_pt, &Out_pt, S.org());
        REAL s2 = Orient2d(&In_pt, &Out_pt, S.dest());
        if (op_db_verbose > 2) {
          printf("  [In, out, S.org(%d)]  s1 = %g\n", S.org()->idx, s1);
          printf("  [In, out, S.dest(%d)] s2 = %g\n", S.dest()->idx, s2);
        }
        if (s1 < 0) {
          do {
            // Must go left of S.org();
            TriEdge searchEdge = S.esym();
            //printf("DBG: adjust S to the left of S.org().\n");
            TriEdge workedge = searchEdge.enext();
            while (true) {
              if (op_db_verbose > 3) {
                printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                       workedge.dest()->idx, workedge.apex()->idx);
              }
              assert(workedge.org() == S.org());
              searchEdge = workedge.esym_enext(); // CW
              if (searchEdge.is_segment()) break;
              if ((searchEdge.esym()).tri->is_hulltri()) break;
              workedge = searchEdge;
            }
            S = searchEdge.esym(); // Update S.
            s1 = Orient2d(&In_pt, &Out_pt, S.org());
          } while (s1 < 0);
        } else if (s2 > 0) {
          // Muts go right of S.dest();
          //printf("DBG: adjust S to the right of S.dest().\n");
          do {
            TriEdge searchEdge = S.esym();
            TriEdge workedge = searchEdge.eprev();
            while (true) {
              if (op_db_verbose > 3) {
                printf("  workE: (%d,%d,%d)\n", workedge.org()->idx,
                       workedge.dest()->idx, workedge.apex()->idx);
              }
              assert(workedge.dest() == S.dest());
              searchEdge = workedge;
              if (searchEdge.is_segment()) break;
              if ((searchEdge.esym()).tri->is_hulltri()) break;
              workedge = searchEdge.esym_eprev();
            }
            S = searchEdge.esym(); // Update S
            s2 = Orient2d(&In_pt, &Out_pt, S.dest());
          } while (s2 > 0);
        }
        if (op_db_verbose > 3) {
          printf("  Found background boundary S: [%d,%d,%d]\n\n",
                 S.org()->idx, S.dest()->idx, S.apex()->idx);
        }
        //===================== Subroutine end =================================
        */
        E.tri->on = S; // Remember this triangle in OMT_domain.

        // Calculate the cut point.
        Vertex *e1 = E.tri->on.org();
        Vertex *e2 = E.tri->on.dest();
        double X0 = N.tri->cct[0];
        double Y0 = N.tri->cct[1];
        double X1 = Out_pt.crd[0];
        double Y1 = Out_pt.crd[1];
        double X2 = e1->crd[0];
        double Y2 = e1->crd[1];
        double X3 = e2->crd[0];
        double Y3 = e2->crd[1];
        double t1, t2;
        if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
          E.tri->cct[0] = X0 + t1 * (X1 - X0);
          E.tri->cct[1] = Y0 + t1 * (Y1 - Y0);
          // Calulcate rr2.
          REAL dx = E.tri->cct[0] - Ux;
          REAL dy = E.tri->cct[1] - Uy;
          E.tri->cct[2] = dx * dx  + dy * dy - U_weight;
        } else {
          printf("!! Warning: Failed at calculating line-line intersection.\n");
        }
      } // if (!N.is_segment())
      // It is either on a segment or a hull edge of OMT_domain.
      E.tri->set_dual_on_bdry();
    } else {
      // The dual vertex of its neighbor lies outside.
      // Move this dual vertex further than the neighbor dual vertex.
      // This is only for visualization useful.
      E.tri->cct[0] = N.tri->cct[0] + len * outerN[0];
      E.tri->cct[1] = N.tri->cct[1] + len * outerN[1];
      E.tri->on = N.tri->on;
      E.tri->set_dual_in_exterior();
    }
    if (op_db_verbose > 10) {
      printf("p:show_vector(%g,%g,0,%g,%g,0)\n", N.tri->cct[0], N.tri->cct[1],
             E.tri->cct[0], E.tri->cct[1]);
    }
  } // if (OMT_domain)

  return 1;
}

// "tri" must be not a hulltri and exterior
int Triangulation::get_tri_orthocenter(Triang* tri)
{
  assert(!tri->is_hulltri());

  REAL Ux, Uy, U_weight;
  REAL Vx, Vy, V_weight;
  REAL Wx, Wy, W_weight;

  Vertex *v1 = tri->vrt[0];
  Vertex *v2 = tri->vrt[1];
  Vertex *v3 = tri->vrt[2];

  Ux = v1->crd[0];
  Uy = v1->crd[1];
  Vx = v2->crd[0];
  Vy = v2->crd[1];
  Wx = v3->crd[0];
  Wy = v3->crd[1];

  // Translate heights to weights.
  U_weight = Ux*Ux + Uy*Uy - v1->crd[2];
  V_weight = Vx*Vx + Vy*Vy - v2->crd[2];
  W_weight = Wx*Wx + Wy*Wy - v3->crd[2];

  if (op_db_verbose > 2) {
    printf(" [%d]: %g,%g,%g, r(%g)\n", v1->idx, Ux, Uy, U_weight, sqrt(fabs(U_weight)));
    printf(" [%d]: %g,%g,%g, r(%g)\n", v2->idx, Vx, Vy, V_weight, sqrt(fabs(V_weight)));
    printf(" [%d]: %g,%g,%g, r(%g)\n", v3->idx, Wx, Wy, W_weight, sqrt(fabs(W_weight)));
  }

  get_orthocenter(Ux, Uy, U_weight,
                  Vx, Vy, V_weight,
                  Wx, Wy, W_weight,
                  &tri->cct[0], &tri->cct[1], &tri->cct[2]);

  if (OMT_domain != NULL) {
    // Determine if this vertex is inside or outside of the OMT domain (or the subdomain
    //   which contains this triangle -- check segments).
    Vertex Mass_pt, pt;
    Mass_pt.crd[0] = (Ux + Vx + Wx) / 3.;
    Mass_pt.crd[1] = (Uy + Vy + Wy) / 3.;
    pt.crd[0] = tri->cct[0];
    pt.crd[1] = tri->cct[1];
    // First search the mass center of "tri" in OMT_domain.
    // It must be inside the domain.
    //int loc = OMT_domain->locate_point(&Mass_pt, tri->on, 0, 0); // encflag = 0
    int loc = OMT_domain->locate_point(&Mass_pt, tri->on, 0); // encflag = 0
    assert(!tri->on.tri->is_hulltri()); 
    // Locate the orthocenter of "tri" from this triangle (containing its mass center). 
    //loc = OMT_domain->locate_point(&pt, tri->on, 0, 1); // encflag = 1 (stop at first segment).
    loc = OMT_domain->locate_point(&pt, tri->on, 1); // encflag = 1 (stop at first segment).
    // Set the interior /exterior flag of the dual vertex.
    // Default, the dual (orothocenter) is in the interior.
    if (loc == LOC_IN_OUTSIDE) {
      tri->set_dual_in_exterior();
    } else if (loc == LOC_IN_TRI) {
      if (tri->on.tri->is_exterior()) {
        tri->set_dual_in_exterior();
      }
    } else if (loc == LOC_ON_EDGE) {
      if (tri->on.is_segment()) {
        tri->set_dual_on_bdry();
      }
    } else if (loc == LOC_ON_VERT) {
      // On a vertex of the background mesh, to be done.
      TriEdge E = tri->on;
      do {
        if (tri->on.tri->is_hulltri() || tri->on.tri->is_exterior()) {
          tri->set_dual_in_exterior(); break;
        } else if (tri->on.is_segment()) {
          tri->set_dual_on_bdry(); break;
        }
        tri->on = tri->on.eprev_esym(); // CCW rotate
        assert(tri->on.org() == E.org());
      } while (E.tri != tri->on.tri);
    } else if (loc == LOC_ENC_SEG) {
      // This dual vertex is behind a segment, which means it lies outside of the
      //   subdomain contains the triangle. We treat it as lying in outside.
      //   This way, a cut point on the segment will be calculated for Voronoi cells.
      assert(tri->on.is_segment());
      tri->set_dual_in_exterior();
    } 
  } // if (OMT_domain != NULL)

  return 1;
}

//==============================================================================
// The power cell of a mesh vertex is given by the convex hull of a set of
//   corners, which are Voronoi vertices and cutting vertices (of Voronoi edges
//   with domain boundary edges). The set of corners are returned by
//   an array ``pptlist" ordered in counterclockwise direction.
//   It assumes that a ``OMT_domain" must be provided.

int Triangulation::get_powercell(Vertex *mesh_vertex, Vertex** pptlist, int* ptnum)
{
  // A background mesh must be provided (maybe itself).
  assert(OMT_domain != NULL);
  *pptlist = NULL;
  *ptnum = 0;

  // At the moment, we skip this operation for boundary vertices.
  int vcount = 0; // Count the number of Voronoi vertices.
  TriEdge E = mesh_vertex->adj;
  do {
    vcount++;
    if (E.is_segment()) { //if (E.tri->is_hulltri()) {
      if (op_db_verbose > 3) {
        printf("  Vertex %d is a boundary vertex. Skipped.\n", mesh_vertex->idx);
      }
      return 0;
    }
    E = E.eprev_esym(); // ccw
  } while (E.tri != mesh_vertex->adj.tri);

  if (op_db_verbose) {
    printf("  Found %d adjacent triangles at vertex %d.\n", vcount, mesh_vertex->idx);
  }

  /*  We assume circumcenters are already calculated.
  if (calcflag) {
    // Calculate the circumcenters of triangles at this vertex.
    TriEdge E = mesh_vertex->adj;
    do {
      if (!E.tri->is_hulltri()) {
        get_tri_orthocenter(E.tri);
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);

    E = mesh_vertex->adj;
    do {
      if (E.tri->is_hulltri()) {
        get_hulltri_orthocenter(E.tri);
      }
      E = E.eprev_esym(); // ccw
    } while (E.tri != mesh_vertex->adj.tri);
  }
  */

  // N_Last_Exit is a boundary or hull edge ([a,b,-1]) of the OMT_domain.
  TriEdge N, N_Last_Exit;
  N_Last_Exit.tri = NULL;

  // We need to start from an interior dual edge -- need one interior Voronoi vertex.
  // So the initial condition (N_Last_Exit.tri == NULL) is valid.
  //printf("\n DBG start: \n");
  E = mesh_vertex->adj;
  N = E.esym();
  bool is_interor_dual_edge = (!N.tri->is_dual_in_exterior() && !N.tri->is_dual_on_bdry());
  if (!is_interor_dual_edge) {
    do {
      //printf("  E: [%d,%d,%d] %d\n", E.org()->idx, E.dest()->idx, E.apex()->idx, E.triis_dual_in_exterior());
      //printf("  N: [%d,%d,%d] %d\n", N.org()->idx, N.dest()->idx, N.apex()->idx, N.triis_dual_in_exterior());
      E = E.eprev_esym(); // ccw
      N = E.esym();
      is_interor_dual_edge = (!N.tri->is_dual_in_exterior() && !N.tri->is_dual_on_bdry());
      if (is_interor_dual_edge) break;
    } while (E.tri != mesh_vertex->adj.tri);
    mesh_vertex->adj = E; // Must update this.
  }
  //printf(" DBG end: \n");
  if (!is_interor_dual_edge) {
    printf("All dual edges lie ouside of the OMT_domain.\n");
    return 0;
  }

  // Use an array to save the list of Voronoi or cut vertices (in ccw order).
  // [Remark] Due to the possible cut vertcies, the total number of vertcies
  //   of a power cell may be more than vcount. Make it large enough.
  Vertex *ptlist = new Vertex[vcount * 4];
  int ptcount = 0;

  do {
    if (op_db_verbose > 3) {
      printf("  Get tri: [%d,%d,%d].\n", E.org()->idx, E.dest()->idx, E.apex()->idx);
    }
    N = E.esym();
  
    if (E.tri->is_dual_in_exterior()) {
      if (N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
        // We're still walking in OUTSIDE.
        assert(N_Last_Exit.tri != NULL);
      } else {
        // N's dual vertex lies inside the domain.
        // We're walking from INSIDE to OUTSIDE.
        assert(N_Last_Exit.tri == NULL);
        // Calculating a cut vertex between this dual edge (from N.tri->cct to E.tri->cct)
        //    and a boundary edge (or a segment) S of the background mesh.
        //    We must ensure that the dual edge and S intersect in S's interior.
        TriEdge S = E.tri->on;
        Vertex In_pt, Out_pt;
        In_pt.init();
        Out_pt.init();
        In_pt.crd[0] = N.tri->cct[0];
        In_pt.crd[1] = N.tri->cct[1];
        Out_pt.crd[0] = E.tri->cct[0];
        Out_pt.crd[1] = E.tri->cct[1];

        get_boundary_cut_dualedge(In_pt, Out_pt, S);
        E.tri->on = S; // Update E.tri->on
        //printf("  Last Exit [%d,%d,%d]\n", S.org()->idx, S.dest()->idx, S.apex()->idx);
        
        Vertex *e1 = E.tri->on.org();
        Vertex *e2 = E.tri->on.dest();
        assert((e1 != NULL) && (e2 != NULL));
        double X0 = E.tri->cct[0];
        double Y0 = E.tri->cct[1];
        double X1 = N.tri->cct[0];
        double Y1 = N.tri->cct[1];
        double X2 = e1->crd[0];
        double Y2 = e1->crd[1];
        double X3 = e2->crd[0];
        double Y3 = e2->crd[1];
        double t1 = 0, t2 = 0;
        if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
          double c0 = X0 + t1 * (X1 - X0);
          double c1 = Y0 + t1 * (Y1 - Y0);
          /*
          x1 = (int) ( Sxy * N.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * N.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * c0 + Cx);
          y2 = (int) (-Sxy * c1 + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          */
          Vertex *pt = &(ptlist[ptcount]);
          pt->init();
          pt->crd[0] = c0;
          pt->crd[1] = c1;
          pt->crd[2] = c0*c0 + c1*c1; // not used
          pt->tag = -1; // A cut vertex.
          pt->idx = ptcount+1;
          ptcount++;
        }

        // Remember this boundary edge.
        N_Last_Exit = E.tri->on;
        //printf("  2 Last Exit [%d,%d,%d]\n", N_Last_Exit.org()->idx,
        //       N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
      }
    } else if (E.tri->is_dual_on_bdry()) {
      if (N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
        // We're walking from OUTSIDE to INSIDE
        assert(N_Last_Exit.tri != NULL);
        // Since E's dual vertex lies exactly on boundary, then E.tri->on must contain this dual vertex.
        //   This is guaranteed by the function get_hulltri_orthocenter().
        //   No need to call get_boundary_cut_dualedge().
        //printf("\n DBG start: \n");
        //printf("  N_Last_Exit: [%d,%d,%d]\n", N_Last_Exit.org()->idx, N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
        //printf("  E.tri->on:   [%d,%d,%d]\n", E.tri->on.org()->idx, E.tri->on.dest()->idx, E.tri->on.apex()->idx);
        //printf("\n DBG end: \n");
        if ((E.tri->on.tri != N_Last_Exit.tri) ||
            ((E.tri->on.tri == N_Last_Exit.tri) && (E.tri->on.ver != N_Last_Exit.ver))) {
          // They are on different boundary.
          //printf("  !!! Searching background corners (2) DEBUG!!!\n");
          TriEdge startEdge, endEdge, searchEdge;
          startEdge = N_Last_Exit.esym();
          endEdge = E.tri->on.esym();
          //===================== Subroutine start =================================
          if (op_db_verbose > 3) {
            printf("  Hull startE:   (%d,%d,%d)\n", startEdge.org()->idx,
                   startEdge.dest()->idx, startEdge.apex()->idx);
            printf("  endE:   (%d,%d,%d)\n", endEdge.org()->idx,
                   endEdge.dest()->idx, endEdge.apex()->idx);
          }
          // Start searching from startEdge towards endEdge.
          searchEdge = startEdge;
          while (searchEdge.tri != endEdge.tri) {
            if (op_db_verbose > 3) {
              printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                     searchEdge.dest()->idx, searchEdge.apex()->idx);
            }
            Vertex *pt = &(ptlist[ptcount]);
            pt->init();
            pt->crd[0] = searchEdge.dest()->crd[0];
            pt->crd[1] = searchEdge.dest()->crd[1];
            pt->crd[2] = searchEdge.dest()->crd[2]; // not used
            pt->tag = -2; // A background vertex.
            pt->idx = ptcount+1;
            ptcount++;
            // Go to the next boundary edge (around dest()).
            TriEdge workedge = searchEdge.enext();
            while (true) {
              searchEdge = workedge.esym_enext(); // CW
              if (searchEdge.is_segment()) break;
              if ((searchEdge.esym()).tri->is_hulltri()) break;
              if (op_db_verbose > 3) {
                printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                       workedge.dest()->idx, workedge.apex()->idx);
              }
              workedge = searchEdge;
            }
          }
          //printf("Done!\n");
          //===================== Subroutine end =================================
        }
        // Clear it. We're walking from OUTSIDE to INSIDE already.
        N_Last_Exit.tri = NULL;
      } else {
        // N's dual vertex lies inside the domain.
        // We're walking to the direction from INSIDE to OUTSIDE.
        // Remeber this boundary edge.
        assert(N_Last_Exit.tri == NULL);
        N_Last_Exit = E.tri->on;
        //printf("  2 Last Exit [%d,%d,%d]\n", N_Last_Exit.org()->idx,
        //       N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
        /*
        // They are in the same subdomain.
        x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
        y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
        x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
        y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        */
      }

      Vertex *pt = &(ptlist[ptcount]);
      pt->init();
      pt->crd[0] = E.tri->cct[0];
      pt->crd[1] = E.tri->cct[1];
      pt->crd[2] = E.tri->cct[2];
      pt->tag = E.tri->idx; // A Voronoi vertex.
      pt->idx = ptcount+1;
      ptcount++;
    } else {
      // E's dual vertex lies inside the domain (or subdomain).
      // There must exist a (or part of) dual edge.
      if (!N.tri->is_dual_in_exterior() || N.tri->is_dual_on_bdry()) {
        // We're walking inside of the OMT_domain.
        assert(N_Last_Exit.tri == NULL);
        /*
        // An interior voronoi edge.
        x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
        y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
        x2 = (int) ( Sxy * N.tri->cct[0] + Cx);
        y2 = (int) (-Sxy * N.tri->cct[1] + Cy);
        painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
        */
      } else {
        // The adjacent voronoi vertex is in exterior of the OMT.
        // We're walking from OUTSIDE to INSIDE.
        assert(N_Last_Exit.tri != NULL);
        // Calculating a cut vertex between this dual edge (from N.tri->cct to E.tri->cct)
        //    and a boundary edge (or a segment) S of the background mesh.
        //    We must ensure that the dual edge and S intersect in S's interior.
        TriEdge S = N.tri->on;
        Vertex In_pt, Out_pt;
        In_pt.init();
        Out_pt.init();
        In_pt.crd[0] = E.tri->cct[0];
        In_pt.crd[1] = E.tri->cct[1];
        Out_pt.crd[0] = N.tri->cct[0];
        Out_pt.crd[1] = N.tri->cct[1];

        get_boundary_cut_dualedge(In_pt, Out_pt, S);
        N.tri->on = S; // Update N.tri->on

        //printf("\n DBG start: \n");
        //printf("  N_Last_Exit: [%d,%d,%d]\n", N_Last_Exit.org()->idx, N_Last_Exit.dest()->idx, N_Last_Exit.apex()->idx);
        //printf("  N.tri->on:   [%d,%d,%d]\n", N.tri->on.org()->idx, N.tri->on.dest()->idx, N.tri->on.apex()->idx);
        //printf("\n DBG end: \n");
        if ((N.tri->on.tri != N_Last_Exit.tri) ||
            ((N.tri->on.tri == N_Last_Exit.tri) && (N.tri->on.ver != N_Last_Exit.ver))) {
          // They are on different boundary.
          //printf("  !!! Searching background corners (2) DEBUG!!!\n");
          TriEdge startEdge, endEdge, searchEdge;
          startEdge = N_Last_Exit.esym();
          endEdge = N.tri->on.esym();
          //===================== Subroutine start =================================
          if (op_db_verbose > 3) {
            printf("  startE:   (%d,%d,%d)\n", startEdge.org()->idx,
                   startEdge.dest()->idx, startEdge.apex()->idx);
            printf("  endE:   (%d,%d,%d)\n", endEdge.org()->idx,
                   endEdge.dest()->idx, endEdge.apex()->idx);
          }
          // Start searching from startEdge towards endEdge.
          searchEdge = startEdge;
          while (searchEdge.tri != endEdge.tri) {
            if (op_db_verbose > 3) {
              printf("  Hull searchE: (%d,%d,%d)\n", searchEdge.org()->idx,
                     searchEdge.dest()->idx, searchEdge.apex()->idx);
            }
            Vertex *pt = &(ptlist[ptcount]);
            pt->init();
            pt->crd[0] = searchEdge.dest()->crd[0];
            pt->crd[1] = searchEdge.dest()->crd[1];
            pt->crd[2] = searchEdge.dest()->crd[2]; // not used
            pt->tag = -2; // A background vertex.
            pt->idx = ptcount+1;
            ptcount++;
            // Go to the next boundary edge (around dest()).
            TriEdge workedge = searchEdge.enext();
            while (true) {
              searchEdge = workedge.esym_enext(); // CW
              if (searchEdge.is_segment()) break;
              if ((searchEdge.esym()).tri->is_hulltri()) break;
              if (op_db_verbose > 3) {
                printf("  Hull workE: (%d,%d,%d)\n", workedge.org()->idx,
                       workedge.dest()->idx, workedge.apex()->idx);
              }
              workedge = searchEdge;
            }
          }
          //printf("Done!\n");
          //===================== Subroutine end =================================
        }

        Vertex *e1 = N.tri->on.org();
        Vertex *e2 = N.tri->on.dest();
        assert((e1 != NULL) && (e2 != NULL));
        double X0 = E.tri->cct[0];
        double Y0 = E.tri->cct[1];
        double X1 = N.tri->cct[0];
        double Y1 = N.tri->cct[1];
        double X2 = e1->crd[0];
        double Y2 = e1->crd[1];
        double X3 = e2->crd[0];
        double Y3 = e2->crd[1];
        double t1 = 0, t2 = 0;
        if (line_line_intersection(X0, Y0, X1, Y1, X2, Y2, X3, Y3, &t1, &t2)) {
          double c0 = X0 + t1 * (X1 - X0);
          double c1 = Y0 + t1 * (Y1 - Y0);
          /*
          x1 = (int) ( Sxy * E.tri->cct[0] + Cx);
          y1 = (int) (-Sxy * E.tri->cct[1] + Cy);
          x2 = (int) ( Sxy * c0 + Cx);
          y2 = (int) (-Sxy * c1 + Cy);
          painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          */
          Vertex *pt = &(ptlist[ptcount]);
          pt->init();
          pt->crd[0] = c0;
          pt->crd[1] = c1;
          pt->crd[2] = c0*c0 + c1*c1; // not used
          pt->tag = -1; // A cut vertex.
          pt->idx = ptcount+1;
          ptcount++;
        }
      }

      Vertex *pt = &(ptlist[ptcount]);
      pt->init();
      pt->crd[0] = E.tri->cct[0];
      pt->crd[1] = E.tri->cct[1];
      pt->crd[2] = E.tri->cct[2];
      pt->tag = E.tri->idx; // A Voronoi vertex.
      pt->idx = ptcount+1;
      ptcount++;

      // Clear it, we're inside the OMT_domain again.
      N_Last_Exit.tri = NULL;
    }
  
    // Go to the next edge.
    E = E.eprev_esym(); // ccw
  } while (E.tri != mesh_vertex->adj.tri);

  *pptlist = ptlist;
  *ptnum = ptcount;

  return 1;
}

//==============================================================================

int Triangulation::get_mass_center(Vertex* ptlist, int ptnum, REAL mc[2])
{
  // Calculate cell mass center
  mc[0] = mc[1] = 0.;
  
  // The formula from Wikipedia.
  double Area = 0.; // signed area.
  for (int i = 0; i < ptnum; i++) {
    Vertex *v1 = &(ptlist[i]);
    Vertex *v2 = &(ptlist[(i+1)%ptnum]);
    double x0 = v1->crd[0];
    double y0 = v1->crd[1];
    double x1 = v2->crd[0];
    double y1 = v2->crd[1];
    Area += 0.5 * (x0 * y1 - x1 * y0);
  }

  double mcx = 0., mcy = 0.;
  for (int i = 0; i < ptnum; i++) {
    Vertex *v1 = &(ptlist[i]);
    Vertex *v2 = &(ptlist[(i+1)%ptnum]);
    double x0 = v1->crd[0];
    double y0 = v1->crd[1];
    double x1 = v2->crd[0];
    double y1 = v2->crd[1];
    mcx += (x0 + x1)*(x0 * y1 - x1 * y0);
    mcy += (y0 + y1)*(x0 * y1 - x1 * y0);
  }
  mcx /= (6 * Area);
  mcy /= (6 * Area);

  mc[0] = mcx;
  mc[1] = mcy;
  return 1;
}
