#include <QtWidgets>
#include <QInputDialog>
#include <QSignalMapper>
#include <QDir>

#if defined(QT_PRINTSUPPORT_LIB)
#include <QtPrintSupport/qtprintsupportglobal.h>
#if QT_CONFIG(printdialog)
#include <QPrinter>
#include <QPrintDialog>
#endif
#endif

#include "scribblearea.h"
#include "detri2qt.h"
#include <assert.h>

using namespace detri2;

//! [0]
ScribbleArea::ScribbleArea(QToolBar *toolbar_top, QToolBar *toolbar_left,
                           QStatusBar *statusbar, QMenuBar *menubar,
                           QWidget *parent)
    : QWidget(parent)
{
    _menubar = menubar;

    _toolbar_top  = toolbar_top;
    _toolbar_left = toolbar_left;
    _statusbar = statusbar;

    m_statusLeft = new QLabel("Left", this);
    m_statusLeft->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    m_statusMiddle = new QLabel("Middle", this);
    m_statusMiddle->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    m_statusRight = new QLabel("Right", this);
    m_statusRight->setFrameStyle(QFrame::Panel | QFrame::Sunken);

    _statusbar->addPermanentWidget(m_statusLeft, 1);
    _statusbar->addPermanentWidget(m_statusMiddle, 2);
    _statusbar->addPermanentWidget(m_statusRight, 4);

    createActions();
    createMenus();

    _geom = NULL;
    _tria = NULL;
    _active_vertex  = NULL;
    _active_segment = NULL;
    _active_trilist = NULL;

    reset_scribble();
    reset_preference();

    // Initialize an empty geometry
    detri2_init_with_bbox(&_geom, -1., 1., -1., 1.);

    QImage newImage(QSize(width(), height()), QImage::Format_RGB32);
    _canvas[0] = _canvas[1] = newImage;
    //qDebug() << "canvas width = " << _canvas.size().width() << " height = " << _canvas.size().height();

    _background_img = NULL;

    _newfilepabel = NULL;
    _meshoptionPanel = NULL;
    _hdeoptionPanel = NULL;
    _sortoptionpanel = NULL;
    _vertexPanel = NULL;
    _segmentPanel = NULL;
    _subdomainPanel = NULL;
    _imgView = NULL;
    _preferdialog = NULL;

    _file_last_used_path = QDir::home().dirName(); // Default is my home directory

    // to receive key press events
    //setFocusPolicy(Qt::StrongFocus);
    setMouseTracking(true);
}

//=============================================================================

void ScribbleArea::clean_scribble() // only called by deletion
{
    if (_tria != NULL) {
      delete _tria;
      _tria = NULL;
    }

    if (_geom != NULL) {
      delete _geom;
      _geom = NULL;
    }
}

void ScribbleArea::reset_scribble()
{
    _cx = _cy = 0.;
    _scale_xy = 1.0;

    // Mouse actions.
    //_trans_mode = true;
    transAct->setChecked(true);  // default
    _insert_mode = 0; 
    insertVertexAct->setChecked(false); //_insert_vertex = 0;
    insertSegmentAct->setChecked(false); //_insert_segment = 0;
    insertRectAct->setChecked(false); //_insert_rect = 0;
    insertArcAct->setChecked(false); //_insert_arc = 0;
    _select_mode = 0;
    assignVertexAct->setChecked(false); //_assign_vertex = 0;
    assignVertexWeightAct->setChecked(false);     //_assign_weight = 0;
    assignSegmentAct->setChecked(false);     //_assign_segment = 0;
    assignSubdomainAct->setChecked(false);     //_assign_subdomain = 0;

    viewPowerCellAct->setChecked(false);     //_view_power_cell = 0;
    viewMassCenterAct->setChecked(false);
    viewTriangleInfoAct->setChecked(false);     //_view_triangle_info = false;
    viewOthocircleAct->setChecked(false);     //_view_triangle_orthocircle = 0;
    viewOthocenterAct->setChecked(false);     //_view_triangle_orthocenter = 0;

    zoomSelectAct->setChecked(false);     //_zoom_selected_region = 0;
    _select_x = _select_y = _select_width = _select_height = 0;

    viewVerticesAct->setChecked(true);     //_show_geom_vertices = true;
    viewSegmentsAct->setChecked(true);     //_show_geom_edges = true;
    viewSubdomainsAct->setChecked(false);     //_show_geom_subdomains = false;
    viewBboxAct->setChecked(true);     //_show_geom_bbox = true;
    viewBgMeshAct->setChecked(false);     //_show_geom_tris = true;
    viewWeightsAct->setChecked(false);     //_show_weights = false;
    viewConvexHullAct->setChecked(false);     //_show_convex_hull = false;
    viewTrianglesAct->setChecked(true);     //_show_triangles = true;
    viewOthocirclesAct->setChecked(false);     //_show_ortho_circle = false;
    viewPowerDiagramAct->setChecked(false);     //_show_voronoi = false;
    viewMassCentersAct->setChecked(false);
    viewLaplacianCentersAct->setChecked(false);
    viewVertexLabelsAct->setChecked(false);     //_show_vertex_labels = false;
    viewSegmentLabelsAct->setChecked(false);     //_show_segment_labels = false;
    viewMeshVertexLabelsAct->setChecked(false);

    _mouse_sens=0.5;
    _mouse_down=_mouse_down_move=0;
    _mouse_last_x=_mouse_last_y=0;

    if (_active_vertex) {
      // Clear all active (selected) vertices.
      Vertex *vrt = _active_vertex, *bak;
      while (vrt != NULL) {
        bak = vrt; // for clearing the ->Next
        vrt = vrt->Next;
        bak->Next = NULL;
      }
      _active_vertex = NULL;
    }
    if (_active_segment) {
      _active_segment = NULL;
    }
    if (_active_trilist) {
      delete _active_trilist;
      _active_trilist = NULL;
    }

    m_statusMiddle->setText("Move");
    m_statusRight->setText("Hold mouse to move the scene");

    _object_valid = 0;
}

//=============================================================================

void ScribbleArea::reset_preference()
{
    _draw_vertex_size = 4;
    _draw_segment_width = 2;
    _draw_edge_width = 1;
    _draw_dual_edge_width = 2;
    _draw_ccircle_width = 2;

    _color_canvas_background = QColor(Qt::white);
    _color_vertex = QColor(Qt::black);
    _color_edge = QColor(Qt::black);
    _color_dual_edge = QColor(Qt::green);
    _color_segment = QColor(Qt::blue);
    _color_ccircle = QColor(Qt::cyan);
    _color_selected = QColor(Qt::red);

    _font_bbox_label = QFont("Helvetica [Cronyx]", 12);
    _font_vertex_label = QFont("Times", 12, QFont::Normal);
    _font_segment_label = QFont("Times", 12, QFont::Bold); // QFont::Normal
    _font_subdomain_label = QFont("Times", 16, QFont::Normal);
    _font_meshvertex_label = QFont("Times", 10, QFont::Normal);

    _enable_debug = 1; // default
    _file_debug = QString("/Users/si/tmp/dump");
}

//=============================================================================

void ScribbleArea::createActions()
{
    QPixmap pix_file_new(":/images/file_new.png"); // 32x32
    QPixmap pix_file_open(":/images/file_open.png");
    QPixmap pix_file_save(":/images/file_save_as.png");
    //pix_file_new = pix_file_new.scaled(16,16);
    //pix_file_open = pix_file_open.scaled(16,16);
    //pix_file_save = pix_file_save.scaled(16,16);

    QPixmap pix_move(":/images/icons8-move-32.png");
    QPixmap pix_zoomIn(":/images/icons8-zoom-in-32.png");
    QPixmap pix_zoomOut(":/images/icons8-zoom-out-32.png");

    QPixmap pix_point(":/images/icon-point.png");
    QPixmap pix_line(":/images/icon-line.png");
    QPixmap pix_rect(":/images/icon-rectangle.png");
    QPixmap pix_arc(":/images/icon-circle.png");
    //pix_point = pix_point.scaled(16,16);
    //pix_line = pix_line.scaled(16,16);
    //pix_rect = pix_rect.scaled(16,16);
    //pix_arc = pix_arc.scaled(16,16);

    QPixmap pix_mesh(":/images/icon-mesh.png");
    QPixmap pix_dt(":/images/icon-dt.png");
    QPixmap pix_fdt(":/images/icon-fdt.png");
    QPixmap pix_mesh_clean(":/images/icon-mesh-clean.png");
    //pix_mesh = pix_mesh.scaled(16,16);
    //pix_dt = pix_dt.scaled(16,16);
    //pix_fdt = pix_fdt.scaled(16,16);
    //pix_mesh_clean = pix_mesh_clean.scaled(16,16);

    QPixmap pix_select_point(":/images/icon-select-point.png");
    QPixmap pix_select_segment(":/images/icon-select-segment.png");
    QPixmap pix_select_subdomain(":/images/icon-select-subdomain.png");

    QPixmap pix_delete(":/images/icon-delete.png");

    newAct = new QAction(QIcon(pix_file_new), tr("&New..."), this);
    //newAct = new QAction(tr("&New..."), this);
    newAct->setShortcuts(QKeySequence::New);
    newAct->setStatusTip(tr("Create a new geometry"));
    connect(newAct, &QAction::triggered, this, &ScribbleArea::newFile);

    openAct = new QAction(QIcon(pix_file_open), tr("&Open..."), this);
    openAct->setShortcuts(QKeySequence::Open);
    openAct->setStatusTip(tr("Open a file contains a geometry or a mesh"));
    connect(openAct, &QAction::triggered, this, &ScribbleArea::openFile);

    addAct = new QAction(tr("&Add..."), this);
    addAct->setShortcut(QKeySequence(tr("Ctrl+A")));
    addAct->setStatusTip(tr("Add a geometry"));
    connect(addAct, &QAction::triggered, this, &ScribbleArea::addFile);

    saveAct = new QAction(QIcon(pix_file_save), tr("&Save As..."), this);
    saveAct->setShortcuts(QKeySequence::Save);
    saveAct->setStatusTip(tr("Save geometry/mesh/image to a file"));
    connect(saveAct, &QAction::triggered, this, &ScribbleArea::saveFile);

    closeAct = new QAction(tr("Close"), this);
    closeAct->setShortcuts(QKeySequence::Close);
    connect(closeAct, &QAction::triggered, this, &ScribbleArea::closeFile);

    UserPreferAct = new QAction(tr("User preference ..."), this);
    //UserPreferAct->setShortcut(QKeySequence(tr("U")));
    connect(UserPreferAct, &QAction::triggered, this, &ScribbleArea::open_preferDialog);

    UserExitAct = new QAction(tr("Quit Detri2"), this);
    UserExitAct->setShortcuts(QKeySequence::Quit);
    //connect(UserExitAct, &QAction::triggered, this, &QCoreApplication::quit);
    connect(UserExitAct, &QAction::triggered, this, &ScribbleArea::on_user_exit);

    // edit menu actions
    insertVertexAct = new QAction(QIcon(pix_point), tr("Add Vertices"), this);
    insertVertexAct->setShortcut(QKeySequence(tr("A")));
    insertVertexAct->setCheckable(true);
    connect(insertVertexAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    insertSegmentAct = new QAction(QIcon(pix_line), tr("Add Segments"), this);
    insertSegmentAct->setShortcut(QKeySequence(tr("E")));
    insertSegmentAct->setCheckable(true);
    connect(insertSegmentAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    insertRectAct = new QAction(QIcon(pix_rect), tr("Add Rectangles..."), this);
    insertRectAct->setShortcut(QKeySequence(tr("B")));
    insertRectAct->setCheckable(true);
    connect(insertRectAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    insertArcAct = new QAction(QIcon(pix_arc), tr("Add Arcs..."), this);
    insertArcAct->setShortcut(QKeySequence(tr("C")));
    insertArcAct->setCheckable(true);
    connect(insertArcAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    assignVertexAct = new QAction(QIcon(pix_select_point), tr("Select Vertex"), this);
    assignVertexAct->setShortcut(QKeySequence(tr("V")));
    assignVertexAct->setCheckable(true);
    connect(assignVertexAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    assignSegmentAct = new QAction(QIcon(pix_select_segment), tr("Select Segments"), this);
    assignSegmentAct->setShortcut(QKeySequence(tr("S")));
    assignSegmentAct->setCheckable(true);
    connect(assignSegmentAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    assignSubdomainAct = new QAction(QIcon(pix_select_subdomain), tr("Select Subdomains"), this);
    assignSubdomainAct->setShortcut(QKeySequence(tr("R")));
    assignSubdomainAct->setCheckable(true);
    connect(assignSubdomainAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    assignVertexWeightAct = new QAction(tr("Assign Weights"), this);
    assignVertexWeightAct->setShortcut(QKeySequence(tr("W")));
    assignVertexWeightAct->setCheckable(true);
    connect(assignVertexWeightAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    setAttributeAct = new QAction(tr("Assign Attributes..."), this);
    setAttributeAct->setShortcut(Qt::Key_Space);
    connect(setAttributeAct, &QAction::triggered, this, &ScribbleArea::edit_attributes_act);

    deleteAct = new QAction(QIcon(pix_delete), tr("Delete selected item"), this);
    deleteAct->setShortcut(QKeySequence(tr("X")));
    connect(deleteAct, &QAction::triggered, this, &ScribbleArea::edit_delete_act);

    scaleAct = new QAction(tr("Scale..."), this);
    //scaleAct->setShortcut(QKeySequence(tr("X")));
    connect(scaleAct, &QAction::triggered, this, &ScribbleArea::edit_scale_act);

    // View menu actions
    viewVerticesAct = new QAction(tr("Show Vertices"), this);
    viewVerticesAct->setShortcut(QKeySequence(tr("Alt+P")));
    viewVerticesAct->setCheckable(true);
    viewVerticesAct->setChecked(true);
    connect(viewVerticesAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewWeightsAct = new QAction(tr("Show weights"), this);
    viewWeightsAct->setShortcut(QKeySequence(tr("Alt+W")));
    viewWeightsAct->setCheckable(true);
    viewWeightsAct->setChecked(true);
    connect(viewWeightsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewVertexLabelsAct = new QAction(tr("Vertex Labels"), this);
    viewVertexLabelsAct->setShortcut(QKeySequence(tr("Alt+L")));
    viewVertexLabelsAct->setCheckable(true);
    viewVertexLabelsAct->setChecked(false);
    connect(viewVertexLabelsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewSegmentsAct = new QAction(tr("Show Segments"), this);
    viewSegmentsAct->setShortcut(QKeySequence(tr("Alt+S")));
    viewSegmentsAct->setCheckable(true);
    viewSegmentsAct->setChecked(true);
    connect(viewSegmentsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewSegmentLabelsAct = new QAction(tr("Segments Labels"), this);
    //viewSegmentsAct->setShortcut(QKeySequence(tr("Alt+S")));
    viewSegmentLabelsAct->setCheckable(true);
    viewSegmentLabelsAct->setChecked(false);
    connect(viewSegmentLabelsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewSubdomainsAct = new QAction(tr("Subdomains"), this);
    viewSubdomainsAct->setShortcut(QKeySequence(tr("Alt+R")));
    viewSubdomainsAct->setCheckable(true);
    //viewSubdomainsAct->setChecked(true);
    connect(viewSubdomainsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewConvexHullAct = new QAction(tr("Show hull"), this);
    viewConvexHullAct->setShortcut(QKeySequence(tr("Alt+H")));
    viewConvexHullAct->setCheckable(true);
    //viewConvexHullAct->setChecked(true);
    connect(viewConvexHullAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewTrianglesAct = new QAction(tr("Show mesh"), this);
    viewTrianglesAct->setShortcut(QKeySequence(tr("Alt+T")));
    viewTrianglesAct->setCheckable(true);
    viewTrianglesAct->setChecked(true);
    connect(viewTrianglesAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewMeshVertexLabelsAct = new QAction(tr("Mesh Vertex Labels"), this);
    //viewMeshVertexLabelsAct->setShortcut(QKeySequence(tr("Alt+L")));
    viewMeshVertexLabelsAct->setCheckable(true);
    viewMeshVertexLabelsAct->setChecked(false);
    connect(viewMeshVertexLabelsAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewTriangleInfoAct = new QAction(tr("Show Triangle"), this);
    viewTriangleInfoAct->setCheckable(true);
    viewTriangleInfoAct->setChecked(false);
    connect(viewTriangleInfoAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewOthocirclesAct = new QAction(tr("Show orthocircles"), this);
    viewOthocirclesAct->setShortcut(QKeySequence(tr("Alt+O")));
    viewOthocirclesAct->setCheckable(true);
    viewOthocirclesAct->setChecked(false);
    connect(viewOthocirclesAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewOthocircleAct = new QAction(tr("Tri orthocircle"), this);
    //viewOthocircleAct->setShortcut(QKeySequence(tr("Alt+O")));
    viewOthocircleAct->setCheckable(true);
    viewOthocircleAct->setChecked(false);
    connect(viewOthocircleAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewOthocenterAct = new QAction(tr("Tri orthocenter"), this);
    //viewOthocircleAct->setShortcut(QKeySequence(tr("Alt+O")));
    viewOthocenterAct->setCheckable(true);
    viewOthocenterAct->setChecked(false);
    connect(viewOthocenterAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewPowerDiagramAct = new QAction(tr("Power (Voronoi) diagram"), this);
    viewPowerDiagramAct->setShortcut(QKeySequence(tr("Alt+V")));
    viewPowerDiagramAct->setCheckable(true);
    viewPowerDiagramAct->setChecked(false);
    connect(viewPowerDiagramAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewMassCentersAct = new QAction(tr("Cell Mass centers"), this);;
    //viewMassCentersAct->setShortcut(QKeySequence(tr("V")));
    viewMassCentersAct->setCheckable(true);
    viewMassCentersAct->setChecked(false);
    connect(viewMassCentersAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewLaplacianCentersAct = new QAction(tr("Laplacian centers"), this);;
    //viewMassCentersAct->setShortcut(QKeySequence(tr("V")));
    viewLaplacianCentersAct->setCheckable(true);
    viewLaplacianCentersAct->setChecked(false);
    connect(viewLaplacianCentersAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewPowerCellAct = new QAction(tr("Power (Voronoi) cell"), this);;
    //viewPowerCellAct->setShortcut(QKeySequence(tr("V")));
    viewPowerCellAct->setCheckable(true);
    viewPowerCellAct->setChecked(false);
    connect(viewPowerCellAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewMassCenterAct = new QAction(tr("Cell Mass center"), this);;
    //viewMassCenterAct->setShortcut(QKeySequence(tr("V")));
    viewMassCenterAct->setCheckable(true);
    viewMassCenterAct->setChecked(false);
    connect(viewMassCenterAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewBboxAct = new QAction(tr("Bounding box"), this);
    viewBboxAct->setShortcut(QKeySequence(tr("Alt+B")));
    viewBboxAct->setCheckable(true);
    viewBboxAct->setChecked(true);
    connect(viewBboxAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    viewBgMeshAct = new QAction(tr("Background Grid"), this);
    viewBgMeshAct->setCheckable(true);
    viewBgMeshAct->setChecked(true); // for debugging (set false for release version).
    connect(viewBgMeshAct, &QAction::triggered, this, &ScribbleArea::view_update_act);

    MoveLeftAct = new QAction(tr("Move Left"), this);
    MoveLeftAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Left));
    MoveRightAct = new QAction(tr("Move Right"), this);
    MoveRightAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Right));
    MoveUpAct = new QAction(tr("Move Up"), this);
    MoveUpAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Up));
    MoveDownAct = new QAction(tr("Move Down"), this);
    MoveDownAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Down));

    // https://stackoverflow.com/questions/5153157/passing-an-argument-to-a-slot
    QSignalMapper* signalMapper = new QSignalMapper (this);
    connect(MoveLeftAct, SIGNAL(triggered()), signalMapper, SLOT(map())) ;
    connect(MoveRightAct, SIGNAL(triggered()), signalMapper, SLOT(map())) ;
    connect(MoveUpAct, SIGNAL(triggered()), signalMapper, SLOT(map())) ;
    connect(MoveDownAct, SIGNAL(triggered()), signalMapper, SLOT(map())) ;

    signalMapper->setMapping (MoveLeftAct, 1) ;
    signalMapper->setMapping (MoveRightAct, 2) ;
    signalMapper->setMapping (MoveUpAct, 3) ;
    signalMapper->setMapping (MoveDownAct, 4) ;
    connect(signalMapper, SIGNAL(mapped(int)), this, SLOT(view_trans_act(int))) ;

    transAct = new QAction(QIcon(pix_move),tr("Move by Mouse"), this);
    transAct->setShortcut(QKeySequence(tr("Ctrl+1")));
    transAct->setCheckable(true);
    transAct->setEnabled(true);
    connect(transAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    zoomInAct = new QAction(QIcon(pix_zoomIn), tr("Zoom In"), this);
    zoomInAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Plus)); // QKeySequence::ZoomIn
    zoomOutAct = new QAction(QIcon(pix_zoomOut), tr("Zoom Out"), this);
    zoomOutAct->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_Minus)); // QKeySequence::ZoomOut

    QSignalMapper* signalMapperZoom = new QSignalMapper (this);
    connect(zoomInAct, SIGNAL(triggered()), signalMapperZoom, SLOT(map())) ;
    connect(zoomOutAct, SIGNAL(triggered()), signalMapperZoom, SLOT(map())) ;
    //connect(zoomSelectAct, SIGNAL(triggered()), signalMapperZoom, SLOT(map())) ;
    signalMapperZoom->setMapping (zoomInAct, 1) ;
    signalMapperZoom->setMapping (zoomOutAct, 2) ;
    //signalMapperZoom->setMapping (zoomSelectAct, 3) ;
    connect(signalMapperZoom, SIGNAL(mapped(int)), this, SLOT(view_zoom_act(int))) ;

    zoomSelectAct = new QAction(tr("Zoom to Select"), this);
    zoomSelectAct->setShortcut(QKeySequence(tr("Ctrl+8")));
    zoomSelectAct->setCheckable(true);
    zoomSelectAct->setEnabled(true);
    connect(zoomSelectAct, &QAction::triggered, this, &ScribbleArea::edit_mouse_act);

    viewResetAct = new QAction(tr("Reset View"), this);
    viewResetAct->setShortcut(QKeySequence(tr("Ctrl+0")));
    connect(viewResetAct, &QAction::triggered, this, &ScribbleArea::view_reset);

    // mesh Actions
    weightedDelaunayAct = new QAction(QIcon(pix_dt), tr("Genrate weighted Delaunay"), this);
    weightedDelaunayAct->setShortcut(QKeySequence(tr("Ctrl+D")));
    connect(weightedDelaunayAct, &QAction::triggered, this, &ScribbleArea::weighted_Delaunay_act);

    furthestDelaunayAct = new QAction(QIcon(pix_fdt), tr("Genrate furthest Delaunay"), this);
    furthestDelaunayAct->setShortcut(QKeySequence(tr("Ctrl+F")));
    connect(furthestDelaunayAct, &QAction::triggered, this, &ScribbleArea::furthest_Delaunay_act);

    cleanMeshAct = new QAction(QIcon(pix_mesh_clean), tr("Clean Mesh"), this);
    //cleanMeshAct->setShortcut(QKeySequence(tr("Alt+C")));
    connect(cleanMeshAct, &QAction::triggered, this, &ScribbleArea::clean_mesh_act);
    //connect(cleanMeshAct, SIGNAL(QAction::triggered), this, SLOT(ScribbleArea::clean_mesh_act));

    detri2OptionsAct = new QAction(QIcon(pix_mesh), tr("Generate Mesh..."), this);
    detri2OptionsAct->setShortcut(QKeySequence(tr("Ctrl+M")));
    connect(detri2OptionsAct, &QAction::triggered, this, &ScribbleArea::detri2_options_act);

    detri2HDEAct = new QAction(tr("Detri2 HDE..."), this);
    //detri2HDEAct->setShortcut(QKeySequence(tr("Ctrl+A")));
    connect(detri2HDEAct, &QAction::triggered, this, &ScribbleArea::detri2_HDE_option_act);

    detri2SortAct = new QAction(tr("Sorting..."), this);
    connect(detri2SortAct, &QAction::triggered, this, &ScribbleArea::detri2_Sort_option_act);

    // Toolbar Top
    _toolbar_top->addAction(newAct);
    _toolbar_top->addAction(openAct);
    _toolbar_top->addAction(saveAct);

    _toolbar_top->addSeparator();
    _toolbar_top->addAction(transAct);
    _toolbar_top->addAction(zoomInAct);
    _toolbar_top->addAction(zoomOutAct);

    _toolbar_top->addSeparator();
    _toolbar_top->addAction(insertVertexAct);
    _toolbar_top->addAction(insertSegmentAct);
    _toolbar_top->addAction(insertRectAct);
    _toolbar_top->addAction(insertArcAct);

    _toolbar_top->addSeparator();
    _toolbar_top->addAction(assignVertexAct);
    _toolbar_top->addAction(assignSegmentAct);
    _toolbar_top->addAction(assignSubdomainAct);
    _toolbar_top->addSeparator();

    _toolbar_top->addAction(deleteAct);

    _toolbar_top->addSeparator();
    _toolbar_top->addAction(weightedDelaunayAct);
    //_toolbar_top->addAction(furthestDelaunayAct);
    _toolbar_top->addAction(detri2OptionsAct);
    _toolbar_top->addAction(cleanMeshAct);
    _toolbar_top->addSeparator();

    _toolbar_left->addAction(viewResetAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewBboxAct);
    //_toolbar_left->addAction(viewBgMeshAct);
    _toolbar_left->addAction(viewVerticesAct);
    _toolbar_left->addAction(viewVertexLabelsAct);
    _toolbar_left->addAction(viewSegmentsAct);
    _toolbar_left->addAction(viewSegmentLabelsAct);
    _toolbar_left->addAction(viewSubdomainsAct);
    _toolbar_left->addAction(viewConvexHullAct);
    _toolbar_left->addAction(viewTrianglesAct);
    _toolbar_left->addAction(viewMeshVertexLabelsAct);
    _toolbar_left->addAction(viewOthocirclesAct);
    _toolbar_left->addAction(viewPowerDiagramAct);
    _toolbar_left->addAction(viewMassCentersAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewWeightsAct);
    _toolbar_left->addAction(assignVertexWeightAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewTriangleInfoAct);
    _toolbar_left->addAction(viewOthocircleAct);
    _toolbar_left->addAction(viewOthocenterAct);
    _toolbar_left->addAction(viewPowerCellAct);
    _toolbar_left->addAction(viewMassCenterAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(zoomSelectAct);

    // Debug options
    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewLaplacianCentersAct);
}

void ScribbleArea::createMenus()
{
    //QMenuBar *menubar = new QMenuBar(this);
    QMenuBar *menubar = _menubar;

    //QMenu *fileMenu = new QMenu();
    QMenu *fileMenu = menubar->addMenu(tr("File"));
    fileMenu->addAction(newAct);
    fileMenu->addAction(openAct);
    fileMenu->addAction(addAct);
    fileMenu->addAction(saveAct);
    fileMenu->addAction(closeAct);
    fileMenu->addSeparator();
    fileMenu->addAction(UserPreferAct);
    fileMenu->addSeparator();
    fileMenu->addAction(UserExitAct);

    //QMenu *editMenu = new QMenu();
    QMenu *editMenu = menubar->addMenu(tr("Edit"));
    //editMenu->addAction(transAct);
    //editMenu->addAction(zoomSelectAct);
    //editMenu->addSeparator();
    editMenu->addAction(insertVertexAct);
    editMenu->addAction(insertSegmentAct);
    editMenu->addAction(insertRectAct);
    editMenu->addAction(insertArcAct);
    editMenu->addSeparator();
    //editMenu->addAction(selectAct);
    editMenu->addAction(assignVertexAct);
    editMenu->addAction(assignSegmentAct);
    editMenu->addAction(assignSubdomainAct);

    //selectAct->setChecked(true);
    assignVertexAct->setChecked(true); // default
    editMenu->addSeparator();
    editMenu->addAction(setAttributeAct);
    editMenu->addAction(assignVertexWeightAct);
    editMenu->addSeparator();
    editMenu->addAction(deleteAct);
    editMenu->addSeparator();
    editMenu->addAction(scaleAct);

    QMenu *viewMenu = menubar->addMenu(tr("View"));
    QMenu* subMenuViewGeom = viewMenu->addMenu("View Geometry");
    subMenuViewGeom->addAction(viewVerticesAct);
    subMenuViewGeom->addAction(viewWeightsAct);
    subMenuViewGeom->addAction(viewVertexLabelsAct);
    subMenuViewGeom->addAction(viewSegmentsAct);
    subMenuViewGeom->addAction(viewSegmentLabelsAct);
    subMenuViewGeom->addAction(viewSubdomainsAct);
    subMenuViewGeom->addAction(viewBboxAct);
    subMenuViewGeom->addAction(viewBgMeshAct);
    QMenu* subMenuViewMesh = viewMenu->addMenu("View Mesh");
    subMenuViewMesh->addAction(viewTrianglesAct);
    subMenuViewGeom->addAction(viewMeshVertexLabelsAct);
    subMenuViewMesh->addAction(viewOthocirclesAct);
    subMenuViewMesh->addAction(viewOthocircleAct);
    subMenuViewMesh->addAction(viewOthocenterAct);
    subMenuViewMesh->addAction(viewPowerDiagramAct);
    subMenuViewMesh->addAction(viewMassCentersAct);
    subMenuViewMesh->addAction(viewLaplacianCentersAct);
    subMenuViewMesh->addAction(viewPowerCellAct);
    subMenuViewMesh->addAction(viewMassCenterAct);
    viewMenu->addSeparator();
    viewMenu->addAction(transAct);
    viewMenu->addSeparator();
    viewMenu->addAction(MoveLeftAct);
    viewMenu->addAction(MoveRightAct);
    viewMenu->addAction(MoveUpAct);
    viewMenu->addAction(MoveDownAct);
    viewMenu->addSeparator();
    viewMenu->addAction(zoomInAct);
    viewMenu->addAction(zoomOutAct);
    viewMenu->addAction(zoomSelectAct);
    viewMenu->addSeparator();
    viewMenu->addAction(viewResetAct);

    QMenu *meshMenu = menubar->addMenu(tr("Generate"));
    meshMenu->addAction(detri2OptionsAct);
    meshMenu->addAction(detri2HDEAct);
    meshMenu->addAction(detri2SortAct);
    meshMenu->addSeparator();
    meshMenu->addAction(weightedDelaunayAct);
    meshMenu->addAction(furthestDelaunayAct);
    meshMenu->addSeparator();
    meshMenu->addAction(cleanMeshAct);

    // Mouse actions
    QActionGroup *actionGroup = new QActionGroup(this);
    actionGroup->addAction(transAct);
    actionGroup->addAction(zoomSelectAct);
    //actionGroup->addAction(insertAct);
    actionGroup->addAction(insertVertexAct);
    actionGroup->addAction(insertSegmentAct);
    actionGroup->addAction(insertRectAct);
    actionGroup->addAction(insertArcAct);
    //actionGroup->addAction(selectAct);
    actionGroup->addAction(assignVertexAct);
    actionGroup->addAction(assignSegmentAct);
    actionGroup->addAction(assignSubdomainAct);
    actionGroup->addAction(assignVertexWeightAct);
    actionGroup->addAction(viewTriangleInfoAct);
    actionGroup->addAction(viewPowerCellAct);
    actionGroup->addAction(viewMassCenterAct);
    actionGroup->addAction(viewOthocircleAct);
    actionGroup->addAction(viewOthocenterAct);
}

//=============================================================================

void ScribbleArea::edit_mouse_act() // mouse action
{
    //_trans_mode = transAct->isChecked();
    if (transAct->isChecked()) {
      m_statusMiddle->setText("Move");
      m_statusRight->setText("Hold mouse to move the scene");
    }
    //_zoom_selected_region = zoomSelectAct->isChecked();
    if (zoomSelectAct->isChecked()) {
      m_statusMiddle->setText("Zoom by region");
      m_statusRight->setText("Hold mouse to select a region");
    }

    //_insert_vertex = insertVertexAct->isChecked();
    if (insertVertexAct->isChecked()) {
      m_statusMiddle->setText("Add vertices");
      m_statusRight->setText("Press (hold) mouse to add (move) a vertex");
    }
    //_insert_segment = insertSegmentAct->isChecked();
    if (insertSegmentAct->isChecked()) {
      m_statusMiddle->setText("Add segments");
      m_statusRight->setText("Select two vertices to add a segment");
    }
    //_insert_rect = insertRectAct->isChecked();
    if (insertRectAct->isChecked()) {
      m_statusMiddle->setText("Add rectangles");
      m_statusRight->setText("Use mouse to select a corner, then drag mouse");
    }
    //_insert_arc = insertArcAct->isChecked();
    if (insertArcAct->isChecked()) {
      m_statusMiddle->setText("Add circles/arcs");
      m_statusRight->setText("Use mouse to select a center, then drag mouse");
    }
    //_insert_mode = (_insert_vertex || _insert_segment || _insert_rect || _insert_arc);
    _insert_mode = (insertVertexAct->isChecked() ||
                    insertSegmentAct->isChecked() ||
                    insertRectAct->isChecked() ||
                    insertArcAct->isChecked());

    //_assign_vertex = assignVertexAct->isChecked();
    if (assignVertexAct->isChecked()) {
      m_statusMiddle->setText("Select/Assign vertex");
      m_statusRight->setText("Select a vertex, then press Space key, Hold Shift key to select multiple vertices");
    }
    //_assign_weight = assignVertexWeightAct->isChecked();
    if (assignVertexWeightAct->isChecked()) {
      m_statusMiddle->setText("Select/Assign vertex weight");
      m_statusRight->setText("Select a vertex, then hold and drag mouse");
    }
    //_assign_segment = assignSegmentAct->isChecked();
    if (assignSegmentAct->isChecked()) {
      m_statusMiddle->setText("Select/Assign segment");
      m_statusRight->setText("Select a segment, then press Space key");
    }
    //_assign_subdomain = assignSubdomainAct->isChecked();
    if (assignSubdomainAct->isChecked()) {
      m_statusMiddle->setText("Select/Assign subdomain");
      m_statusRight->setText("Select a region, then press Space key");
    }

    //_select_mode = (_assign_vertex || _assign_weight || _assign_segment || _assign_subdomain);
    _select_mode = (assignVertexAct->isChecked() ||
                    assignVertexWeightAct->isChecked() ||
                    assignSegmentAct->isChecked() ||
                    assignSubdomainAct->isChecked());

    //_view_triangle_info = viewTriangleInfoAct->isChecked();
    if (viewTriangleInfoAct->isChecked()) {
      m_statusMiddle->setText("View triangle information");
      m_statusRight->setText("Select a triangle");
    }

    //_view_power_cell = viewPowerCellAct->isChecked();
    if (viewPowerCellAct->isChecked()) {
      m_statusMiddle->setText("View Power (Voronoi) cell");
      m_statusRight->setText("Select a vertex");
    }

    if (viewMassCenterAct->isChecked()) {
      m_statusMiddle->setText("View mass center of Power cell");
      m_statusRight->setText("Select a vertex");
    }

    //_view_triangle_orthocircle = viewOthocircleAct->isChecked();
    if (viewOthocircleAct->isChecked()) {
      m_statusMiddle->setText("View Triangle Orthocircle (circumcircle)");
      m_statusRight->setText("Select a triangle");
    }

    //_view_triangle_orthocenter = viewOthocenterAct->isChecked();
    if (viewOthocenterAct->isChecked()) {
      m_statusMiddle->setText("View Triangle Orthocenter (circumcenter)");
      m_statusRight->setText("Select a triangle");
    }

    // clean selected vertex/segment
    if (_active_vertex) {
      set_active_vertex(NULL);
    }
    if (_active_segment) {
      set_active_segment(NULL);
    }
    if (_active_trilist) {
      set_active_trilist(NULL);
    }
    if (_active_sdm_vertex) {
      _active_sdm_vertex = NULL;
    }
    update();
}

//=============================================================================

void ScribbleArea::edit_attributes_act()
{
    qDebug()<<"space key is pressed";
    if (_active_vertex && (_active_vertex->Next == NULL)) {
      // Only one active vertex is selected.
      if (_vertexPanel == NULL) {
        _vertexPanel = new vertexPanel(this);
      }
      _vertexPanel->_idx->setText(QString("index = ") + QString::number(_active_vertex->idx));
      _vertexPanel->_x->setText(QString::number(_active_vertex->crd[0]));
      _vertexPanel->_y->setText(QString::number(_active_vertex->crd[1]));
      _vertexPanel->_w->setText(QString::number(_active_vertex->wei));
      _vertexPanel->_val->setText(QString::number(_active_vertex->val));
      char strtyp[64];
      detri2_get_vertex_typ(_active_vertex, strtyp);
      _vertexPanel->_typ->setText(QString(strtyp));
      _vertexPanel->show();
    } else if (_active_segment) {
      if (_segmentPanel == NULL) {
        _segmentPanel = new segmentPanel(this);
      }
      Vertex *e1 = _active_segment->vrt[0];
      Vertex *e2 = _active_segment->vrt[1];
      _segmentPanel->_ends->setText(QString("segment: ") +
           QString::number(e1->idx) + QString(",") + QString::number(e2->idx));
      _segmentPanel->_tag->setText(QString::number(_active_segment->tag));
      _segmentPanel->_maxlen->setText(QString::number(_active_segment->val));
      _segmentPanel->show();
    } else if (_active_trilist) {
      if (_subdomainPanel == NULL) {
        _subdomainPanel = new subdomainPanel(this);
      }
      // Initialise the values.
      _subdomainPanel->_holeflag->setChecked(false);
      _subdomainPanel->_tag->setText(QString("-1"));
      _subdomainPanel->_maxarea->setText(QString("0"));
      // determine which region from _geom->in_sdms;      
      _active_sdm_vertex = NULL;
      if (detri2_get_subdomain(_geom, _active_trilist, &_active_sdm_vertex)) {
        if (_active_sdm_vertex->tag == 0) {
          _subdomainPanel->_holeflag->setChecked(true);
        }
        _subdomainPanel->_tag->setText(QString::number(_active_sdm_vertex->tag));
        _subdomainPanel->_maxarea->setText(QString::number(_active_sdm_vertex->val));
      }
      _subdomainPanel->show();
    }
}

//=============================================================================

void ScribbleArea::edit_delete_act()
{
    //if (_geom->check_mesh(0,0)) {
    //  qDebug()<<"!!!!!! Error before removing a vertex";
    //  return;
    //}

    if (_active_vertex) {
      // only remove it if it is a free vertex.
      arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
      int count = 0;
      Vertex *vrt = _active_vertex, *next_vrt;
      while (vrt != NULL) {
        next_vrt = vrt->Next; // The next vertex.
        if (!vrt->is_fixed()) { // Do not remove a segment vertex.
          //if (_active_vertex->Pair != NULL) {
          //  if (_tria != NULL) {
          //    delete _tria;
          //    _tria = NULL;
          //    _object_valid=0; // need redraw
          //  }
          //}
          _geom->remove_point(vrt, fqueue);
          _geom->lawson_flip(NULL, 0, fqueue); // hullflag==0
          count++;
        } else {
          QMessageBox msgBox;
          msgBox.setText("Can't remove a corner vertex.");
          msgBox.setStandardButtons(QMessageBox::Close);
          msgBox.exec();
        }
        vrt = next_vrt;
      }
      _active_vertex = NULL;
      delete fqueue;
      QString ss = "Removed "+QString::number(count)+" vertices";
      m_statusRight->setText(ss);
      if (count > 0) {
        _object_valid=0; // redraw the scene
        update();
      }
    } else if (_active_segment) {
      // Only delete this segment if it does not belong to any subdomain.
      int count = 0;
      Triang *delseg = _active_segment, *next_seg;
      while (delseg != NULL) {
        next_seg = delseg->Next; // bakup
        Vertex *e1 = delseg->vrt[0];
        Vertex *e2 = delseg->vrt[1];
        //if ((e1->Pair != NULL) || (e2->Pair != NULL)) {
        //  if (_tria != NULL) {
        //    delete _tria;
        //    _tria = NULL;
        //    _object_valid=0; // need redraw
        //  }
        //}
        // Only clear the segment. Do not re-draw the scene.
        //set_active_segment(NULL);
        // Remove the segment from _geom.
        //if (!detri2_remove_segment(delseg)) {
        //  assert(0); // We assume _geom is a CDT. This should not happen.
        //}
        // Assume the triangulation is a CDT, no Steiner point in segment.
        // We must find exactly a segment with the endpoints [e1,e2]
        /*  replace by the code below. get_edge() and remove_segment()
        TriEdge E = e1->adj;
        do {
          assert(E.org() == e1);
          if (E.dest() == e2) break;
          E = E.eprev_esym();
        } while (E.tri != e1->adj.tri);
        if (E.dest() == e2) {
          // Found the segment.
          E.clear_segment();
          TriEdge N = E.esym();
          N.clear_segment();
          // remove this segment and update the vertex-to-segment map.
          detri2_remove_segment_at_vertex(e1, delseg);
          detri2_remove_segment_at_vertex(e2, delseg);
          delseg->set_deleted();
          _geom->tr_segs->dealloc(delseg);
          _geom->ct_segments--;
          // For debug only
          //if (_geom->check_mesh(0,0)) {
          //  qDebug()<<"!! Error after removing a segment";
          //}
          count++;
        }
        */
        TriEdge E;
        if (_geom->get_edge(e1, e2, E)) {
          assert(E.is_segment());
          Triang *seg = E.get_segment();
          _geom->remove_segment(seg);
          assert(!E.is_segment());
          count++;
        }
        delseg = next_seg;
      }
      QString ss = "Removed "+QString::number(count)+"segments";
      m_statusRight->setText(ss);
      _active_segment = NULL;
      if (count > 0) {
        // at least one selected segment is deleted. redraw.
        _object_valid = 0;
        update();
      }
    }
}

//=============================================================================

void ScribbleArea::edit_scale_act()
{
    qDebug()<<"scale act";
    if (_geom->tr_steiners && _geom->tr_steiners->objects > 0) {
      detri2_scale_points(_geom, -1, 1, -1, 1);
      calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
      _object_valid = 0;
      update();
    }
}

//=============================================================================

void ScribbleArea::view_update_act()
{
    //_show_geom_vertices = viewVerticesAct->isChecked();
    //_show_weights = viewWeightsAct->isChecked();
    //_show_vertex_labels = viewVertexLabelsAct->isChecked();
    //_show_geom_edges = viewSegmentsAct->isChecked();
    //_show_geom_subdomains = viewSubdomainsAct->isChecked();
    //_show_segment_labels = viewSegmentLabelsAct->isChecked();
    //_show_geom_bbox = viewBboxAct->isChecked();
    //_show_geom_tris = viewBgMeshAct->isChecked();
    //_show_convex_hull = viewConvexHullAct->isChecked();
    //_show_triangles = viewTrianglesAct->isChecked();
    //_show_ortho_circle = viewOthocirclesAct->isChecked();
    //_show_voronoi = viewPowerDiagramAct->isChecked();

    // Some option changed, re-draw the scene.
    _object_valid = 0;
    update();
}

//=============================================================================

void ScribbleArea::view_trans_act(int i)
{
    qDebug()<<" signal i="<<i;
    if (i == 1) {
      // view_moveleft()
      double dx = -40.0, dy = 0.0;
      _cx += (int) (_mouse_sens * dx);
      _cy += (int) (_mouse_sens * dy);
      _object_valid=0; // invalidate_object();
      update(); //updateGL();
    } else if (i == 2) {
      // view_moveright()
      double dx = 40.0, dy = 0.0;
      _cx += (int) (_mouse_sens * dx);
      _cy += (int) (_mouse_sens * dy);
      _object_valid=0; // invalidate_object();
      update(); //updateGL();
    } else if (i == 3) {
      // view_moveup()
      double dx = 0.0, dy = -40.0;
      _cx += (int) (_mouse_sens * dx);
      _cy += (int) (_mouse_sens * dy);
      _object_valid=0; // invalidate_object();
      update(); //updateGL();
    } else if (i == 4) {
      // view_movedown()
      double dx = 0.0, dy = 40.0;
      _cx += (int) (_mouse_sens * dx);
      _cy += (int) (_mouse_sens * dy);
      _object_valid=0; // invalidate_object();
      update(); //updateGL();
    }
}

//=============================================================================

void ScribbleArea::view_zoom_act(int i)
{
    if (i == 1) {
        double dy = 0.1;
        _scale_xy *= (1.0 + _mouse_sens * dy);
        _object_valid=0;
        update();
    } else if (i == 2) {
        double dy = -0.1;
        _scale_xy *= (1.0 + _mouse_sens * dy);
        _object_valid=0;
        update();
    } /*else if (i == 3) {
        _zoom_selected_region = 1;
    }*/
}

//=============================================================================

void ScribbleArea::view_reset()
{
      reset_scribble();
      calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
      update(); //updateGL();
}

//=============================================================================

void ScribbleArea::detri2_options_act()
{
    if (_meshoptionPanel == NULL) {
      _meshoptionPanel = new meshoptionPanel(this);
    }
    _meshoptionPanel->show();
}

void ScribbleArea::generate_mesh_act()
{
  int furthest = 0;
  int weighted = 1;
  int cdt = 0;

  if ((_tria != NULL) && (_tria->tr_tris != NULL)) {
    // An initial triangulation already exists.
    qDebug()<<"An initial triangulation already exists -- refine or smooth it";
    if (_meshoptionPanel) {
      furthest = (int) _meshoptionPanel->op_FDT->isChecked();
      if (furthest) _tria->op_dt_nearest = -1;
      weighted = (int) _meshoptionPanel->op_weighted->isChecked();

      // Save file options
      _tria->io_dump_to_ucd = (int) _meshoptionPanel->op_dump_to_ucd->isChecked();
      _tria->io_dump_lift_map = (int) _meshoptionPanel->op_dump_lift_map->isChecked();
      if (_tria->io_dump_to_ucd) {
        if (!_meshoptionPanel->line_dump_file_name->text().isEmpty()) {
          qDebug()<<"dump file name = "<<_meshoptionPanel->line_dump_file_name->text();
          QByteArray ba = _meshoptionPanel->line_dump_file_name->text().toLatin1();
          const char *c_str2 = ba.data();
          strcpy(_tria->io_outfilename, c_str2);
        } else {
          // Use a default filename.
          if (_tria->io_outfilename[0] == 0) {
            strcpy(_tria->io_outfilename, "dump");
          }
        }
      }

      if (_meshoptionPanel->op_Convexhull->isChecked()) {
        // Incrementally update convex hull
        _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
        _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
        //_show_triangles = false;
        viewTrianglesAct->setChecked(false);
        //_show_convex_hull = true;
        viewConvexHullAct->setChecked(true);
      }
      else if (_meshoptionPanel->op_Delaunay->isChecked()) {
        // Incrementally update Delaunay triangulation.
        _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
        _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
        TriEdge E, tt[4];
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
        for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
          if (v->is_deleted()) continue;
          //v->print();
          // An laready inserted vertex must be v->Pair != NULL.
          if (v->Pair == NULL) {
            qDebug()<<"Insert a new vertex id="<<v->idx;
            // This is a vertex to be inserted.
            if (_tria->tr_steiners == NULL) {
              qDebug()<<"Create Steiner points pool tr_steiners";
              _tria->tr_steiners = new arraypool(sizeof(Vertex), 8);
            }
            Vertex *newvrt = (Vertex *) _tria->tr_steiners->alloc();
            newvrt->init();
            newvrt->crd[0] = v->crd[0];
            newvrt->crd[1] = v->crd[1];
            newvrt->crd[2] = v->crd[0]*v->crd[0]+v->crd[1]*v->crd[1];
            if (weighted) {
              newvrt->crd[2] -= v->wei;
            }
            newvrt->idx = (_tria->ct_in_vrts + _tria->tr_steiners->objects - 1)
                        + _tria->io_firstindex;
            newvrt->typ = STEINERVERTEX;
            E.tri = NULL;
            qDebug() << "Locate point";
            //int loc = _tria->locate_point(newvrt, E, 0, 0); // encflag = 0
            int loc = _tria->locate_point(newvrt, E, 0); // encflag = 0
            tt[0] = E;
            int hullflag = 0;
            if (loc == LOC_IN_OUTSIDE) {
              qDebug() << "  loc = OUTSIDE";
              // This is only allowed when E is not a hull segment.
              if (!E.is_segment()) {
                hullflag = 1;
                int fflag = FLIP_13;
                _tria->flip(tt, &newvrt, fflag, fqueue);
              } else {
                // Do not insert this vertex.
                qDebug() << "  Point not inserted!!";
                newvrt->set_deleted();
                _tria->tr_steiners->dealloc(newvrt);
                newvrt = NULL;
              }
            } else if (loc == LOC_IN_TRI) {
              qDebug() << "  loc = In triangle";
              int fflag = FLIP_13;
              _tria->flip(tt, &newvrt, fflag, fqueue);
            } else if (loc == LOC_ON_EDGE) {
              qDebug() << "  loc = ON edge";
              int fflag = FLIP_24;
              _tria->flip(tt, &newvrt, fflag, fqueue);
            } else {
              qDebug() << "  Other location loc = "<< loc;
              qDebug() << "  Point not inserted!!";
              newvrt->set_deleted();
              _tria->tr_steiners->dealloc(newvrt);
              newvrt = NULL;
            }
            if (newvrt) {
              // The vertex is inserted
              v->Pair = newvrt;
              _tria->lawson_flip(newvrt, hullflag, fqueue);
            }
          } // if (v->Pair == NULL)
        } // i
        delete fqueue;
      }
      else if (_meshoptionPanel->op_CDT->isChecked()) {
        // Insert new segments.
        if (_tria->tr_segs == NULL) {
          _tria->tr_segs = new arraypool(sizeof(Triang), 8);
        }
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
        for (int i = 0; i < _geom->tr_segs->used_items; i++) {
          Triang *seg = (Triang *) _geom->tr_segs->get(i);
          if (seg->is_deleted()) continue;
          Vertex *e1 = seg->vrt[0]->Pair;
          Vertex *e2 = seg->vrt[1]->Pair;
          int tag = seg->tag;
          REAL val = seg->val;
          /*
          TriEdge E; E.tri = NULL; // Used to return the segment.
          // if (!detri2_insert_segment(_tria, e1, e2, tag, E)) {
          if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
            if (_tria->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
              Triang *newseg = NULL;
              _tria->insert_segment(E, &newseg);
              newseg->tag = tag;
            } else {
              printf("!! Failed to recover segment [%d,%d] tga(%d).\n",
                     e1->idx, e2->idx, tag);
            }
            if (fqueue->objects > 0) {
              _tria->lawson_flip(NULL, 0, fqueue);
            }
          }
          */
          if (!detri2_insert_segment(_tria, e1, e2, tag, val, fqueue)) {
            printf("!! Failed to recover segment [%d,%d] tag(%d).\n", e1->idx, e2->idx, tag);
          }
        }
        delete fqueue;
      }
      else if (_meshoptionPanel->op_Quality->isChecked()) {
        // Do Delaunay refinement
        if ((_tria->ct_exteriors == 0) &&
            (_tria->tr_nonconvex == false)) {
          _tria->set_subdomains();
        }
        if (_tria->ct_exteriors == _tria->tr_tris->objects) {
          QMessageBox msg;
          msg.setText("Mesh domain is not closed!");
          msg.exec();
          return;
        }
        if (_tria->ct_exteriors > 0) {
          _tria->remove_exteriors(); // Remove exterior triangles.
          _tria->tr_nonconvex = true;
        }
        if (_tria->tr_tris->objects > 0) {
          _tria->op_quality = 1;
          _tria->op_minangle = (_meshoptionPanel->line_op_minangle->text()).toDouble();
          _tria->op_maxarea = (_meshoptionPanel->line_op_maxarea->text()).toDouble();
          _tria->op_minlen = (_meshoptionPanel->line_op_minlen->text()).toDouble();
          _tria->op_metric = (int) _meshoptionPanel->op_metric->isChecked();
          if (_tria->op_metric > 0) {
            qDebug()<<"Mesh size enabled";
            _tria->OMT_domain = _geom; // Set the background mesh.
          } else {
            _tria->OMT_domain = NULL;
          }
          //qDebug()<<"Quality mesh generation";
          //_tria->op_db_verbose = 4; // debug
          _tria->delaunay_refinement();
          _tria->mesh_statistics();
          _tria->quality_statistics();
        }
      }
      else if (_meshoptionPanel->op_Flip->isChecked()) {
        // Do Lawson flips.
        int bak = _tria->op_no_incremental_flip;
        _tria->op_no_incremental_flip = 0;
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
        _tria->lawson_flip(NULL, 1, fqueue); // hullflag = 1;
        _tria->op_no_incremental_flip = bak;
        delete fqueue;
      }
      else if (_meshoptionPanel->op_Smooth->isChecked()) {
        // Smooth mesh vertices
        int curidx = _meshoptionPanel->_cb_smooth_methods->currentIndex();
        qDebug()<<"choosed count = "<< curidx;
        int iter = _meshoptionPanel->line_smooth_iters->text().toInt();
        if (iter > 0) {
          //_tria->op_db_verbose = 10; // for debug
          _tria->OMT_domain = _geom;
          for (int i = 0; i < iter; i++) {
            //qDebug()<<"\n iter="<<i<<"\n";
            _tria->smooth_vertices(curidx+1);
          }
          _tria->OMT_domain = NULL;
          qDebug()<<"\n finished after "<<iter<<" iterations\n";
        }
        _tria->mesh_statistics();
        _tria->quality_statistics();
      }

      _object_valid=0;
      update(); //updateGL();
    } // if (_meshoptionPanel)

    return;
  } // if ((_tria != NULL) && (_tria->tr_tris != NULL))

  if (_meshoptionPanel) {
    if (_meshoptionPanel->op_Flip->isChecked() ||
        _meshoptionPanel->op_Smooth->isChecked()) {
      QMessageBox msg;
      msg.setText("Please first generate a mesh!");
      msg.exec();
      return;
    }
  }

  qDebug()<<"====== Generating a new triangulation.===========\n";

  if (_tria != NULL) delete _tria;
  _tria = NULL;

  _tria = new Triangulation();

  if (_meshoptionPanel) {
    furthest = (int) _meshoptionPanel->op_FDT->isChecked();
    if (furthest) _tria->op_dt_nearest = -1;
    weighted = (int) _meshoptionPanel->op_weighted->isChecked();

    if (_meshoptionPanel->op_Convexhull->isChecked()) {
      _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
      _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
      //_show_triangles = false;
      viewTrianglesAct->setChecked(false);
      //_show_convex_hull = true;
      viewConvexHullAct->setChecked(true);
    }

    if (_meshoptionPanel->op_Delaunay->isChecked()) {
      _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
      _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
    }

    cdt = (int) _meshoptionPanel->op_CDT->isChecked();

    if (_meshoptionPanel->op_Quality->isChecked()) {
      _tria->op_quality = 1;
      _tria->op_minangle = (_meshoptionPanel->line_op_minangle->text()).toDouble();
      _tria->op_maxarea = (_meshoptionPanel->line_op_maxarea->text()).toDouble();
      _tria->op_minlen = (_meshoptionPanel->line_op_minlen->text()).toDouble();
      _tria->op_metric = (int) _meshoptionPanel->op_metric->isChecked();
    }

    // Save file options
    _tria->io_dump_to_ucd = (int) _meshoptionPanel->op_dump_to_ucd->isChecked();
    _tria->io_dump_lift_map = (int) _meshoptionPanel->op_dump_lift_map->isChecked();
    if (_tria->io_dump_to_ucd) {
      if (!_meshoptionPanel->line_dump_file_name->text().isEmpty()) {
        qDebug()<<"dump file name = "<<_meshoptionPanel->line_dump_file_name->text();
        QByteArray ba = _meshoptionPanel->line_dump_file_name->text().toLatin1();
        const char *c_str2 = ba.data();
        strcpy(_tria->io_outfilename, c_str2);
      } else {
        // Use a default filename.
        if (_tria->io_outfilename[0] == 0) {
          strcpy(_tria->io_outfilename, "dump");
        }
      }
    }
  }

  if (_tria->op_metric > 0) {
     qDebug()<<"Mesh size enabled";
    _tria->OMT_domain = _geom; // Set the background mesh.
    _tria->op_metric = METRIC_ISO; //  = 1
  } else {
    _tria->OMT_domain = NULL;
    _tria->op_metric = METRIC_EUCLIDEAN; // = 0
  }

  //_tria->op_db_verbose = 4; // Debug

  detri2_generate(_geom, &_tria, weighted, cdt);

  _object_valid=0;
  update(); //updateGL();
}

void ScribbleArea::generate_mesh_ani_act()
{
  int furthest = 0;
  int weighted = 1;
  int cdt = 0;

  if ((_tria != NULL) && (_tria->tr_tris != NULL)) {
    // An initial triangulation already exists.
    if (_meshoptionPanel) {
      furthest = (int) _meshoptionPanel->op_FDT->isChecked();
      if (furthest) _tria->op_dt_nearest = -1;
      weighted = (int) _meshoptionPanel->op_weighted->isChecked();

      // Save file options
      _tria->io_dump_to_ucd = 0;
      _tria->io_dump_lift_map = 0;
      /*
      _tria->io_dump_to_ucd = (int) _meshoptionPanel->op_dump_to_ucd->isChecked();
      _tria->io_dump_lift_map = (int) _meshoptionPanel->op_dump_lift_map->isChecked();
      if (_tria->io_dump_to_ucd) {
        if (!_meshoptionPanel->line_dump_file_name->text().isEmpty()) {
          qDebug()<<"dump file name = "<<_meshoptionPanel->line_dump_file_name->text();
          QByteArray ba = _meshoptionPanel->line_dump_file_name->text().toLatin1();
          const char *c_str2 = ba.data();
          strcpy(_tria->io_outfilename, c_str2);
        } else {
          // Use a default filename.
          if (_tria->io_outfilename[0] == 0) {
            strcpy(_tria->io_outfilename, "dump");
          }
        }
      }
      */
      // These two values are cleaned every time after using.
      assert(_tria->vw_win == NULL);
      assert(_tria->vw_imgs == NULL);

      if (_meshoptionPanel->op_Convexhull->isChecked()) {
        // Incrementally update convex hull
        _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
        _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
        //_show_triangles = false;
        viewTrianglesAct->setChecked(false);
        //_show_convex_hull = true;
        viewConvexHullAct->setChecked(true);
        _tria->op_no_incremental_flip = true; // default for convex hull construction.
      }
      else if (_meshoptionPanel->op_Delaunay->isChecked()) {
        // Incrementally update Delaunay triangulation.
        _tria->vw_win = (void *) this;
        _tria->vw_imgs = new arraypool(sizeof(QImage *), 10);
        _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
        _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
        TriEdge E, tt[4];
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
        for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
          if (v->is_deleted()) continue;
          //v->print();
          // An laready inserted vertex must be v->Pair != NULL.
          if (v->Pair == NULL) {
            qDebug()<<"Insert a new vertex id="<<v->idx;
            // This is a vertex to be inserted.
            if (_tria->tr_steiners == NULL) {
              qDebug()<<"Create Steiner points pool tr_steiners";
              _tria->tr_steiners = new arraypool(sizeof(Vertex), 8);
            }
            Vertex *newvrt = (Vertex *) _tria->tr_steiners->alloc();
            newvrt->init();
            newvrt->crd[0] = v->crd[0];
            newvrt->crd[1] = v->crd[1];
            newvrt->crd[2] = v->crd[0]*v->crd[0]+v->crd[1]*v->crd[1];
            if (weighted) {
              newvrt->crd[2] -= v->wei;
            }
            newvrt->idx = (_tria->ct_in_vrts + _tria->tr_steiners->objects - 1)
                        + _tria->io_firstindex;
            newvrt->typ = STEINERVERTEX;
            E.tri = NULL;
            qDebug() << "Locate point";
            //int loc = _tria->locate_point(newvrt, E, 0, 0); // encflag = 0
            int loc = _tria->locate_point(newvrt, E, 0); // encflag = 0
            tt[0] = E;
            int hullflag = 0;
            if (loc == LOC_IN_OUTSIDE) {
              qDebug() << "  loc = OUTSIDE";
              // This is only allowed when E is not a hull segment.
              if (!E.is_segment()) {
                hullflag = 1;
                int fflag = FLIP_13;
                _tria->flip(tt, &newvrt, fflag, fqueue);
              } else {
                // Do not insert this vertex.
                qDebug() << "  Point not inserted!!";
                newvrt->set_deleted();
                _tria->tr_steiners->dealloc(newvrt);
                newvrt = NULL;
              }
            } else if (loc == LOC_IN_TRI) {
              qDebug() << "  loc = In triangle";
              int fflag = FLIP_13;
              _tria->flip(tt, &newvrt, fflag, fqueue);
            } else if (loc == LOC_ON_EDGE) {
              qDebug() << "  loc = ON edge";
              int fflag = FLIP_24;
              _tria->flip(tt, &newvrt, fflag, fqueue);
            } else {
              qDebug() << "  Other location loc = "<< loc;
              qDebug() << "  Point not inserted!!";
              newvrt->set_deleted();
              _tria->tr_steiners->dealloc(newvrt);
              newvrt = NULL;
            }
            if (newvrt) {
              // The vertex is inserted
              v->Pair = newvrt;
              _tria->lawson_flip(newvrt, hullflag, fqueue);
            }
          } // if (v->Pair == NULL)
        } // i
        delete fqueue;
      }
      else if (_meshoptionPanel->op_CDT->isChecked()) {
        // Insert new segments.
        _tria->vw_win = (void *) this;
        _tria->vw_imgs = new arraypool(sizeof(QImage *), 10);
        if (_tria->tr_segs == NULL) {
          _tria->tr_segs = new arraypool(sizeof(Triang), 8);
        }
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 4);
        for (int i = 0; i < _geom->tr_segs->used_items; i++) {
          Triang *seg = (Triang *) _geom->tr_segs->get(i);
          if (seg->is_deleted()) continue;
          Vertex *e1 = seg->vrt[0]->Pair;
          Vertex *e2 = seg->vrt[1]->Pair;
          int tag = seg->tag;
          REAL val = seg->val;
          /*
          TriEdge E; E.tri = NULL; // Used to return the segment.
          //if (!detri2_insert_segment(_tria, e1, e2, tag, E)) {
          if ((e1 != NULL) && (e2 != NULL) && (e1 != e2)) {
            if (_tria->recover_edge(e1, e2, E, fqueue) == INTERSECT_SHARE_EDGE) {
              if (!E.is_segment()) {
                _tria->insert_segment(E, tag, NULL);
              }
            } else {
              printf("!! Failed to recover segment [%d,%d] tga(%d).\n",
                     e1->idx, e2->idx, tag);
            }
            if (fqueue->objects > 0) {
              _tria->lawson_flip(NULL, 0, fqueue);
            }
          }
          */
          if (!detri2_insert_segment(_tria, e1, e2, tag, val, fqueue)) {
            //printf("!! Failed to insert segment [%d,%d] tag(%d).\n", e1->idx, e2->idx, tag);
            // It is possible that this segment already been inserted.
          }
        }
        delete fqueue;
      }
      else if (_meshoptionPanel->op_Quality->isChecked()) {
        if ((_tria->ct_exteriors == 0) &&
            (_tria->tr_nonconvex == false)) {
          _tria->set_subdomains();
        }
        if (_tria->ct_exteriors == _tria->tr_tris->objects) {
          QMessageBox msg;
          msg.setText("Mesh domain is not closed!");
          msg.exec();
          return;
        }
        if (_tria->ct_exteriors > 0) {
          _tria->remove_exteriors(); // Remove exterior triangles.
          _tria->tr_nonconvex = true;
        }
        if (_tria->tr_tris->objects > 0) {
          _tria->vw_win = (void *) this;
          _tria->vw_imgs = new arraypool(sizeof(QImage *), 10);
          _tria->op_quality = 1;
          _tria->op_minangle = (_meshoptionPanel->line_op_minangle->text()).toDouble();
          _tria->op_maxarea = (_meshoptionPanel->line_op_maxarea->text()).toDouble();
          _tria->op_minlen = (_meshoptionPanel->line_op_minlen->text()).toDouble();
          _tria->op_metric = (int) _meshoptionPanel->op_metric->isChecked();
          if (_tria->op_metric > 0) {
            qDebug()<<"Mesh size enabled";
            _tria->OMT_domain = _geom; // Set the background mesh.
          } else {
            _tria->OMT_domain = NULL;
          }
          _tria->delaunay_refinement();
          _tria->mesh_statistics();
          _tria->quality_statistics();
        }
      }
      else if (_meshoptionPanel->op_Flip->isChecked()) {
        // Use the Delaunay criteria.
        _tria->vw_win = (void *) this;
        _tria->vw_imgs = new arraypool(sizeof(QImage *), 10);
        int bak = _tria->op_no_incremental_flip;
        _tria->op_no_incremental_flip = 0;
        arraypool *fqueue = new arraypool(sizeof(TriEdge), 8);
        _tria->lawson_flip(NULL, 1, fqueue); // hullflag = 1;
        _tria->op_no_incremental_flip = bak;
        delete fqueue;
      } // op_Flip

      //_object_valid=0;
      //update(); //updateGL();
      if (_imgView == NULL) {
        _imgView = new ImageView();
      }
      _imgView->loadImages(_tria->vw_imgs);

      // Release all images (allocated in flip.cpp)
      for (int i = 0; i < _tria->vw_imgs->objects; i++) {
        QImage **pimg = (QImage **) _tria->vw_imgs->get(i);
        delete *(pimg);
      }
      delete _tria->vw_imgs;
      _tria->vw_imgs=NULL;
      _tria->vw_win=NULL;

      _imgView->show();
    } // if (_meshoptionPanel)

    return;
  } // if ((_tria != NULL) && (_tria->tr_tris != NULL))

  if (_tria != NULL) {
    delete _tria;
  }
  _tria = new Triangulation();

  if (_meshoptionPanel) {
    furthest = (int) _meshoptionPanel->op_FDT->isChecked();
    if (furthest) _tria->op_dt_nearest = -1;
    weighted = (int) _meshoptionPanel->op_weighted->isChecked();

    if (_meshoptionPanel->op_Convexhull->isChecked()) {
      _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
      _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
      viewTrianglesAct->setChecked(false); //_show_triangles = false;
      viewConvexHullAct->setChecked(true); //_show_convex_hull = true;
      _tria->op_no_incremental_flip = true; // default for convex hull construciton
    }

    if (_meshoptionPanel->op_Delaunay->isChecked()) {
      _tria->so_nosort = (int) _meshoptionPanel->op_no_sort->isChecked();
      _tria->op_no_incremental_flip = (int) _meshoptionPanel->op_no_flip->isChecked();
    }

    cdt = (int) _meshoptionPanel->op_CDT->isChecked();

    if (_meshoptionPanel->op_Quality->isChecked()) {
      _tria->op_quality = 1;
      _tria->op_minangle = (_meshoptionPanel->line_op_minangle->text()).toDouble();
      _tria->op_maxarea = (_meshoptionPanel->line_op_maxarea->text()).toDouble();
      _tria->op_minlen = (_meshoptionPanel->line_op_minlen->text()).toDouble();
      _tria->op_metric = (int) _meshoptionPanel->op_metric->isChecked();
    }
  }

  if (_tria->op_metric > 0) {
    _tria->OMT_domain = _geom; // Set the background mesh.
  } else {
    _tria->OMT_domain = NULL;
  }

  _tria->vw_win = (void *) this;
  _tria->vw_imgs = new arraypool(sizeof(QImage *), 10);

  //_show_geom_bbox = false; // Do not show background mesh
  viewBboxAct->setChecked(false);

  detri2_generate(_geom, &_tria, weighted, cdt);

  //_active_vertex = NULL;
  //_object_valid=0;
  //update(); //updateGL();

  if (_imgView == NULL) {
    _imgView = new ImageView();
  }
  _imgView->loadImages(_tria->vw_imgs);

  // Release all images (allocated in flip.cpp)
  for (int i = 0; i < _tria->vw_imgs->objects; i++) {
    QImage **pimg = (QImage **) _tria->vw_imgs->get(i);
    delete *(pimg);
  }
  delete _tria->vw_imgs;
  _tria->vw_imgs=NULL;
  _tria->vw_win=NULL;

  _imgView->show();
}

void ScribbleArea::weighted_Delaunay_act()
{
    if (_geom->tr_steiners->objects >= 3) {
      // Generate triangulation (mesh) from _geom
      int weighted = 1;
      int cdt = (_geom->tr_segs && (_geom->tr_segs->objects > 0));
      detri2_Delaunay(_geom, &_tria, 0, weighted, cdt);
      _object_valid=0;
      update(); //updateGL();
    }
}

void ScribbleArea::furthest_Delaunay_act()
{
    if (_geom->tr_steiners->objects >= 3) {
      int weighted = 1;
      int cdt = 0;
      if (_meshoptionPanel) {
        weighted = (int) _meshoptionPanel->op_weighted->isChecked();
      }
      detri2_Delaunay(_geom, &_tria, 1, weighted, cdt);
      _object_valid=0;
      update(); //updateGL();
    }
}

void ScribbleArea::detri2_HDE_option_act()
{
    if (_hdeoptionPanel == NULL) {
      _hdeoptionPanel = new hdeoptionPanel(this, NULL);
    }
    _hdeoptionPanel->show();
}

void ScribbleArea::detri2_HDE_act()
{
    if ((_tria != NULL) && (_tria->tr_tris != NULL) &&
        (_tria->tr_tris->objects > 0)) {
      qDebug()<<"get parameters for HDE mesh adaptation";
      double s1 = _hdeoptionPanel->hde_s1->text().toDouble();
      double s2 = _hdeoptionPanel->hde_s2->text().toDouble();
      double min_len = _hdeoptionPanel->hde_min_len->text().toDouble();
      double min_angle = _hdeoptionPanel->hde_min_angle->text().toDouble();
      int size_iter = _hdeoptionPanel->hde_max_size_iter->text().toInt();
      int qual_iter = _hdeoptionPanel->hde_max_quality_iter->text().toInt();
      int use_intpoints = _hdeoptionPanel->hde_use_intpoints->text().toInt();
      int use_smooth = (int) _hdeoptionPanel->hde_use_smooth->isChecked();
      int smooth_iter = _hdeoptionPanel->hde_smooth_iter->text().toInt();
      double smooth_alpha = _hdeoptionPanel->hde_smooth_alpha->text().toDouble();
      //int verbose = _hdeoptionPanel->line_op_verbose->text().toInt();
      int report_error = _hdeoptionPanel->hde_report_error_only->isChecked();

      int test_fun = _hdeoptionPanel->hde_testfun_ComboBox->currentIndex();
      if (test_fun == -1) {
        qDebug()<<"No test function is selected";
        return;
      }
      qDebug()<<"done!";

      detri2_HDE_mesh_adapt(&_tria,
                            s1,
                            s2,
                            min_len,
                            min_angle,
                            size_iter,
                            qual_iter,
                            use_intpoints,
                            use_smooth,
                            smooth_iter,
                            smooth_alpha,
                            report_error, // verbose
                            test_fun+1);
      _object_valid=0;
      update(); //updateGL();
    } else {
      QMessageBox msg;
      msg.setText("No initial mesh!");
      msg.exec();
    }
}

void ScribbleArea::detri2_Sort_option_act()
{
    if (_sortoptionpanel == NULL) {
      _sortoptionpanel = new sortOptionPanel(this, NULL);
    }
    _sortoptionpanel->show();
}

void ScribbleArea::detri2_Sort_act()
{
    if (_sortoptionpanel == NULL) {
      return; // no sort option is available.
    }

    if (_sortoptionpanel->op_draw_Hilbert_curve->isChecked()) {
      // Show nth-order Hilbert curve
      qDebug()<<"Draw Hilbert curve";
      int order = _sortoptionpanel->hilbert_order->text().toInt();
      int ei = _sortoptionpanel->hilbert_e->text().toInt();
      int fi = _sortoptionpanel->hilbert_f->text().toInt();
      if (ei == fi) {
        QMessageBox msg;
        msg.setText(tr("Entry and exit must not be the same"));
        msg.exec();
        return;
      }

      // Get the grey code of e = ei^(ei>>1), and f = fi^(fi>>1)
      //   (see page 7, Theorem 2.1, Hamilton2006)
      int e = ei ^ (ei >> 1);
      int f = fi ^ (fi >> 1);
      // Calculate the intra sub-cube direction (see page 11, Hamilton06).
      int d = (e^f) >> 1;
      printf("  e = %d, f = %d, d = %d\n", e, f, d);

      int hidex; int *phindex;
      if (_sortoptionpanel->show_Hilbert_index->isChecked()) {
        hidex = 0; phindex = &hidex;
      } else {
        phindex = NULL;
      }

      double bxmin = _geom->io_xmin;
      double bxmax = _geom->io_xmax;
      double bymin = _geom->io_ymin;
      double bymax = _geom->io_ymax;
      if (_geom->tr_steiners->objects > 0) {
        // Use the bbox of the point set.
        for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
          if (v->is_deleted()) continue;
          bxmin = (bxmin < v->crd[0] ? bxmin : v->crd[0]);
          bxmax = (bxmax > v->crd[0] ? bxmax : v->crd[0]);
          bymin = (bymin < v->crd[1] ? bymin : v->crd[1]);
          bymax = (bymax > v->crd[1] ? bymax : v->crd[1]);
        }
      }

      QPainter painter(&_canvas[0]);
      QPen pen;
      pen.setColor(qRgb(255, 204, 0));
      pen.setWidth(2);
      painter.setPen(pen);
      void *Painter = (void *) &painter;

      detri2_draw_hilbert_init(2);
      detri2_draw_Hilbert_curve(bxmin, bxmax, bymin, bymax,
                                e,d,order,0, // depth = 0
                                NULL, NULL, phindex,
                                Painter, _cx, _cy, _scale_xy);
      //_object_valid=0;
      update(); //updateGL();
      return;
    }

    if (_geom->tr_steiners->objects == 0) {
      QMessageBox msg;
      msg.setText(tr("The vertex set is empty!"));
      msg.exec();
      return;
    }

    // Sort vertices accroding to selected method.
    // Get the set of vertices.
    int ptnum = _geom->tr_steiners->objects;
    Vertex **vlist = new Vertex*[ptnum];
    int idx = 0;
    for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      vlist[idx] = v;
      idx++;
    }
    assert(idx == ptnum);

    if (_sortoptionpanel->op_random_sort->isChecked()) {

    } else if (_sortoptionpanel->op_sweepline_sort->isChecked()) {
      // get the normal of the sweepline.
      double vx = _sortoptionpanel->sweep_x->text().toDouble();
      double vy = _sortoptionpanel->sweep_y->text().toDouble();
      qDebug()<<"Sweepline sort, normal ("<<vx<<","<<vy<<")";
      detri2_sweepline_sort(vlist, ptnum, vx, vy);
    } else if (_sortoptionpanel->op_Hilbert_sort->isChecked()) {

    } else if (_sortoptionpanel->op_BRIO_Hilbert_sort->isChecked()) {

    }

    // Prepare a list of points.
    double *ptlist = new double[ptnum * 2];
    for (int i = 0; i < ptnum; i++) {
      double *pt = &(ptlist[i*2]);
      pt[0] = vlist[i]->crd[0];
      pt[1] = vlist[i]->crd[1];
    }
    delete [] vlist;

    // Clean the current geometry and mesh.
    delete _geom;
    _geom = NULL;

    // Using the sorted points re-create the geometry.
    detri2_add_geometry(&_geom, ptnum, ptlist, 0, NULL, NULL);

    delete [] ptlist;

    // Re-draw
    _object_valid = 0;
    update();
}

void ScribbleArea::detri2_mesh_smooth_act()
{
    /*
    if ((_tria != NULL) && (_tria->tr_tris != NULL) &&
        (_tria->tr_tris->objects > 0)) {
      double tf = 10;
      double abstol = 1e-6;
      double dt = 1e-6;
      detri2_mesh_smooth_mmpde(&_tria, tf, abstol, dt);
      _object_valid=0;
      update(); //updateGL();
    } else {
      QMessageBox msg;
      msg.setText("No initial mesh!");
      msg.exec();
    }
    */
}

void ScribbleArea::clean_mesh_act()
{
  if (_tria != NULL) {
    m_statusRight->setText("Clear triangulation");
    delete _tria;
    _tria = NULL;
    // Clear all vertex->Pair pointers in _geom
    for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
      Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
      if (v->is_deleted()) continue;
      v->Pair = NULL;
    }
    _object_valid=0;
    update(); //updateGL();
  }
}

//=============================================================================

void ScribbleArea::newFile()
{
    if (_newfilepabel != NULL) {
      delete _newfilepabel;
    }
    _newfilepabel = new newfilePanel(this, 0);
    _newfilepabel->show();
}

void ScribbleArea::addFile()
{
    if (_newfilepabel != NULL) {
      delete _newfilepabel;
    }
    _newfilepabel = new newfilePanel(this, 1);
    _newfilepabel->show();
}

void ScribbleArea::openFile()
{
    // Read input from file(s).
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
        _file_last_used_path,
        tr("Geometry & meshes (*.node *.poly *.smesh *.mesh *.ele *.uv.m)"));

    QByteArray ba = fileName.toLatin1();
    const char *c_str2 = ba.data();
    if (c_str2!=NULL) {
      _file_last_used_path = QFileInfo(fileName).path(); // store path for next time

      // Reset everything
      clean_scribble();
      reset_scribble();

      if (!detri2_reconstruct(&_geom, &_tria, c_str2)) {
        // Failed to open file, create a new geometry.
        detri2_init_with_bbox(&_geom, -1., 1., -1., 1.);
      }
      calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
      update(); //updateGL();
    }
}

void ScribbleArea::saveFile()
{
    QFileDialog::Options options; // = QFlag(fileDialogOptionsWidget->value());
    QString selectedFilter;

    QString fileName = QFileDialog::getSaveFileName(this,
                                          tr("Save File As"),
                                          _file_last_used_path,
                                          tr("*.node (save points);;\
                                             *.ele (save mesh);;\
                                             *.poly (save geometry);;\
                                             *.smesh (save mesh in smesh format);;\
                                             *.mesh (save mesh in INRIA, FreeFem++ format);;\
                                             *.inp (save mesh in UCD format, for Paraview);;\
                                             *.weight (save vertex weights);;\
                                             *.mtr (save vertex metrics);;\
                                             *.png (save screen in PNG format)"),
                                             &selectedFilter,
                                             options);

    qDebug()<<"selectedFilter: "<<selectedFilter;

    if (!fileName.isEmpty()) {
      _file_last_used_path = QFileInfo(fileName).path(); // store path for next time

      QByteArray ba = fileName.toLatin1();
      const char *c_str2 = ba.data();
      qDebug()<< "save filename = "<< c_str2;
      char filename[256];
      strcpy(filename, c_str2);

      // First try to recongize the file format from the .suffix
      if (!strcmp(&filename[strlen(filename) - 5], ".node")) {
        if (_geom->tr_steiners->objects > 0) {
          filename[strlen(filename) - 5] = '\0';
          strcpy(_geom->io_outfilename, filename);
          _geom->save_nodes(1); // only save Steiner points.
        }
      } else if (!strcmp(&filename[strlen(filename) - 4], ".ele")) {
        if (_tria && (_tria->tr_tris->objects > 0)) {
          filename[strlen(filename) - 4] = '\0';
          strcpy(_tria->io_outfilename, filename);
          //if (_tria->ct_exteriors > 0) {
          //  _tria->remove_exteriors();
          //}
          _tria->save_triangulation();
          if (_tria->tr_segs) {
            _tria->save_edges(); // save segment edges.
          }
        }
      } else if (!strcmp(&filename[strlen(filename) - 5], ".poly")) {
        if (_geom->tr_steiners->objects > 0) {
          filename[strlen(filename) - 5] = '\0';
          strcpy(_geom->io_outfilename, filename);
          _geom->save_poly(1); // only save Steiner points.
        }
      } else if (!strcmp(&filename[strlen(filename) - 6], ".smesh")) {
        if (_tria && (_tria->tr_tris->objects > 0)) {
          filename[strlen(filename) - 6] = '\0';
          strcpy(_tria->io_outfilename, filename);
          //if (_tria->ct_exteriors > 0) {
          //  _tria->remove_exteriors();
          //}
          _tria->save_smesh();
        }
      } else if (!strcmp(&filename[strlen(filename) - 5], ".mesh")) {
        if (_tria && (_tria->tr_tris->objects > 0)) {
          filename[strlen(filename) - 5] = '\0';
          strcpy(_tria->io_outfilename, filename);
          //if (_tria->ct_exteriors > 0) {
          //  _tria->remove_exteriors();
          //}
          _tria->save_inria_mesh();
        }
      } else if (!strcmp(&filename[strlen(filename) - 4], ".inp")) {
        if (_tria && (_tria->tr_tris->objects > 0)) {
          filename[strlen(filename) - 4] = '\0';
          strcpy(_tria->io_outfilename, filename);
          //if (_tria->ct_exteriors > 0) {
          //  _tria->remove_exteriors();
          //}
          _tria->save_to_ucd(0, 1); // meshidx = 0, save_val=1
        }
      } else if (!strcmp(&filename[strlen(filename) - 7], ".weight")) {
        if (_geom->tr_steiners->objects > 0) {
          filename[strlen(filename) - 7] = '\0';
          strcpy(_geom->io_outfilename, filename);
          _geom->save_weights(1); // only save Steiner points.
        }
      } else if (!strcmp(&filename[strlen(filename) - 4], ".mtr")) {
        if (_geom->tr_steiners->objects > 0) {
          filename[strlen(filename) - 4] = '\0';
          strcpy(_geom->io_outfilename, filename);
          _geom->save_metric(1); // only save Steiner points.
        }
      } else if (!strcmp(&filename[strlen(filename) - 4], ".png")) {
        _canvas[0].save(fileName, 0, 50);
      } else {
        //QMessageBox msg;
        //msg.setText("Unknown format");
          // Use the selected file format from the provided list.
          if (selectedFilter.contains("*.node", Qt::CaseInsensitive)) {
            if (_geom->tr_steiners->objects > 0) {
              //if (!strcmp(&filename[strlen(filename) - 5], ".node")) {
              //  filename[strlen(filename) - 5] = '\0';
              //}
              strcpy(_geom->io_outfilename, filename);
              _geom->save_nodes(1); // only save Steiner points.
            }
          } else if (selectedFilter.contains("*.ele", Qt::CaseInsensitive)) {
            if (_tria && (_tria->tr_tris->objects > 0)) {
              //if (!strcmp(&filename[strlen(filename) - 4], ".ele")) {
              //  filename[strlen(filename) - 4] = '\0';
              //}
              strcpy(_tria->io_outfilename, filename);
              //if (_tria->ct_exteriors > 0) {
              //  _tria->remove_exteriors();
              //}
              _tria->save_triangulation();
              if (_tria->tr_segs) {
                _tria->save_edges(); // save segment edges.
              }
            }
          } else if (selectedFilter.contains("*.poly", Qt::CaseInsensitive)) {
            if (_geom->tr_steiners->objects > 0) {
              //if (!strcmp(&filename[strlen(filename) - 5], ".poly")) {
              //  filename[strlen(filename) - 5] = '\0';
              //}
              strcpy(_geom->io_outfilename, filename);
              _geom->save_poly(1); // only save Steiner points.
            }
          } else if (selectedFilter.contains("*.smesh", Qt::CaseInsensitive)) {
            if (_tria && (_tria->tr_tris->objects > 0)) {
              //if (!strcmp(&filename[strlen(filename) - 6], ".smesh")) {
              //  filename[strlen(filename) - 6] = '\0';
              //}
              strcpy(_tria->io_outfilename, filename);
              //if (_tria->ct_exteriors > 0) {
              //  _tria->remove_exteriors();
              //}
              _tria->save_smesh();
            }
          } else if (selectedFilter.contains("*.mesh", Qt::CaseInsensitive)) {
            if (_tria && (_tria->tr_tris->objects > 0)) {
              //if (!strcmp(&filename[strlen(filename) - 5], ".mesh")) {
              //  filename[strlen(filename) - 5] = '\0';
              //}
              strcpy(_tria->io_outfilename, filename);
              //if (_tria->ct_exteriors > 0) {
              //  _tria->remove_exteriors();
              //}
              _tria->save_inria_mesh();
            }
          } else if (selectedFilter.contains("*.inp", Qt::CaseInsensitive)) {
            if (_tria && (_tria->tr_tris->objects > 0)) {
              //if (!strcmp(&filename[strlen(filename) - 4], ".inp")) {
              //  filename[strlen(filename) - 4] = '\0';
              //}
              strcpy(_tria->io_outfilename, filename);
              //if (_tria->ct_exteriors > 0) {
              //  _tria->remove_exteriors();
              //}
              _tria->save_to_ucd(0, 1); // meshidx = 0, save_val=1
            }
          } else if (selectedFilter.contains("*.weight", Qt::CaseInsensitive)) {
            if (_geom->tr_steiners->objects > 0) {
              //if (!strcmp(&filename[strlen(filename) - 7], ".weight")) {
              //  filename[strlen(filename) - 7] = '\0';
              //}
              strcpy(_geom->io_outfilename, filename);
              _geom->save_weights(1); // only save Steiner points.
            }
          } else if (selectedFilter.contains("*.mtr", Qt::CaseInsensitive)) {
            if (_geom->tr_steiners->objects > 0) {
              //if (!strcmp(&filename[strlen(filename) - 4], ".mtr")) {
              //  filename[strlen(filename) - 4] = '\0';
              //}
              strcpy(_geom->io_outfilename, filename);
              _geom->save_metric(1); // only save Steiner points.
            }
          } else if (selectedFilter.contains("*.png", Qt::CaseInsensitive)) {
            //if (!strcmp(&filename[strlen(filename) - 4], ".png")) {
            //  filename[strlen(filename) - 4] = '\0';
            //}
            strcat(filename, ".png");
            qDebug()<<"save png: "<<filename;
            _canvas[0].save(fileName, 0, 50);
          } else {
            QMessageBox msg;
            msg.setText("Unknown format");
          }
      } /////////////////////////////////////////
    }
}

void ScribbleArea::closeFile()
{
    clean_scribble();
    reset_scribble();

    detri2_init_with_bbox(&_geom, -1., 1., -1., 1.);
    calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
    update(); //updateGL();
}

void ScribbleArea::open_preferDialog()
{
    if (_preferdialog == NULL) { // Only open it once.
      _preferdialog = new PreferDialog(this, NULL);
      _preferdialog->setWindowModality(Qt::WindowModal);
    }
    _preferdialog->show();
}

void ScribbleArea::on_user_exit()
{
    qDebug()<<"quit detri2";
    // Do something before closing detri2

    QApplication::quit();
}

//=============================================================================

void ScribbleArea::calc_trans(double xmin, double xmax, double ymin, double ymax)
{   /*
    qDebug() << "calc_trans";
    double w = (double)width() - 20;
    double h = (double)height() - 20;
    //qDebug() << "width="<<w << ", height="<<height()<<endl;
    double dx = xmax - xmin;
    double dy = ymax - ymin;
    double sx =  w / dx;
    double sy =  -h / dy;
    //_cx = w / 2. - (_min[0]+_max[0]) / 2. * sx;
    //_cy = h / 2. - (_min[1]+_max[1]) / 2. * sy;
    _cx = w / 2. - (xmin+xmax) / 2. * sx + 10;
    _cy = h / 2. - (ymin+ymax) / 2. * sy + 10;
    _scale_xy = sx < fabs(sy) ? sx : fabs(sy);
    */

    double w = (double)width()-20;
    double h = (double)height()-20;
    /*
    if (h > w) {
      _scale_xy = h / (ymax - ymin);
    } else {
      _scale_xy = w / (xmax - xmin);
    }
    _cx = -_scale_xy * xmin + 10;
    _cy =  _scale_xy * ymax + 10;
    */

    double dx = xmax - xmin;
    double dy = ymax - ymin;
    double min_hw = h < w ? h : w;
    if (dy < dx) {
      _scale_xy = min_hw / dy;
    } else {
      _scale_xy = min_hw / dx;
    }
    _cx = w / 2. - (xmin + xmax) / 2.0 * _scale_xy + 10;
    _cy = h / 2. + (ymin + ymax) / 2.0 * _scale_xy + 10;

    qDebug()<<"width="<<w<<", height="<<h;
    qDebug()<<"xmin="<<xmin<<", xmax="<<xmax<<", ymin="<<ymin<<", ymax="<<ymax;
    qDebug()<<"Cx="<<_cx<<", Cy="<<_cy<<", Sxy="<<_scale_xy;

    //_cx += _transX;
    //_cy += _transY;
    //_scale_xy *= _zoom;
    //qDebug()<<"Translated: Cx="<<_cx<<", Cy="<<_cy<<", Sxy="<<_scale_xy;
}

void ScribbleArea::get_selection_region(double& xmin, double& xmax, double& ymin, double& ymax)
{
    xmin =  (_select_x - _cx) / _scale_xy;
    ymin = -(_select_y + _select_height - _cy) / _scale_xy;
    xmax =  (_select_x + _select_width - _cx) / _scale_xy;
    ymax = -(_select_y - _cy) / _scale_xy;
}

//=============================================================================
// If thevertex == NULL, only clear the old active vertex
int ScribbleArea::set_active_vertex(Vertex *thevertex)
{
  QPainter painter(&_canvas[0]);
  QPen pen;
  QBrush brush;
  if (_active_vertex != NULL) {
    // First deactive the previous selected vertices.
    //int count = 0;
    Vertex *vrt = _active_vertex, *bak;
    while (vrt != NULL) {      
      //count++;
      // Use the vertex color to re-draw it.
      pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
      pen.setWidth(1); // default
      painter.setPen(pen);
      brush.setColor(_color_vertex); // qRgb(102, 102, 102)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      int px = (int) ( _scale_xy * vrt->crd[0] + _cx);
      int py = (int) (-_scale_xy * vrt->crd[1] + _cy);
      painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
      bak = vrt; // for clearing the ->Next
      vrt = vrt->Next;
      bak->Next = NULL;
    }
    assert(_active_vertex->Next == NULL); // debug
    _active_vertex = NULL;
    painter.setBrush(Qt::NoBrush);
    //QString ss = "Cleared "+ QString::number(count) + " active vertices";
    //m_statusRight->setText(ss);
  }
  if (thevertex != NULL) {
    _active_vertex = thevertex;
    pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
    pen.setWidth(1); // default
    painter.setPen(pen);
    brush.setColor(_color_selected); // qRgb(255, 0, 0)
    brush.setStyle(Qt::SolidPattern);
    painter.setBrush(brush);
    int px = (int) ( _scale_xy * _active_vertex->crd[0] + _cx);
    int py = (int) (-_scale_xy * _active_vertex->crd[1] + _cy);
    painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
    painter.setBrush(Qt::NoBrush);
    assert(thevertex->Next == NULL); // debug
    //QString ss = "Selected an active vertex id="+QString::number(thevertex->idx);
    //m_statusRight->setText(ss);
  }
  //_canvas[1] = _canvas[0];
  return 1;
}

//=============================================================================

int ScribbleArea::is_vertex_active(Vertex *thevertex)
{
  if (_active_vertex != NULL) {
    Vertex *vrt = _active_vertex;
    while (vrt != NULL) {
      if (vrt == thevertex) {
        return 1;
      }
      vrt = vrt->Next;
    }
  }
  return 0;
}

int ScribbleArea::remove_active_vertex(detri2::Vertex *thevertex)
{
    if (_active_vertex != NULL) {
      if (thevertex == _active_vertex) {
        _active_vertex = _active_vertex->Next;
      } else {
        // Search and remove it from the list.
        Vertex *vrt = _active_vertex;
        Vertex *prev_vrt = vrt; // Bakup
        while (vrt != NULL) {
          if (vrt == thevertex) {
            prev_vrt->Next = vrt->Next;
            break;
          }
          prev_vrt = vrt; // Bakup
          vrt = vrt->Next;
        }
      }
      thevertex->Next = NULL;

      QPainter painter(&_canvas[0]);
      QPen pen;
      QBrush brush;
      // Use the vertex color to re-draw it.
      pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
      pen.setWidth(1); // default
      painter.setPen(pen);
      brush.setColor(_color_vertex); // qRgb(102, 102, 102)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      int px = (int) ( _scale_xy * thevertex->crd[0] + _cx);
      int py = (int) (-_scale_xy * thevertex->crd[1] + _cy);
      painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
      painter.setBrush(Qt::NoBrush);
      QString ss = "Removed an active vertex";
      m_statusRight->setText(ss);
    }
    return 1;
}

int ScribbleArea::add_active_vertex(Vertex *thevertex)
{
  QPainter painter(&_canvas[0]);
  QPen pen;
  QBrush brush;

  pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
  pen.setWidth(1); // default
  painter.setPen(pen);
  brush.setColor(_color_selected); // qRgb(255, 0, 0)
  brush.setStyle(Qt::SolidPattern);
  painter.setBrush(brush);
  int px = (int) ( _scale_xy * thevertex->crd[0] + _cx);
  int py = (int) (-_scale_xy * thevertex->crd[1] + _cy);
  painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
  painter.setBrush(Qt::NoBrush);

  int count = 0;
  if (_active_vertex == NULL) {
    _active_vertex = thevertex;
    count++;
  } else {
    Vertex *vrt = _active_vertex;
    count++;
    while (vrt->Next != NULL) {
      count++;
      vrt = vrt->Next;
    }
    vrt->Next = thevertex; count++;
  }
  assert(thevertex->Next == NULL);
  QString ss = "There are "+ QString::number(count) + " active vertices";
  m_statusRight->setText(ss);
  return 1;
}

//=============================================================================

int ScribbleArea::set_active_segment(detri2::Triang *theseg)
{
    QPainter painter(&_canvas[0]);
    QPen pen;
    QBrush brush;
    if (_active_segment != NULL) {
      // First clean the previous selected segment.
      //int count = 0;
      Triang *seg = _active_segment, *bak;
      while (seg != NULL) {
        //count++;
        Vertex *e1 = seg->vrt[0];
        Vertex *e2 = seg->vrt[1];
        pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
        pen.setWidth(_draw_segment_width); // default
        painter.setPen(pen);
        detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
        pen.setColor(qRgb(102, 102, 102)); // vertex color
        pen.setWidth(1); // default
        painter.setPen(pen);
        brush.setColor(_color_vertex); // qRgb(102, 102, 102)
        brush.setStyle(Qt::SolidPattern);
        painter.setBrush(brush);
        detri2_draw_point(e1, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
        detri2_draw_point(e2, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
        painter.setBrush(Qt::NoBrush);
        bak = seg; // For clearing ->Next
        seg = seg->Next;
        bak->Next = NULL;
      }
      _active_segment = NULL;
    }
    if (theseg != NULL) {
      _active_segment = theseg;
      Vertex *e1 = _active_segment->vrt[0];
      Vertex *e2 = _active_segment->vrt[1];
      pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
      pen.setWidth(_draw_segment_width); // default
      painter.setPen(pen);
      detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
      //pen.setColor(qRgb(255, 0, 0)); // selected color
      //pen.setWidth(1); // default
      //painter.setPen(pen);
      brush.setColor(_color_selected); // qRgb(255, 0, 0)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      detri2_draw_point(e1, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
      detri2_draw_point(e2, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
      painter.setBrush(Qt::NoBrush);
      _active_segment->Next = NULL;
    }
    return 1;
}

int ScribbleArea::is_segment_active(Triang *theseg)
{
  if (_active_segment != NULL) {
    Triang *seg = _active_segment;
    while (seg != NULL) {
      if (seg == theseg) {
        return 1;
      }
      seg = seg->Next;
    }
  }
  return 0;
}

int  ScribbleArea::remove_active_segment(Triang *theseg)
{
    if (_active_segment != NULL) {
      if (theseg == _active_segment) {
        _active_segment = _active_segment->Next;
      } else {
        // Search and remove it.
        Triang *seg = _active_segment;
        Triang *pre_seg = seg; // Bakup
        while (seg != NULL) {
          if (seg == theseg) {
            pre_seg->Next = seg->Next;
            break;
          }
          pre_seg = seg; // Bakup
          seg = seg->Next;
        }
      }
      theseg->Next = NULL;

      // Clear the segment
      QPainter painter(&_canvas[0]);
      QPen pen;
      QBrush brush;
      // First clean the previous selected segment.
      Vertex *e1 = theseg->vrt[0];
      Vertex *e2 = theseg->vrt[1];
      pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
      pen.setWidth(_draw_segment_width); // default
      painter.setPen(pen);
      detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
      pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
      pen.setWidth(1); // default
      painter.setPen(pen);
      brush.setColor(_color_vertex); // qRgb(102, 102, 102)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      detri2_draw_point(e1, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
      detri2_draw_point(e2, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
      painter.setBrush(Qt::NoBrush);
      QString ss = "Removed an active segment";
      m_statusRight->setText(ss);
    }
    return 1;
}

int ScribbleArea::add_active_segment(Triang *theseg)
{
    QPainter painter(&_canvas[0]);
    QPen pen;
    QBrush brush;
    //int ptsize=4;

    Vertex *e1 = theseg->vrt[0];
    Vertex *e2 = theseg->vrt[1];
    pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
    pen.setWidth(_draw_segment_width); // default
    painter.setPen(pen);
    detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
    //pen.setColor(qRgb(255, 0, 0)); // selected color
    //pen.setWidth(1); // default
    //painter.setPen(pen);
    brush.setColor(_color_selected); // qRgb(255, 0, 0)
    brush.setStyle(Qt::SolidPattern);
    painter.setBrush(brush);
    detri2_draw_point(e1, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
    detri2_draw_point(e2, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
    painter.setBrush(Qt::NoBrush);

    int count = 0;
    if (_active_segment == NULL) {
      _active_segment = theseg;
      count++;
    } else {
      Triang *seg = _active_segment;
      count++;
      while (seg->Next != NULL) {
        count++;
        seg = seg->Next;
      }
      seg->Next = theseg; count++;
    }
    assert(theseg->Next == NULL);
    QString ss = "There are "+ QString::number(count) + " active segments";
    m_statusRight->setText(ss);
    return 1;
}

//=============================================================================

int  ScribbleArea::set_active_trilist(detri2::arraypool *newtrilist)
{
    QPainter painter(&_canvas[0]);
    QPen pen;
    QBrush brush;

    if (_active_trilist != NULL) {
      // Clear the current triangles.
      brush.setColor(_color_canvas_background); // white
      brush.setStyle(Qt::SolidPattern);
      //painter.setBrush(brush);
      for (int i = 0; i < _active_trilist->objects; i++) {
        Triang** ptri = (Triang **) _active_trilist->get(i);
        Vertex *p1 = (*ptri)->vrt[0];
        Vertex *p2 = (*ptri)->vrt[1];
        Vertex *p3 = (*ptri)->vrt[2];
        //detri2_draw_triangle(p1, p2, p3, &painter, _cx, _cy, _scale_xy);
        int x1 = (int) ( _scale_xy * p1->crd[0] + _cx);
        int y1 = (int) (-_scale_xy * p1->crd[1] + _cy);
        int x2 = (int) ( _scale_xy * p2->crd[0] + _cx);
        int y2 = (int) (-_scale_xy * p2->crd[1] + _cy);
        int x3 = (int) ( _scale_xy * p3->crd[0] + _cx);
        int y3 = (int) (-_scale_xy * p3->crd[1] + _cy);
        QPainterPath path;
        path.moveTo (x1, y1);
        path.lineTo (x2, y2);
        path.lineTo (x3, y3);
        path.lineTo (x1, y1);
        painter.setPen (Qt :: NoPen);
        painter.fillPath (path, brush);
      }
      //if (_show_geom_edges) {
      if (viewSegmentsAct->isChecked()) {
        pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
        pen.setWidth(_draw_segment_width); // default
        painter.setPen(pen);
        TriEdge E;
        for (int i = 0; i < _active_trilist->objects; i++) {
          E.tri = * (Triang **) _active_trilist->get(i);
          for (E.ver = 0; E.ver < 3; E.ver++) {
            if (E.is_segment()) {
              Vertex *e1 = E.org();
              Vertex *e2 = E.dest();
              detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
            }
          }
        }
      }
      //if (_show_geom_vertices) {
      if (viewVerticesAct->isChecked()) {
        // Re-draw segments and vertices of this subdomain
        pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
        pen.setWidth(1); // default
        painter.setPen(pen);
        brush.setColor(_color_vertex); // qRgb(102, 102, 102)
        brush.setStyle(Qt::SolidPattern);
        painter.setBrush(brush);
        TriEdge E;
        for (int i = 0; i < _active_trilist->objects; i++) {
          E.tri = * (Triang **) _active_trilist->get(i);
          for (E.ver = 0; E.ver < 3; E.ver++) {
            detri2_draw_point(E.org(), _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
          }
        }
        painter.setBrush(Qt::NoBrush);
      }
      delete _active_trilist;
      _active_trilist = NULL;
    }

    if (newtrilist != NULL) {
      _active_trilist = newtrilist;
      brush.setColor(_color_selected); // red qRgb(255, 0, 0)
      brush.setStyle(Qt::SolidPattern);
      for (int i = 0; i < _active_trilist->objects; i++) {
        Triang** ptri = (Triang **) _active_trilist->get(i);
        Vertex *p1 = (*ptri)->vrt[0];
        Vertex *p2 = (*ptri)->vrt[1];
        Vertex *p3 = (*ptri)->vrt[2];
        //detri2_draw_triangle(p1, p2, p3, &painter, _cx, _cy, _scale_xy);
        int x1 = (int) ( _scale_xy * p1->crd[0] + _cx);
        int y1 = (int) (-_scale_xy * p1->crd[1] + _cy);
        int x2 = (int) ( _scale_xy * p2->crd[0] + _cx);
        int y2 = (int) (-_scale_xy * p2->crd[1] + _cy);
        int x3 = (int) ( _scale_xy * p3->crd[0] + _cx);
        int y3 = (int) (-_scale_xy * p3->crd[1] + _cy);
        QPainterPath path;
        path.moveTo (x1, y1);
        path.lineTo (x2, y2);
        path.lineTo (x3, y3);
        path.lineTo (x1, y1);
        painter.setPen (Qt :: NoPen);
        painter.fillPath (path, brush);
      }
      //if (_show_geom_edges) {
      if (viewSegmentsAct->isChecked()) {
        pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
        pen.setWidth(_draw_segment_width); // default
        painter.setPen(pen);
        TriEdge E;
        for (int i = 0; i < _active_trilist->objects; i++) {
          E.tri = * (Triang **) _active_trilist->get(i);
          //assert(E.tri->is_infected());
          for (E.ver = 0; E.ver < 3; E.ver++) {
            if (E.is_segment()) {
              Vertex *e1 = E.org();
              Vertex *e2 = E.dest();
              detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
            }
          }
        }
      }
      //if (_show_geom_vertices) {
      if (viewVerticesAct->isChecked()) {
        // Re-draw segments and vertices of this subdomain
        pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
        pen.setWidth(1); // default
        painter.setPen(pen);
        brush.setColor(_color_vertex); // qRgb(102, 102, 102)
        brush.setStyle(Qt::SolidPattern);
        painter.setBrush(brush);
        TriEdge E;
        for (int i = 0; i < _active_trilist->objects; i++) {
          E.tri = * (Triang **) _active_trilist->get(i);
          for (E.ver = 0; E.ver < 3; E.ver++) {
            detri2_draw_point(E.org(), _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
          }
        }
        painter.setBrush(Qt::NoBrush);
      }
    }

    return 1;
}

//=============================================================================

void ScribbleArea::mousePressEvent(QMouseEvent *event)
{
    //if (event->button() == Qt::RightButton) {
    //  qDebug() << " Right mouse button clicked";
    //} // if (event->button() == Qt::RightButton)
    //else
    if (event->button() == Qt::LeftButton) {
      QPoint lastPoint = event->pos();
      //qDebug() << "mouse pressed at x=" << lastPoint.x() << ", y =" << lastPoint.y();
      // Left button is pushed down.
      _mouse_down=1;
      _mouse_last_x=lastPoint.x();
      _mouse_last_y=lastPoint.y();

      if (_insert_mode) {
        //if (event->modifiers() == Qt::AltModifier) { // Alt key is pressed.
        //  qDebug() << "Alt+ left-mouse-button pressed";
        //if (_insert_rect) {
        if (insertRectAct->isChecked()) {
          // Staring to select a region.
          _select_width = _select_height = 0;
          // Backup the current scene.
          _canvas[1] = _canvas[0];
        } // _insert_rect

        //else if (_insert_arc) {
        else if (insertArcAct->isChecked()) {
          // Backup the current scene.
          _canvas[1] = _canvas[0];
        }

        //else if (_insert_segment) {
        else if (insertSegmentAct->isChecked()) {
          if (_active_vertex) {
            //qDebug() << " active vertex: idx ="<<_active_vertex->idx;
            // Try to insert a segment from the previous selected vertex to the current vertex.
            double x =  (_mouse_last_x - _cx) / _scale_xy;
            double y = -(_mouse_last_y - _cy) / _scale_xy;
            double dist = 0.; TriEdge E; //NE;
            int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
            if ((dist / _geom->io_diagonal) < 1.e-2) { // very close to the vertex E.org()
              Vertex  *v = E.org();
              qDebug() << " selected vertex idx = "<<v->idx;
              if (v != _active_vertex) {
                // Try to insert an edge between _active_vertex and v.
                //int chk = detri2_detect_intersection(_geom, _active_vertex, v, E);
                int chk = _geom->detect_intersection(_active_vertex, v, E);
                qDebug() << "  detect_intersection result = "<<chk;
                int insertflag = true;
                if (chk == INTERSECT_NONE || chk == INTERSECT_SHARE_EDGE) {
                  insertflag = true;
                } else if (chk == INTERSECT_CONFLICT) {
                  m_statusRight->setText("!! Segment already exists");
                  insertflag = false;
                } else if ((chk == INTERSECT_SEGMENT) || (chk == INTERSECT_VERTEX)) {
                  //m_statusRight->setText("!! Segment intersects an existing segment");
                  // see example: http://doc.qt.io/archives/qt-4.8/qmessagebox.html
                  // see http://doc.qt.io/qt-5/qmessagebox.html#StandardButton-enum
                  QMessageBox msgBox;
                  msgBox.setText("This segment intersects existing segment(s).");
                  msgBox.setInformativeText("Do you want to intersect them?");
                  msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
                  msgBox.setDefaultButton(QMessageBox::Yes);
                  int ret = msgBox.exec();
                  if (ret == QMessageBox::Yes) {
                    insertflag = true;
                  } else {
                    insertflag = false;
                  }
                }
                if (insertflag) {
                  // Insert this segment (with intersection)
                  //detri2_insert_segment_intersect(_geom, _active_vertex, v, -1, E);
                  _geom->insert_segment_intersect(_active_vertex, v, -1, 0);
                  // draw this segment.
                  {
                    QPainter painter(&_canvas[0]);
                    QPen pen;
                    pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
                    pen.setWidth(_draw_segment_width); // //pen.setWidth(2); default
                    painter.setPen(pen);
                    detri2_draw_line(_active_vertex, v, &painter, _cx, _cy, _scale_xy);
                  }
                  m_statusRight->setText("Inserted a new segment");
                  // Update v as the active vertex.
                  set_active_vertex(v); //_canvas[1] = _canvas[0];
                  //_object_valid = 0; // Do not re-draw the scene.
                  update();
                }
              } else { // v == _active_vertex
                m_statusRight->setText("Please re-select the first vertex of the segment");
                // Unselect the first vertex.
                set_active_vertex(NULL); // Unslect it.
                update();
              } // if (v != _active_vertex)
            } // if ((dist / _geom->io_diagonal) < 1.e-2)
          } else {
            // Select the first vertex of the segment.
            double x =  (_mouse_last_x - _cx) / _scale_xy;
            double y = -(_mouse_last_y - _cy) / _scale_xy;
            double dist = 0.; TriEdge E;//, NE;
            detri2_locate_point(_geom, x, y, E, dist, 1);
            if ((dist / _geom->io_diagonal) < 1.e-2) {
              if (E.org()->typ == STEINERVERTEX) {
                set_active_vertex(E.org());
                m_statusRight->setText("Please select the second vertex of the segment");
                //assert(_active_vertex = E.org());
                // Update the mouse position to be at the _active_vertex
                //   for restoring its position in MouseReleaseEvent().
                _mouse_last_x =  _scale_xy * _active_vertex->crd[0] + _cx;
                _mouse_last_y = -_scale_xy * _active_vertex->crd[1] + _cy;
                // do not re-draw other objects.
                update();
              }
            }
          } // no _active_vertex
        } // insert a segment

        //else if (_insert_vertex) {
        else if (insertVertexAct->isChecked()) {
          // default insert or select a vertex
          double x =  (_mouse_last_x - _cx) / _scale_xy;
          double y = -(_mouse_last_y - _cy) / _scale_xy;
          double dist = 0.; TriEdge E, NE;
          E.tri = NULL, NE.tri = NULL;
          printf("debug: locate point for inserting a point.\n");
          int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
          if (loc == LOC_IN_OUTSIDE) {
            // Lies outside the bbox.
            detri2_enlarge_bbox(&_geom, x, y);
            calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
            //qDebug()<< "Adding a new vertex at x="<<x<<" y="<<y;
            E.tri = 0; E.ver = 0; // Re=locate the point (x,y).
            detri2_insert_point(_geom, x, y, E, loc);
            set_active_vertex(E.org());
            _object_valid=0;
            update(); //updateGL();
          } else {
            if ((dist / _geom->io_diagonal) < 1.e-2) {
              // Very close to an existing vertex. Select it and make it active vertex.
              // So it can be moved by draging the mouse.
              qDebug()<< "Select an existing vertex dist = "<< dist;
              if (E.org()->typ == STEINERVERTEX) {
                QString ss = "Selected vertex idx=" + QString::number(E.org()->idx);
                m_statusRight->setText(ss);
                set_active_vertex(E.org());
                // Update the mouse position to be at the _active_vertex
                //   for restoring its position in MouseReleaseEvent().
                _mouse_last_x =  _scale_xy * _active_vertex->crd[0] + _cx;
                _mouse_last_y = -_scale_xy * _active_vertex->crd[1] + _cy;
                // do not re-draw other objects.
                update();
              }
            } else {
              // Insert a vertex (might split a segment as well).
              bool insertflag = true;
              if (loc == LOC_IN_TRI) {
                // Check if this edge is too close to a segment.
                printf("  debug: insert vertex: locatring an edge.\n");
                NE = E;
                detri2_locate_edge(_geom, x, y, NE);
                printf("   aaaaaaaaaa\n");
                if (NE.is_segment()) {
                  printf("   bbbbbbbbbbbb\n");
                  Vertex p, prj; p.init(); prj.init();
                  p.crd[0] = x; p.crd[1] = y;
                  detri2_projpt2edge(&p, NE.org(), NE.dest(), &prj);
                  printf("   ccccccccccccc\n");
                  double dd = _geom->get_distance(&p, &prj);
                  double len = _geom->get_distance(NE.org(), NE.dest());
                  if ((dd / len) < 0.1) {
                    // Ask if we need to proj the point onto this segment.
                    QMessageBox msgBox;
                    msgBox.setText("A segment is nearby.");
                    msgBox.setInformativeText("Do you want to project this point onto it?");
                    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
                    msgBox.setDefaultButton(QMessageBox::Yes);
                    int ret = msgBox.exec();
                    if (ret == QMessageBox::Yes) {
                      // Project this point.
                      //Vertex p, prj; p.init(); prj.init();
                      //p.crd[0] = x; p.crd[1] = y;
                      //detri2_projpt2edge(&p, NE.org(), NE.dest(), &prj);
                      x = prj.crd[0];
                      y = prj.crd[1];
                      E = NE;
                      loc = LOC_ON_EDGE;
                      insertflag = true;
                    } else if (ret == QMessageBox::No) {
                      // Do not split a near segment.
                      insertflag = true;
                    } else if (ret == QMessageBox::Cancel) {
                      // Do not insert this point.
                      insertflag = false;
                    }
                  } // if ((dd / len) < 0.1)
                }
              } else if ((loc == LOC_ON_EDGE) && E.is_segment()) {
                // Ask split a segment?
                // Project this point onto the segment.
                Vertex p, prj; p.init(); prj.init();
                p.crd[0] = x; p.crd[1] = y;
                detri2_projpt2edge(&p, NE.org(), NE.dest(), &prj);
                x = prj.crd[0];
                y = prj.crd[1];
              } else if (loc == LOC_ON_VERT) {
                assert(0); // Should be handled in above.
              }
              if (insertflag) {
                qDebug()<< "Adding a new vertex at x="<<x<<" y="<<y;
                detri2_insert_point(_geom, x, y, E, loc);
                set_active_vertex(E.org());
                // do not re-draw //_object_valid=0;
                update(); //updateGL();
              }
            }
          }
        } // insert a vertex
      } // if (_insert_mode)

      else if (_select_mode) {
        //if (_select_mode && (event->modifiers() == Qt::AltModifier)) {
        double x =  (_mouse_last_x - _cx) / _scale_xy;
        double y = -(_mouse_last_y - _cy) / _scale_xy;
        double dist = 0.; TriEdge E, NE;
        
        //if (_assign_subdomain) {
        if (assignSubdomainAct->isChecked()) {
          qDebug()<<" assign subdomain";
          int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
          if (loc != LOC_IN_OUTSIDE) {
            arraypool *newtrilist = new arraypool(sizeof(Triang *), 8);
            detri2_get_trilist(E.tri, newtrilist);
            if (_active_trilist != NULL) {
              // There exists an already selected subdomain.
              Triang *testtri = * (Triang **) _active_trilist->get(0);
              if (detri2_is_tri_in_trilist(newtrilist, testtri)) {
                // This is the same subdomain.
                delete newtrilist;
                set_active_trilist(NULL); // Unselect it.
              } else {
                if (!detri2_is_exterior_trilist(newtrilist)) {
                  qDebug()<<" set active subdomain";
                  set_active_trilist(newtrilist);
                } else {
                  // This means that this triangle belongs to exterior.
                  delete newtrilist;
                  set_active_trilist(NULL); // Unselect it.
                }
              }
              update();
            } else {
              if (!detri2_is_exterior_trilist(newtrilist)) {
                qDebug()<<" set active subdomain";
                set_active_trilist(newtrilist);
                update();
              } else {
                // This means that this triangle belongs to exterior.               
                delete newtrilist;
              }
            }
          }
        } // if (_assign_subdomain)
        
        //else if (_assign_segment) {
        else if (assignSegmentAct->isChecked()) {
          detri2_locate_edge(_geom, x, y, E);
          //TriEdge NE;
          //int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
          //if ((loc == LOC_IN_TRI) && (NE.tri != NULL)) {
          //  E = NE;
          //}
          if (E.is_segment()) {
            Triang *seg = E.get_segment();
            if (seg != NULL) {
              // We got a segment.
              if (event->modifiers() & Qt::ShiftModifier) {
                if (is_segment_active(seg)) {
                  remove_active_segment(seg);
                } else {
                  add_active_segment(seg);
                }
              } else {
                // Without Shift-key pressed, only select one segment.
                if (_active_segment && (_active_segment->Next != NULL)) {
                  set_active_segment(NULL); // Unselect multiple segment.
                }
                if (_active_segment && (_active_segment == seg)) {
                  set_active_segment(NULL); // unselect it.
                } else {
                  set_active_segment(seg);
                }
              }
              update();
            }
          } // E.is_segment()
        } // select a segment

        //else if (_assign_vertex) { // select vertex
        else if (assignVertexAct->isChecked()) { // select vertex
          //double x =  (_mouse_last_x - _cx) / _scale_xy;
          //double y = -(_mouse_last_y - _cy) / _scale_xy;
          TriEdge NE; //double dist = 0.; TriEdge E;
          if (event->modifiers() & Qt::AltModifier) {
            // Staring to select a region.
            _select_width = _select_height = 0;
            // Backup the current scene.
            _canvas[1] = _canvas[0];
          } else if (event->modifiers() & Qt::ShiftModifier) {
            //qDebug() << "Shift key is pressed.";
            int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
            if (loc != LOC_IN_OUTSIDE) {
              // Check if this vertex E.org() already be selected.
              Vertex *thevertex = E.org();
              if (is_vertex_active(thevertex)) {
                remove_active_vertex(thevertex);
              } else {
                add_active_vertex(thevertex);
              }
              update();
            } // loc > 0
          } else { 
            // Without Shift-key pressed, only select one vertex.
            if (_active_vertex && (_active_vertex->Next != NULL)) {
              set_active_vertex(NULL); // Unselect multiple vertices.
            }
            int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
            if (loc != LOC_IN_OUTSIDE) {
              if (_active_vertex &&
                 (_active_vertex == E.org())) {
                set_active_vertex(NULL); // Unslect it.
                update();
              } else {
                // Do not select a vertex of the Bbox.
                if (E.org()->typ == STEINERVERTEX) {
                  QString ss = "Selected vertex idx=" + QString::number(E.org()->idx);
                  m_statusRight->setText(ss);
                  set_active_vertex(E.org());
                  update();
                }
              }
            } // if (loc > 0)
          }
        } // select a vertex or assign vertex weight

        //else if (_assign_weight) {
        else if (assignVertexWeightAct->isChecked()) {
          // Select a vertex
          //double x =  (_mouse_last_x - _cx) / _scale_xy;
          //double y = -(_mouse_last_y - _cy) / _scale_xy;
          TriEdge NE; //double dist = 0.; TriEdge E;
          int loc = detri2_locate_point(_geom, x, y, E, dist, 1);
          if (loc != LOC_IN_OUTSIDE) {
            if (_active_vertex && (_active_vertex == E.org())) {
              // the active vertex does not change.
              //if (_assign_weight) {
              //if (assignVertexWeightAct->isChecked()) { // this must be true
                // Bakup the curent image.
                _canvas[1] = _canvas[0];
              //} else {
              //  set_active_vertex(NULL); // Unslect it.
              //  update();
              //}
            } else {
              // Do not select a vertex of the Bbox.
              if (E.org()->typ == STEINERVERTEX) {
                QString ss = "Selected vertex idx=" + QString::number(E.org()->idx);
                m_statusRight->setText(ss);
                set_active_vertex(E.org());
                update();
              }
            }
          }
        } // _assign_weight
      } // select_mode

      //else if (_view_power_cell) {
      else if (viewPowerCellAct->isChecked() || viewMassCenterAct->isChecked()) {
        if (_tria && _tria->tr_tris->objects > 0) {
          //qDebug() << " view power cell";
          double x =  (_mouse_last_x - _cx) / _scale_xy;
          double y = -(_mouse_last_y - _cy) / _scale_xy;
          TriEdge E, NE; E.tri=NULL; double dist=0.;
          detri2_locate_point(_tria, x, y, E, dist, 1);
          /* [2018-11-05] done in the above routine.
          if (loc == LOC_IN_OUTSIDE) {
            qDebug() << "  loc="<<loc;
            // Outside, choose the closer vertex of E.
            Vertex pt; pt.crd[0] = x; pt.crd[1] = y;
            Vertex *v1 = E.org();
            Vertex *v2 = E.dest();
            double d1 = _tria->get_distance(&pt, v1);
            double d2 = _tria->get_distance(&pt, v2);
            if (d1 > d2) {
              E = E.esym(); // dest() is more closer.
            }
          }
          */
          // E.org() returns the nearest vertex to pt.
          // Draw the selected mesh vertex.
          Vertex *vt = E.org();
          qDebug() << "  mesh pt= ["<<vt->crd[0]<<","<<vt->crd[1]<<"]";
          QPainter painter(&_canvas[0]);
          QPen pen;
          pen.setColor(_color_dual_edge); // qRgb(153, 153, 255)
          pen.setWidth(_draw_dual_edge_width); // 4
          painter.setPen(pen);
          QBrush brush;
          if (viewMassCenterAct->isChecked()) {
            brush.setColor(_color_dual_edge);
            brush.setStyle(Qt::SolidPattern);
            painter.setBrush(brush);
          }
          //int px = (int) ( _scale_xy * vt->crd[0] + _cx);
          //int py = (int) (-_scale_xy * vt->crd[1] + _cy);
          //painter.drawEllipse(QPoint(px,py), 8, 8);
          //detri2_draw_point(vt, 8, &painter, _cx, _cy, _scale_xy);
          detri2_draw_voronoi_cell(_tria, _geom, vt, viewMassCenterAct->isChecked(),
                                   _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
          //detri2_draw_voronoi_cell_debug(_tria, _geom, vt, &painter, _cx, _cy, _scale_xy);
          if (viewMassCenterAct->isChecked()) {
            painter.setBrush(Qt::NoBrush);
          }
          update();
        } // if (_tria)
      }

      //else if (_view_triangle_orthocircle || _view_triangle_orthocenter || _view_triangle_info) {
      else if (viewTriangleInfoAct->isChecked() ||
               viewOthocircleAct->isChecked() ||
               viewOthocenterAct->isChecked()) {
        //qDebug()<<" view triangle orthocircle & information";
        if (_tria && _tria->tr_tris->objects > 0) {
          double x =  (_mouse_last_x - _cx) / _scale_xy;
          double y = -(_mouse_last_y - _cy) / _scale_xy;
          TriEdge E, NE; E.tri=NULL; double dist=0.;
          int loc = detri2_locate_point(_tria, x, y, E, dist, 1);
          qDebug() << "  loc="<<loc;
          if (loc != LOC_IN_OUTSIDE) {
            QPainter painter(&_canvas[0]);
            QPen pen;
            //if (_view_triangle_orthocircle || _view_triangle_orthocenter) {
            if (viewOthocircleAct->isChecked() || viewOthocenterAct->isChecked()) {
              assert(_tria->OMT_domain == NULL);
              _tria->get_tri_orthocenter(E.tri);
              int px = (int) ( _scale_xy * E.tri->cct[0] + _cx);
              int py = (int) (-_scale_xy * E.tri->cct[1] + _cy);
              //if (_view_triangle_orthocenter) {
              if (viewOthocenterAct->isChecked()) {
                // Draw the orthocenter.
                pen.setColor(qRgb(244, 66, 235)); // color
                pen.setWidth(1); // default
                painter.setPen(pen);
                QBrush brush;
                brush.setColor(qRgb(244, 66, 235));
                brush.setStyle(Qt::SolidPattern);
                painter.setBrush(brush);
                painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
                painter.setBrush(Qt::NoBrush);
              } else {
                // Draw orthocirle
                pen.setColor(qRgb(244, 66, 235));
                pen.setWidth(4); // default
                painter.setPen(pen);
                // Weight may be negative
                if (E.tri->cct[2] > 0) {
                  qDebug()<<"  radius = "<< sqrt(E.tri->cct[2]);
                  int rr = (int) ( _scale_xy * sqrt(E.tri->cct[2]));
                  painter.drawEllipse(QPoint(px,py), rr, rr);
                } else {
                  QString text;
                  text.sprintf("%.2f", E.tri->cct[2]);
                  painter.drawText(QPoint(px,py), text);
                }
              }
            //} else if (_view_triangle_info) {
            } else if (viewTriangleInfoAct->isChecked()) {
              pen.setColor(qRgb(0, 255, 255));
              pen.setWidth(2); // default
              painter.setPen(pen);
              Vertex *pa = E.org();
              Vertex *pb = E.dest();
              Vertex *pc = E.apex();
              double midx = (pa->crd[0] + pb->crd[0] + pc->crd[0]) / 3.0;
              double midy = (pa->crd[1] + pb->crd[1] + pc->crd[1]) / 3.0;
              detri2_draw_line(pa, pb, &painter, _cx, _cy, _scale_xy);
              detri2_draw_line(pb, pc, &painter, _cx, _cy, _scale_xy);
              detri2_draw_line(pc, pa, &painter, _cx, _cy, _scale_xy);
              int px = (int) ( _scale_xy * midx + _cx);
              int py = (int) (-_scale_xy * midy + _cy);
              //painter.drawText(QPoint(px,py), QString::number(E.tri->tag));
              QString ss = "tri [";
              ss += QString::number(pa->idx);
              ss += ",";
              ss += QString::number(pb->idx);
              ss += ",";
              ss += QString::number(pc->idx);
              ss += "] tag(";
              ss += QString::number(E.tri->tag);
              ss += ")";
              if (E.tri->val > 0) {
                ss += (" area("+QString::number(E.tri->val)+")");
              }
              m_statusRight->setText(ss);
            }
            update();
          }
        } // if (_tria)
      }

      //else if (_zoom_selected_region) {
      else if (zoomSelectAct->isChecked()) {
        // Staring to select a region.
        _select_width = _select_height = 0;
        // Backup the current scene.
        _canvas[1] = _canvas[0];
      }
    } // if (event->button() == Qt::LeftButton)
}

void ScribbleArea::mouseMoveEvent(QMouseEvent *event)
{
    if ((event->buttons() & Qt::LeftButton)) {
      QPoint lastPoint = event->pos();
      //qDebug() << "mouse moved to x=" << lastPoint.x() << ", y =" << lastPoint.y();
      double dx,dy;
      int x=lastPoint.x();
      int y=lastPoint.y();
      dx=(double)(x-_mouse_last_x);
      dy=(double)(y-_mouse_last_y);
      _mouse_down_move=1;

      //if (_trans_mode) {
      if (transAct->isChecked()) {
        // Default, move the whole object.
        _mouse_last_x=x;
        _mouse_last_y=y;
        _cx += _mouse_sens * dx;
        _cy += _mouse_sens * dy;
        _object_valid=0; // invalidate_object();
        update(); //updateGL();
      }

      else if (zoomSelectAct->isChecked() ||
               insertRectAct->isChecked() ||
               (assignVertexAct->isChecked() && (event->modifiers() & Qt::AltModifier))) {
        // Draw a rectangle of a selected region.
        //_select_width = dx;
        //_select_height = dy;
        _canvas[0] = _canvas[1]; // restore the image.
        QPainter painter(&_canvas[0]);
        QPen pen(qRgb(255,0,255));
        pen.setWidth(2);
        painter.setPen(pen);
        QRect rect(_mouse_last_x, _mouse_last_y, dx, dy);
        painter.drawRect(rect);
        update();
      } // if (_zoom_selected_region)

      //else if (_insert_arc) {
      else if (insertArcAct->isChecked()) {
        _canvas[0] = _canvas[1]; // restore the image.
        // Draw a circle centered at it.
        int U = _mouse_last_x; //  (int) ( _scale_xy * _mouse_last_x + _cx);
        int V = _mouse_last_y; // (int) (-_scale_xy * _mouse_last_y + _cy);
        //int dx = lastPoint.x() - U;
        //int dy = lastPoint.y() - V;
        int R = (int) sqrt((double)(dx*dx + dy*dy));
        QPainter painter(&_canvas[0]);
        QPen pen(qRgb(255,0,255)); // red
        pen.setWidth(2);
        painter.setPen(pen);
        //qDebug()<<"  draw circle"<<"cx="<<U<<" cy="<<V<<" R="<<R;
        painter.drawEllipse(QPoint(U,V), R, R);
        update();
      }

      //else if (_assign_weight) {
      else if (assignVertexWeightAct->isChecked()) {
        if (_active_vertex) {
          // A vertex was selected.
          //double px =  ((double)lastPoint.x() - _cx) / _scale_xy;
          //double py = -((double)lastPoint.y() - _cy) / _scale_xy;
          //double u = _active_vertex->crd[0];
          //double v = _active_vertex->crd[1];
          //double w = (px - u)*(px - u) + (py - v)*(py - v);
          //qDebug() << " weight="<<w;
          //_active_vertex->wei = w;
          // Draw a circle centered at it.
          int U = (int) ( _scale_xy * _active_vertex->crd[0] + _cx);
          int V = (int) (-_scale_xy * _active_vertex->crd[1] + _cy);
          int dx = lastPoint.x() - U;
          int dy = lastPoint.y() - V;
          int R = (int) sqrt((double)(dx*dx + dy*dy));
          _canvas[0] = _canvas[1]; // restore the image.
          QPainter painter(&_canvas[0]);
          QPen pen(qRgb(255,0,0)); // red
          pen.setWidth(2);
          painter.setPen(pen);
          painter.drawEllipse(QPoint(U,V), R, R);
          update();
        } else {
          m_statusRight->setText("Please select a vertex");
        }
      } // if (_assign_weight)

      else if (_insert_mode) {
        // Move selected vertex.
        if ((_active_vertex != NULL) && (_active_vertex->Next == NULL)) {
          // Only a vertex was selected.
          double px =  ((double)lastPoint.x() - _cx) / _scale_xy;
          double py = -((double)lastPoint.y() - _cy) / _scale_xy;
          double u = _active_vertex->crd[0];
          double v = _active_vertex->crd[1];
          // Move the vertex if this (px, py) is close to (u, v);
          double dist = sqrt((px - u)*(px - u) + (py - v)*(py - v));
          //qDebug() << " dist="<<dist << ", ratio="<<(dist / _geom->io_diagonal);
          if ((dist / _geom->io_diagonal) < 0.01) { // 0.01 is like the moving speed.
            // get the weight of _active_vertex->crd
            //double w = u*u + v*v - _active_vertex->crd[2];
            _active_vertex->crd[0] = px;
            _active_vertex->crd[1] = py;
            _active_vertex->crd[2] = px*px + py*py;
            if (_active_vertex->Pair != NULL) {
              // update the corresponding mesh vertex.
              Vertex *mv = _active_vertex->Pair;
              mv->crd[0] = px;
              mv->crd[1] = py;
              mv->crd[2] = _active_vertex->crd[2] - _active_vertex->wei; // include weights
            }
            _object_valid = 0;
            update();
          }
        }
      } // if (_insert_mode)

      /*
      else {
        // Default, move the whole object.
        _mouse_last_x=x;
        _mouse_last_y=y;
        _cx += _mouse_sens * dx;
        _cy += _mouse_sens * dy;
        _object_valid=0; // invalidate_object();
        update(); //updateGL();
      }
      */
    } // if ((event->buttons() & Qt::LeftButton))

    else {
      // show the current mouse location on StatusBar
      QPoint lastPoint = event->pos();
      //qDebug() << "mouse moved to x=" << lastPoint.x() << ", y =" << lastPoint.y();
      double x =  ((double)lastPoint.x() - _cx) / _scale_xy;
      double y = -((double)lastPoint.y() - _cy) / _scale_xy;
      //qDebug() << "position x=" << x << ", y =" << y;
      QString ss = QString::number(x, 'f', 2) + "," + QString::number(y, 'f', 2);
      m_statusLeft->setText(ss);
    }
}

void ScribbleArea::mouseReleaseEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
      // Left button is pushed down.
      QPoint lastPoint = event->pos();
      //qDebug() << "Mouse released at x="<<lastPoint.x()<<", y="<<lastPoint.y();

      //if (_zoom_selected_region) {
      if (zoomSelectAct->isChecked()) {
        if (_mouse_down_move) {
          // Default, a region is selected.
          _select_x = _mouse_last_x;
          _select_y = _mouse_last_y;
          _select_width = lastPoint.x()-_mouse_last_x;
          _select_height = lastPoint.y()-_mouse_last_y;
          //qDebug() << "Selection rect: x=" << _select_x << " y=" << _select_y
          //         << " w=" << _select_width << " h=" << _select_height << endl;
          //debug_selection_region(); // debug
          double xmin, ymin, xmax, ymax;
          get_selection_region(xmin, xmax, ymin, ymax);
          calc_trans(xmin, xmax, ymin, ymax);
          //_zoom_selected_region = 0; // only do it once.
          _object_valid=0;
          update(); //updateGL();
        }
      } // if (_zoom_selected_region)

      else if (assignVertexAct->isChecked() && (event->modifiers() & Qt::AltModifier)) {
        if (_mouse_down_move) {
          // Clear the rectangle.
          _canvas[0] = _canvas[1]; // restore the image.
          // Get the selected rectangle.
          _select_x = _mouse_last_x;
          _select_y = _mouse_last_y;
          _select_width = lastPoint.x()-_mouse_last_x;
          _select_height = lastPoint.y()-_mouse_last_y;
          //qDebug() << "Selection rect: x=" << _select_x << " y=" << _select_y
          //         << " w=" << _select_width << " h=" << _select_height << endl;
          double xmin, ymin, xmax, ymax;
          get_selection_region(xmin, xmax, ymin, ymax);
          //qDebug() << "Selection rect: x=" << xmin << " y=" << ymin
          //         << " w=" << xmax << " h=" << ymax << endl;
          for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
            Vertex *vt = (Vertex *) _geom->tr_steiners->get(i);
            if (vt->is_deleted()) continue;
            // Test if this vertex lies in this region.
            if ((vt->crd[0] > xmin) && (vt->crd[0] < xmax) &&
                (vt->crd[1] > ymin) && (vt->crd[1] < ymax)) {
              if (is_vertex_active(vt)) {
                remove_active_vertex(vt);
              } else {
                add_active_vertex(vt);
              }
            }
          }
          update();
        }
      }

      //else if (_insert_arc) {
      else if (insertArcAct->isChecked()) {
        if (_mouse_down_move) {
          double centx =  (_mouse_last_x - _cx) / _scale_xy;
          double centy = -(_mouse_last_y - _cy) / _scale_xy;
          double dx = lastPoint.x()-_mouse_last_x;
          double dy = lastPoint.y()-_mouse_last_y;
          double rr = sqrt(dx*dx + dy*dy) / _scale_xy;
          GeometryInputPanel *Panel = new GeometryInputPanel(2);
          Panel->centx->setText(QString::number(centx));
          Panel->centy->setText(QString::number(centy));
          Panel->cir_radius->setText(QString::number(rr));
          Panel->cir_angle_start->setText(tr("0"));
          Panel->cir_angle_end->setText(tr("360"));
          Panel->cir_segnum->setText(tr("18"));
          Panel->segment_tag->setText(tr("-1"));
          if(Panel->exec() == QDialog::Accepted){
            double rcx = Panel->centx->text().toDouble();
            double rcy = Panel->centy->text().toDouble();
            double radius = Panel->cir_radius->text().toDouble();
            double ang1 = Panel->cir_angle_start->text().toDouble();
            double ang2 = Panel->cir_angle_end->text().toDouble();
            double segnum = Panel->cir_segnum->text().toDouble();
            int tag = Panel->segment_tag->text().toInt();
            if ((radius > 0) && (ang2 > ang1) && (segnum > 1)) {
              if (Panel->use_trans->isChecked()) {
                // Translate the rectangle.
                double rotcx = Panel->rot_cx->text().toDouble();
                double rotcy = Panel->rot_cy->text().toDouble();
                double rotang = Panel->rot_angle->text().toDouble();
                double tx  = Panel->trans_x->text().toDouble();
                double ty  = Panel->trans_y->text().toDouble();
                detri2_add_circle(&_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                  true, rotcx, rotcy, rotang, tx, ty);
              } else {
                detri2_add_circle(&_geom, rcx, rcy, radius, ang1, ang2, segnum, tag,
                                  false, 0, 0, 0, 0, 0);
              }
              QString ss = QString("Added a circle/arc");
              m_statusRight->setText(ss);
              _object_valid=0;
            } else {
              // Only clear the rectangle.
              _canvas[0] = _canvas[1]; // restore the image.
            }
          } else {
            // Only clear the rectangle.
            _canvas[0] = _canvas[1]; // restore the image.
          }
          delete Panel;
          update(); //updateGL();
        }
      } // if (_insert_arc)

      //else if (_insert_rect) {
      else if (insertRectAct->isChecked()) {
        if (_mouse_down_move) {
          _select_x = _mouse_last_x;
          _select_y = _mouse_last_y;
          _select_width = lastPoint.x()-_mouse_last_x;
          _select_height = lastPoint.y()-_mouse_last_y;
          //qDebug() << "Selection rect: x=" << _select_x << " y=" << _select_y
          //         << " w=" << _select_width << " h=" << _select_height << endl;
          //debug_selection_region(); // debug
          double xmin, ymin, xmax, ymax;
          get_selection_region(xmin, xmax, ymin, ymax);
          // Open the dialog.
          GeometryInputPanel *Panel = new GeometryInputPanel(1);
          Panel->lbl_centx->setText(tr("xmin"));
          Panel->lbl_centy->setText(tr("ymin"));
          Panel->lbl_rect_width->setText(tr("xmax"));
          Panel->lbl_rect_heigh->setText(tr("ymax"));
          Panel->centx->setText(QString::number(xmin));
          Panel->centy->setText(QString::number(ymin));
          Panel->rect_width->setText(QString::number(xmax));
          Panel->rect_height->setText(QString::number(ymax));
          if(Panel->exec() == QDialog::Accepted){
            //qDebug()<<"Add a rectangle";
            xmin = Panel->centx->text().toDouble();
            ymin = Panel->centy->text().toDouble();
            xmax = Panel->rect_width->text().toDouble();
            ymax = Panel->rect_height->text().toDouble();
            int tag = Panel->segment_tag->text().toInt();
            double ctx = 0.5 * (xmin + xmax);
            double cty = 0.5 * (ymin + ymax);
            double wid = xmax - xmin;
            double hei = ymax - ymin;
            if (Panel->use_trans->isChecked()) {
              // Translate the rectangle.
              double rcx = Panel->rot_cx->text().toDouble();
              double rcy = Panel->rot_cy->text().toDouble();
              double ang = Panel->rot_angle->text().toDouble();
              double dx  = Panel->trans_x->text().toDouble();
              double dy  = Panel->trans_y->text().toDouble();
              detri2_add_rectangle(&_geom, ctx, cty, wid, hei, tag,
                                   true, rcx, rcy, ang, dx, dy);
            } else {
              detri2_add_rectangle(&_geom, ctx, cty, wid, hei, tag,
                                   false, 0, 0, 0, 0, 0);
            }
            QString ss = QString("Added a rectnagle");
            m_statusRight->setText(ss);
            _object_valid=0;
          } else {
            // Only clear the rectangle.
            _canvas[0] = _canvas[1]; // restore the image.
          }
          delete Panel;
          update(); //updateGL();
        }
      } // insert_rect

      //else if (_assign_weight) {
      else if (assignVertexWeightAct->isChecked()) {
        if (_active_vertex && _mouse_down_move) {
          _canvas[0] = _canvas[1]; // restore the image, clear the dynamic circle.
          // Set the new weight.
          double old_wei = _active_vertex->wei;
          // Set the new weight.
          double px =  ((double)lastPoint.x() - _cx) / _scale_xy;
          double py = -((double)lastPoint.y() - _cy) / _scale_xy;
          double u = _active_vertex->crd[0];
          double v = _active_vertex->crd[1];
          double w = (px - u)*(px - u) + (py - v)*(py - v);
          //qDebug() << " weight="<<w;
          _active_vertex->wei = w;
          QString ss = QString("Changed vertex weight from ")
                     + QString::number(old_wei) + QString("to ")
                     + QString::number(w);
          m_statusRight->setText(ss);
          //if (_show_weights) {
          if (viewWeightsAct->isChecked()) {
            // Do not draw the scene. //_object_valid=0;
            QPainter painter(&_canvas[0]);
            int U = (int) ( _scale_xy * _active_vertex->crd[0] + _cx);
            int V = (int) (-_scale_xy * _active_vertex->crd[1] + _cy);
            // Clean the old weight.
            if (fabs(old_wei) > 0.) {
              int r = (int) (sqrt(fabs(old_wei)) * _scale_xy);
              QPen pen(_color_canvas_background); // background color qRgb(255,255,255)
              pen.setWidth(2);
              painter.setPen(pen);
              painter.drawEllipse(QPoint(U,V), r, r);
            }
            // Draw the new weight.
            //int dx = lastPoint.x() - U;
            //int dy = lastPoint.y() - V;
            //int R = (int) sqrt((double)(dx*dx + dy*dy));
            int R = (int) (sqrt(w) * _scale_xy);
            //QPainter painter(&_canvas[0]);
            QPen pen(qRgb(255,140,0)); // weight color
            pen.setWidth(2);
            painter.setPen(pen);
            painter.drawEllipse(QPoint(U,V), R, R);
            // clear the active vertex.
            set_active_vertex(NULL);
            // update the background picture.
            _canvas[1] = _canvas[0];
            update();
          }
        }
      } // if (_assign_weight)

      /* // Only for debug
      else if (_assign_subdomain && _active_trilist) {
        int count = 0;
        for (int i = 0; i < _geom->tr_tris->used_items; i++) {
          Triang *tri = (Triang *) _geom->tr_tris->get(i);
          if (tri->is_deleted() || tri->is_hulltri()) continue;
          if (tri->is_infected()) count++;
        }
        qDebug()<<"  Found "<<count<<" infected triangles";
        if (count != _active_trilist->objects) {
          qDebug()<<"  Unmatched infected triangles should be = "<< _active_trilist->objects;
        }
      } // Only for debug
      */

      else if (_insert_mode) {
        if (_active_vertex && _mouse_down_move) {
          // Check if it has been moved.
          double px =  ((double)_mouse_last_x - _cx) / _scale_xy;
          double py = -((double)_mouse_last_y - _cy) / _scale_xy;
          double u = _active_vertex->crd[0];
          double v = _active_vertex->crd[1];
          // Move the vertex if this (px, py) is close to (u, v);
          double dist = sqrt((px - u)*(px - u) + (py - v)*(py - v));
          qDebug() << " Moved distance="<<dist << ", ratio="<<(dist / _geom->io_diagonal);
          if ((dist / _geom->io_diagonal) > 0.01) {
            // Move the vertex in _geom.
            // First put this vertex back to its original position.
            _active_vertex->crd[0] = px;
            _active_vertex->crd[1] = py;
            _active_vertex->crd[2] = px*px + py*py;
            // Move the vertex from (px, py) to (u,v).
            if (!detri2_move_vertex(_geom, &_active_vertex, u, v)) {
              // Not moved. Must update the scene.
              if (_active_vertex->Pair != NULL) {
                Vertex *mv = _active_vertex->Pair;
                mv->crd[0] = px;
                mv->crd[1] = py;
                mv->crd[2] = px*px + py*py - _active_vertex->wei; // include weights
              }
              _object_valid=0;
              update();
            }
          }
        }
      } // if (_insert_mode)

      _mouse_down=_mouse_down_move=0;
      _mouse_last_x=lastPoint.x();
      _mouse_last_y=lastPoint.y();
    }
}

void ScribbleArea::wheelEvent(QWheelEvent *e)
{
    double dy = 0.0; //,scale;
    //qDebug() << "e->delta() = " << e->delta() << endl;
    if (e->delta() > 0)
        dy = 0.1;
    else if (e->delta() < 0)
        dy = -0.1; //if (_zoom > 0.2)

    e->accept();

    //_zoom += (_mouse_sens * dy);p
    //if (_zoom < _zoom_min) _zoom = _zoom_min;
    //if (_zoom > _zoom_max) _zoom = _zoom_max;

    _scale_xy *= (1.0 + _mouse_sens * dy);
    //qDebug()<<"Added scale: Sxy="<<_scale_xy;

    _object_valid=0; // invalidate_object();
    update(); //updateGL();
}

/*
void ScribbleArea::keyPressEvent(QKeyEvent *event)
{
    //if (!isStarted || isPaused || curPiece.shape() == NoShape) {
    //    QFrame::keyPressEvent(event);
    //    return;
    //}
    qDebug() << "A key press is received:  event->key() = "<< event->key();

    switch (event->key()) {
    case Qt::Key_A:
        qDebug() <<" Key A";
        break;
    case Qt::Key_B:
        qDebug() <<" Key B";
        break;
    case Qt::Key_C:
        break;
    case Qt::Key_D:        
        break;
    case Qt::Key_E:
        break;
    case Qt::Key_I:
        break;
    case Qt::Key_O:
        break;
    case Qt::Key_P:
        qDebug() <<" Key P";
        if(event->modifiers() & Qt::AltModifier){
          qDebug() <<" Key Alt+P";
        }
        break;
    case Qt::Key_R:
        break;
    case Qt::Key_S:
        break;
    case Qt::Key_T:
        break;
    case Qt::Key_V:
        break;
    case Qt::Key_W:
        break;
    case Qt::Key_Z:
        break;
    case Qt::Key_Left:
        //tryMove(curPiece, curX - 1, curY);
        break;
    case Qt::Key_Right:
        //tryMove(curPiece, curX + 1, curY);
        break;
    case Qt::Key_Down:
        //tryMove(curPiece.rotatedRight(), curX, curY);
        break;
    case Qt::Key_Up:
        //tryMove(curPiece.rotatedLeft(), curX, curY);
        break;
    case Qt::Key_Space:
        //dropDown();
        break;
    default:
        return; //ScribbleArea::keyPressEvent(event);
    }
}
*/

//=============================================================================

void ScribbleArea::resizeEvent(QResizeEvent *event)
{
    if (width() != _canvas[0].width() || height() != _canvas[0].height()) {
        calc_trans(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax);
        //resizeImage(&_canvas, QSize(width(), height()));
        QImage newImage(QSize(width(), height()), QImage::Format_RGB32);
        //_canvas[0] = _canvas[1] = newImage;
        _canvas[1] = newImage;
        _object_valid=0; // re-draw the scene.
        update();
    }
    QWidget::resizeEvent(event);
}

void ScribbleArea::draw_scene(QPainter &painter)
{
    //if (_show_geom_bbox) {
    if (viewBboxAct->isChecked()) {
      QPen pen;
      pen.setColor(qRgb(204, 204, 204));
      pen.setWidth(2);
      painter.setPen(pen);
      //QFont serifFont("Times", 10, QFont::Bold);
      //QFont sansFont("Helvetica [Cronyx]", 12);
      //painter.setFont(sansFont);
      painter.setFont(_font_bbox_label);
      detri2_draw_bbox(_geom->io_xmin, _geom->io_xmax, _geom->io_ymin, _geom->io_ymax,
                       &painter, _cx, _cy, _scale_xy);

      if (viewVertexLabelsAct->isChecked() ||
          viewMeshVertexLabelsAct->isChecked()) { // Only for debug
        //detri2_draw_vertex_labels(_geom, &painter, _cx, _cy, _scale_xy);
        for (int i = 0; i < _geom->ct_in_vrts; i++) {
          Vertex *v = &(_geom->in_vrts[i]);
          if (v->is_deleted()) continue;
          int x = (int) ( _scale_xy * v->crd[0] + _cx);
          int y = (int) (-_scale_xy * v->crd[1] + _cy);
          //painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
          painter.drawText(x+_draw_vertex_size, y-_draw_vertex_size, QString::number(v->idx));
        }
      }

      //if (_show_geom_tris) {
      if (viewBgMeshAct->isChecked()) {
        QPen pen;
        pen.setColor(qRgb(242, 242, 242)); // bacground mesh edge color
        pen.setWidth(1); // default 1
        painter.setPen(pen);
        detri2_draw_Triangulation(_geom, &painter, _cx, _cy, _scale_xy);
      }
    }

    if (_active_trilist) {
      /* // only for debug
      int count = 0;
      for (int i = 0; i < _geom->tr_tris->used_items; i++) {
        Triang *tri = (Triang *) _geom->tr_tris->get(i);
        if (tri->is_deleted() || tri->is_hulltri()) continue;
        if (tri->is_infected()) count++;
      }
      qDebug()<<"  Found "<<count<<" infected triangles";
      if (count != _active_trilist->objects) {
        qDebug()<<"  Unmatched infected triangles should be = "<< _active_trilist->objects;
      }
      */

      QBrush brush;
      brush.setColor(_color_selected); // red qRgb(255, 0, 0)
      brush.setStyle(Qt::SolidPattern);
      for (int i = 0; i < _active_trilist->objects; i++) {
        Triang** ptri = (Triang **) _active_trilist->get(i);
        Vertex *p1 = (*ptri)->vrt[0];
        Vertex *p2 = (*ptri)->vrt[1];
        Vertex *p3 = (*ptri)->vrt[2];
        //detri2_draw_triangle(p1, p2, p3, &painter, _cx, _cy, _scale_xy);
        int x1 = (int) ( _scale_xy * p1->crd[0] + _cx);
        int y1 = (int) (-_scale_xy * p1->crd[1] + _cy);
        int x2 = (int) ( _scale_xy * p2->crd[0] + _cx);
        int y2 = (int) (-_scale_xy * p2->crd[1] + _cy);
        int x3 = (int) ( _scale_xy * p3->crd[0] + _cx);
        int y3 = (int) (-_scale_xy * p3->crd[1] + _cy);
        QPainterPath path;
        path.moveTo (x1, y1);
        path.lineTo (x2, y2);
        path.lineTo (x3, y3);
        path.lineTo (x1, y1);
        painter.setPen (Qt :: NoPen);
        painter.fillPath (path, brush);
      }
    }

    //if (_show_triangles && _tria && (_tria->tr_tris != NULL)) {
    if (viewTrianglesAct->isChecked() && _tria && (_tria->tr_tris != NULL)) {
      QPen pen;
      pen.setColor(_color_edge); // edge color qRgb(0, 0, 0)
      pen.setWidth(_draw_edge_width); // default 1
      painter.setPen(pen);
      detri2_draw_Triangulation(_tria, &painter, _cx, _cy, _scale_xy);

      //if (_show_vertex_labels) {
      if (viewMeshVertexLabelsAct->isChecked()) {
        QPen pen;
        pen.setColor(qRgb(77, 77, 77)); // dark grey
        pen.setWidth(1); // default
        painter.setPen(pen);
        //QFont serifFont("Times", 12, QFont::Normal); // QFont::Bold
        //QFont sansFont("Helvetica [Cronyx]", 12);
        //painter.setFont(serifFont);
        painter.setFont(_font_meshvertex_label);
        //detri2_draw_vertex_labels(_geom, &painter, _cx, _cy, _scale_xy);

        for (int i = 0; i < _tria->ct_in_vrts; i++) {
          Vertex *v = &(_tria->in_vrts[i]);
          if (v->typ == UNUSEDVERTEX) continue;
          int x = (int) ( _scale_xy * v->crd[0] + _cx);
          int y = (int) (-_scale_xy * v->crd[1] + _cy);
          //painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
          painter.drawText(x+_draw_vertex_size, y-_draw_vertex_size, QString::number(v->idx));
        }
        if (_tria->tr_steiners != NULL) {
          for (int i = 0; i < _tria->tr_steiners->used_items; i++) {
            Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
            if (v->is_deleted()) continue;
            int x = (int) ( _scale_xy * v->crd[0] + _cx);
            int y = (int) (-_scale_xy * v->crd[1] + _cy);
            //painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
            painter.drawText(x+_draw_vertex_size, y-_draw_vertex_size, QString::number(v->idx));
          }
        }
      }
    }

    //if (_show_convex_hull && _tria && (_tria->tr_tris != NULL)) {
    if (viewConvexHullAct->isChecked() && _tria && (_tria->tr_tris != NULL)) {
      QPen pen;
      pen.setColor(qRgb(183, 186, 55)); // edge color
      pen.setWidth(3); // default
      painter.setPen(pen);
      detri2_draw_convex_hull(_tria, &painter, _cx, _cy, _scale_xy);
    }

    //if (_show_ortho_circle && _tria && (_tria->tr_tris != NULL)) {
    if (viewOthocirclesAct->isChecked() && _tria && (_tria->tr_tris != NULL)) {
      QPen pen;
      pen.setColor(_color_ccircle); // light blue qRgb(0,255,255)
      pen.setWidth(_draw_ccircle_width); // default 2
      painter.setPen(pen);
      detri2_draw_orthocircles(_tria, &painter, _cx, _cy, _scale_xy);
    }

    //if (_show_voronoi && _tria && (_tria->tr_tris != NULL)) {
    if ((viewPowerDiagramAct->isChecked() || viewMassCentersAct->isChecked())
        && _tria && (_tria->tr_tris != NULL)) {
      QPen pen;
      pen.setColor(_color_dual_edge); // green qRgb(0,255,0)
      pen.setWidth(_draw_dual_edge_width); // default 2
      painter.setPen(pen);
      if (viewMassCentersAct->isChecked()) {
        QBrush brush;
        brush.setColor(_color_dual_edge);
        brush.setStyle(Qt::SolidPattern);
        painter.setBrush(brush);
        detri2_draw_voronoi_mass_centers(_tria, _geom, viewPowerDiagramAct->isChecked(),
                                         _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
        painter.setBrush(Qt::NoBrush);
      } else {
        detri2_draw_voronoi(_tria, _geom, &painter, _cx, _cy, _scale_xy);
      }
    }

    // Only for debug
    if (viewLaplacianCentersAct->isChecked()) {
      QPen pen;
      pen.setColor(qRgb(0,255,255)); // _color_dual_edge
      pen.setWidth(2); // default 2
      painter.setPen(pen);
      QBrush brush;
      brush.setColor(qRgb(0,255,255));
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      detri2_draw_laplacian_centers(_tria, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
      painter.setBrush(Qt::NoBrush);
    }

    //if (_show_geom_edges && _geom->tr_segs->objects > 0) {
    if (viewSegmentsAct->isChecked() && (_geom->tr_segs->objects > 0)) {
      QPen pen;
      pen.setColor(_color_segment); // segment color qRgb(153, 153, 255)
      pen.setWidth(_draw_segment_width); // default 2
      painter.setPen(pen);
      //detri2_draw_segments(_geom, &painter, _cx, _cy, _scale_xy);
      int x1, y1, x2, y2;
      for (int i = 0; i < _geom->tr_segs->used_items; i++) {
        Triang *seg = (Triang *) _geom->tr_segs->get(i);
        if (seg->is_deleted()) continue;
        x1 = (int) ( _scale_xy * seg->vrt[0]->crd[0] + _cx);
        y1 = (int) (-_scale_xy * seg->vrt[0]->crd[1] + _cy);
        x2 = (int) ( _scale_xy * seg->vrt[1]->crd[0] + _cx);
        y2 = (int) (-_scale_xy * seg->vrt[1]->crd[1] + _cy);
        painter.drawLine(QPoint(x1, y1), QPoint(x2,y2));
      }

      //if (_show_segment_labels && (_geom->tr_segs->objects > 0)) {
      if (viewSegmentLabelsAct->isChecked() && (_geom->tr_segs->objects > 0)) {
        QPen pen;
        pen.setColor(qRgb(0, 0, 204)); // dark blue
        pen.setWidth(1); // default
        painter.setPen(pen);
        QFont serifFont("Times", 12, QFont::Bold); // QFont::Normal
        //QFont sansFont("Helvetica [Cronyx]", 12);
        painter.setFont(serifFont);
        //detri2_draw_segment_labels(_geom, &painter, _cx, _cy, _scale_xy);
        int x1, y1, x2, y2;
        for (int i = 0; i < _geom->tr_segs->used_items; i++) {
          Triang *seg = (Triang *) _geom->tr_segs->get(i);
          if (seg->is_deleted()) continue;
          x1 = (int) ( _scale_xy * seg->vrt[0]->crd[0] + _cx);
          y1 = (int) (-_scale_xy * seg->vrt[0]->crd[1] + _cy);
          x2 = (int) ( _scale_xy * seg->vrt[1]->crd[0] + _cx);
          y2 = (int) (-_scale_xy * seg->vrt[1]->crd[1] + _cy);
          //painter->drawLine(QPoint(x1, y1), QPoint(x2,y2));
          painter.drawText((x1+x2)/2, (y1+y2)/2, QString::number(seg->tag));
        }
      }
    }

    //if (_show_geom_vertices) {
    if (viewVerticesAct->isChecked()) {
      QPen pen;
      pen.setColor(_color_vertex); // vertex color qRgb(102, 102, 102)
      pen.setWidth(1); // default
      painter.setPen(pen);
      QBrush brush;
      brush.setColor(_color_vertex); // qRgb(102, 102, 102)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);
      //detri2_draw_geometry(_geom, &painter, _cx, _cy, _scale_xy);
      for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
        Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        int x = (int) ( _scale_xy * v->crd[0] + _cx);
        int y = (int) (-_scale_xy * v->crd[1] + _cy);
        painter.drawEllipse(QPoint(x,y), _draw_vertex_size, _draw_vertex_size);
      }
      painter.setBrush(Qt::NoBrush);

      //if (_show_vertex_labels) {
      if (viewVertexLabelsAct->isChecked()) {
        QPen pen;
        pen.setColor(qRgb(77, 77, 77)); // dark grey
        pen.setWidth(1); // default
        painter.setPen(pen);
        //QFont serifFont("Times", 12, QFont::Normal); // QFont::Bold
        //QFont sansFont("Helvetica [Cronyx]", 12);
        //painter.setFont(serifFont);
        painter.setFont(_font_vertex_label);
        //detri2_draw_vertex_labels(_geom, &painter, _cx, _cy, _scale_xy);
        for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
          Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
          if (v->is_deleted()) continue;
          int x = (int) ( _scale_xy * v->crd[0] + _cx);
          int y = (int) (-_scale_xy * v->crd[1] + _cy);
          //painter->drawEllipse(QPoint(x,y), ptsize, ptsize);
          painter.drawText(x+_draw_vertex_size, y-_draw_vertex_size, QString::number(v->idx));
        }
      }
    }

    //if (_show_weights) {
    if (viewWeightsAct->isChecked()) {
      QPen pen;
      pen.setColor(qRgb(255,140,0)); // dark orange
      pen.setWidth(2); // default
      painter.setPen(pen);
      //detri2_draw_weights(_geom, _tria, &painter, _cx, _cy, _scale_xy);
      QString text;
      for (int i = 0; i < _geom->tr_steiners->used_items; i++) {
        Vertex *v = (Vertex *) _geom->tr_steiners->get(i);
        if (v->is_deleted()) continue;
        int x = (int) ( _scale_xy * v->crd[0] + _cx);
        int y = (int) (-_scale_xy * v->crd[1] + _cy);
        if (v->wei > 0.0) {
          int r = (int) (sqrt(fabs(v->wei)) * _scale_xy);
          painter.drawEllipse(QPoint(x,y), r, r);
        }
        text.sprintf("%.3f", v->wei);
        painter.drawText(x,y, text);
      }
    }

    //if (_show_geom_subdomains && (_geom->ct_in_sdms > 0)) {
    if (viewSubdomainsAct->isChecked() && (_geom->ct_in_sdms > 0)) {
      QPen pen;
      pen.setColor(qRgb(0, 153, 153));
      pen.setWidth(2); // default
      painter.setPen(pen);
      QFont serifFont("Times", 16, QFont::Normal); // QFont::Bold
      //QFont sansFont("Helvetica [Cronyx]", 12);
      painter.setFont(serifFont);
      //detri2_draw_subdomains(_geom, &painter, _cx, _cy, _scale_xy);
      for (int i = 0; i < _geom->ct_in_sdms; i++) {
        Vertex *v = &(_geom->in_sdms[i]);
        int x = (int) ( _scale_xy * v->crd[0] + _cx);
        int y = (int) (-_scale_xy * v->crd[1] + _cy);
        //painter->drawText(x, y, QString::number(v->tag));
        //QRect rect(x-10, y-10, x+10, y+10);
        painter.drawEllipse(QPoint(x,y),_draw_vertex_size,_draw_vertex_size);
        painter.drawText(x,y, QString::number(v->tag));
        //painter->drawText(rect, Qt::AlignCenter, QString::number(v->tag));
      }
    }

    if (_active_vertex) {
      QPen pen;
      QBrush brush;
      //int ptsize=4;
      pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
      pen.setWidth(1); // default
      painter.setPen(pen);
      brush.setColor(_color_selected); // qRgb(255, 0, 0)
      brush.setStyle(Qt::SolidPattern);
      painter.setBrush(brush);

      Vertex *vrt = _active_vertex;
      while (vrt != NULL) {
        int px = (int) ( _scale_xy * vrt->crd[0] + _cx);
        int py = (int) (-_scale_xy * vrt->crd[1] + _cy);
        //painter.drawEllipse(QPoint(px,py), ptsize, ptsize);
        painter.drawEllipse(QPoint(px,py), _draw_vertex_size, _draw_vertex_size);
        vrt = vrt->Next;
      }

      painter.setBrush(Qt::NoBrush);
    }

    if (_active_segment) {
      QPen pen;
      QBrush brush;
      Triang *seg = _active_segment;
      while (seg != NULL) {
        Vertex *e1 = seg->vrt[0];
        Vertex *e2 = seg->vrt[1];
        pen.setColor(_color_selected); // selected color qRgb(255, 0, 0)
        pen.setWidth(_draw_segment_width); // default 2
        painter.setPen(pen);
        detri2_draw_line(e1, e2, &painter, _cx, _cy, _scale_xy);
        //pen.setColor(qRgb(255, 0, 0)); // selected color
        //pen.setWidth(1); // default
        //painter.setPen(pen);
        brush.setColor(_color_segment); // qRgb(255, 0, 0)
        brush.setStyle(Qt::SolidPattern);
        painter.setBrush(brush);
        detri2_draw_point(e1, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
        detri2_draw_point(e2, _draw_vertex_size, &painter, _cx, _cy, _scale_xy);
        painter.setBrush(Qt::NoBrush);
        seg = seg->Next;
      }
    }
}

// Whenever update() is called (an paintevent is sent). This function will be executed.
void ScribbleArea::paintEvent(QPaintEvent *event)
{
    //qDebug() << "paintEvent!";
    if (!_object_valid) {
      //qDebug() << "draw image";
      // clear everyting.
      //_canvas[1].fill(qRgb(255, 255, 255)); // white
      if (_background_img) {
        _canvas[1] = *_background_img;
      } else {
        _canvas[1].fill(_color_canvas_background);
      }

      QPainter painter(&_canvas[1]);

      draw_scene(painter);

      _canvas[0] = _canvas[1]; // copy to foreground.
      _object_valid = 1;
    } // if (!_object_valid)

    //qDebug() << "show image";
    QPainter thepainter(this);
    QRect dirtyRect = event->rect();
    thepainter.drawImage(dirtyRect, _canvas[0], dirtyRect);
}
